//-----------------------------------------------------------------------------
// GA2K
//
// ISDL, Doshisha University Distributed Genetic Algorithm Software
// Version 1.4.1
//-----------------------------------------------------------------------------
// Parallel Genetic Algorithm Team of Intelligent Systems Design Laborato
// ry, Doshisha University.
//
// Copyright (C) 1999           Ikki Ohmukai,      All rights reserved.
// Copyright (C) 1999,2000,2001 Jun-ichi Yoshida,  All rights reserved.
// Copyright (C) 2001,2002,2003 Tomoyuki Hiroyasu, All rights reserved.
// Copyright (C) 2001,2002,2003 Mitsunori Miki,    All rights reserved.
// Copyright (C) 2002           Masaki Sano,       All rights reserved.
// Copyright (C) 2002,2003      Junji Sawada,      All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The names of the authors should not be used to endorse or promote
//    products derived from this software without specific prior
//    written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
// IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// functions.cpp
//-----------------------------------------------------------------------------

#include <numeric>
#include <cmath>

#include "functions.h"

using namespace std;

static const double PI = 3.1415926535897932384626433832795;


//----------------------------------------------------------------------------
// Ackley [-32.768 32.768) : 0.0
// variable : 0.0
double ackley(const vector<double>& x)
{
	unsigned n = x.size();

	unsigned i = 0;
	const double a = 20.0, b = 0.20, c = 2.0 * PI, E = exp(1.0);
	double val1, val2, sum1, sum2, sum;

	for(val1 = val2 = 0.0, i = 0; i < n; i++) {
		val1 += x[i] * x[i];
		val2 += cos(c * x[i]);
	}

	sum1 = -exp(val2 / n) + E + a;
	sum2 = -exp(-b * sqrt(val1 / n)) * a;
	sum  = sum1 + sum2;

    return -sum;
}


//----------------------------------------------------------------------------
// Rosenbrock [-2.048 2.048) : 0.0
// variable : 1.0
//double rosenbrock(const vector<double>& x)
//{
//	unsigned i;
//
//	double d = 0;
//	for(i = 0; i < x.size() - 1 ; i++ )
//	{
//		double p = x[i + 1] - x[i] * x[i];
//		double q = x[i] - 1;
//		d += p * p * 100 + q * q;
//	}
//	return -d;
//}

//----------------------------------------------------------------------------
// Rosenbrock [-2.048 2.048) : 0.0
// variable : 1.0
double rosenbrock(const vector<double>& x)
{
	unsigned i;

	double d = 0;
	for(i = 1; i < x.size(); i++ )
	{
		double p = x[0] - x[i] * x[i];
		double q = x[i] - 1;
		d += p * p * 100 + q * q;
	}
	return -d;
}


//----------------------------------------------------------------------------
// Rastrigin [-5.12,5.12) : 0.0
// variable : 0.0
double rastrigin(const vector<double>& x)
{
	unsigned n = x.size();
	unsigned i;
	double d;
	for(i = 0, d = 0; i < n; i++)
		d += x[i] * x[i] - 10 * cos(2 * PI * x[i]);
	return -(10 * n + d);

}


//----------------------------------------------------------------------------
// Schwefel [-512 512) : 0.0
// this version of schwefel converges to 0.0, because of hard-coding optimal.
double schwefel(const vector<double>& x)
{
	unsigned n = x.size();
	unsigned i;
	double d;
	for(i = 0, d = 0; i < n; i++)
		d += x[i] * sin(sqrt(fabs(x[i])));
	return ( d - ((418.98288 * n) - 0.00115983) );

//-----------------------------------------------------------------------------
// 10 ӥå / ߷ѿ  ĶϡʲΥɤѤ
//-----------------------------------------------------------------------------
//		static const double schw_ans_10bit
//			= 418.98276401614430142217315733432769775390;
//		static const double schw_ans_20bit
//			= 418.98288727243209450534777715802192687988;
//		static const double schw_ans_30bit
//			= 418.98288727243368612107587978243827819824;
//		static const double schw_ans = schw_ans_10bit;
//
//		return -(d + schw_ans * n);
//-----------------------------------------------------------------------------
}

//----------------------------------------------------------------------------
// Griewank [-512 512) : 0.0
// variable : 0.0
double griewank(const vector<double>& x)
{
	unsigned i, n = x.size();
	double d, e;
	for(i = 0, d = e = 1; i < n; i++){
		d += x[i] * x[i] / 4000;
		e *= cos(x[i] / sqrt(i + 1.0));
	}
	return -(d - e);
}

//----------------------------------------------------------------------------
// Ridge [-64 64) : 0.0
// variable : 0.0
double ridge(const vector<double>& x)
{
	unsigned i, j, n = x.size();
	double d = 0;
	for(i = 0; i < n; i++)
	{
		double p = 0;
		for(j = 0; j <= i; j++)
			p += x[j];
		d += p * p;
	}
	return -d;
}


//----------------------------------------------------------------------------
// Rotated Rastrigin [-5.12,5.12) : 0.0
// variable : 0.0
double rotated_rastrigin(const vector<double>& x)
{
	vector<double> trans(x);
	double cos_value = cos(PI*0.5);
	double sin_value = sin(PI*0.5);
	unsigned n = x.size();

	for(unsigned i=0;i<n-1;i++){
		for(unsigned j=i+1;j<n;j++){
			double temp1 = cos_value*trans[i] - sin_value*trans[j];
			double temp2 = sin_value*trans[i] + cos_value*trans[j];
			trans[i] = temp1;
			trans[j] = temp2;
		}
	}
	double f = (double)10*n;
	for(unsigned i=0;i<n;i++)
		f += trans[i]*trans[i] - 10.0*cos(2.0*PI*trans[i]);

	return -f;
}


//----------------------------------------------------------------------------
// Onemax : 0
// variable : 1
double onemax(const vector<char>& b)
{
	int bit1 = accumulate(b.begin(), b.end(), 0);
	return static_cast<double>( bit1 - static_cast<int>(b.size()) );
}

