//-----------------------------------------------------------------------------
// GA2K
//
// ISDL, Doshisha University Distributed Genetic Algorithm Software
// Version 1.4.1
//-----------------------------------------------------------------------------
// Parallel Genetic Algorithm Team of Intelligent Systems Design Laborato
// ry, Doshisha University.
//
// Copyright (C) 1999           Ikki Ohmukai,      All rights reserved.
// Copyright (C) 1999,2000,2001 Jun-ichi Yoshida,  All rights reserved.
// Copyright (C) 2001,2002,2003 Tomoyuki Hiroyasu, All rights reserved.
// Copyright (C) 2001,2002,2003 Mitsunori Miki,    All rights reserved.
// Copyright (C) 2002           Masaki Sano,       All rights reserved.
// Copyright (C) 2002,2003      Junji Sawada,      All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The names of the authors should not be used to endorse or promote
//    products derived from this software without specific prior
//    written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
// IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// ga2k.h
//-----------------------------------------------------------------------------

#ifndef _GA2K_H_
#define _GA2K_H_


#include <vector>
#include <fstream>
using namespace std;

#include "mtrand.h"
#include "randring.h"
#include "interface.h"

#define INITIAL_POPULATION_FILE "design_variable.ini"
#define LOG_FILE "ga2k.log"
#define FITNESS_FILE "fitness.out"
#define VARIABLE_FILE "variable.out"

class Evaluation;

///////////////////////////////////////////////////////////////////////////////
// global variable
///////////////////////////////////////////////////////////////////////////////

enum FunctionType {
	Rastrigin = 0,
	Schwefel = 1,
	Griewank = 2,
	Rosenbrock = 3,
	Ridge = 4,
	Onemax = 5,
	User = 99
};

enum SelectionType {
	Tournament = 0,
	Roulette = 1
};

extern ObjectiveProblem problem;
extern Evaluation* eval;


///////////////////////////////////////////////////////////////////////////////
// class Gene
///////////////////////////////////////////////////////////////////////////////

class Gene
{
  public:
	//data
	vector<char>	bitString;
	double			fitness;
	vector<double>	variable;

	//operator
	bool operator<(const Gene& gene) const;

	//method (constructor)
	Gene();
	~Gene();
	void initialize(int geneLength);
	void initialize(int geneLength, char value);
	void initialize(int geneLength, const vector<char>& code);

	//method (operator)
	void visualize();
};


///////////////////////////////////////////////////////////////////////////////
// class Evaluation
///////////////////////////////////////////////////////////////////////////////

class Evaluation
{
  public:
	//data
	int				numEvaluate;

	int				geneLength;
	int				variableLength;
	int				numVariable;

	FunctionType	functionType;
	double			openMax;
	double			closedMin;
	double			delta;

	int				typeOfEncode;

	//method (constructor)
	Evaluation();
	virtual ~Evaluation();
	virtual void setting(int argc, char *argv[]);  //

	//method
	virtual void initialize();         //GAλϤ˼¹Ԥ
	virtual void start();              //ɾ˼¹Ԥ
	virtual void fitness(Gene& gene);  //ɾ׻Ԥ
	virtual void end();                //ɾθ˼¹Ԥ
	virtual void finalize();           //GAν˼¹Ԥ


  protected:
	//method
	vector<double> decode(const Gene& gene);


  private:
	//data
	vector<Gene*>	pool;

	//method
	void evaluate(Gene& gene);
};


///////////////////////////////////////////////////////////////////////////////
// class Island
///////////////////////////////////////////////////////////////////////////////

class Island
{
  public:
	//data
	int		populationSize;

	double	crossoverRate;
	double	mutationRate;

	int		numElite;

	double	migrationRate;
	int		migrantSize;
	int		nativeSize;
	int 	tournamentSize;

	int		geneLength;
	int		variableLength;
	int		numVariable;
	double	openMax;
	double	closedMin;
	double	delta;
	double	finValue;

	FunctionType functionType;
	int		typeOfEncode;
	int		numCrossoverPoint;
	SelectionType selectionType;

	mutable int		numEvaluate;

	double			bestFitness;
	vector<double>	bestVariable;

	vector<double>	roulette;
	vector<int>		crossoverTable;

//gene
	vector<Gene>	gene;
	vector<Gene>	migrant;
	vector<Gene>	elite;
	vector<Gene>	newGeneration;

//method (constructor)
	Island();
	~Island();
	void initialize(std::ifstream& fin);
	void initializeElite();

//method (operator)
	void immigrate();
	void crossover();
	void nPointCrossover();
	void mutate();
	void evaluate();
	void reproduce();
	void rouletteSelection();
	void tournamentSelection();
	void elitePreservation();
	void setGeneFromFile(std::ifstream& fin);
};


///////////////////////////////////////////////////////////////////////////////
// class Globe
///////////////////////////////////////////////////////////////////////////////

class Globe
{
  public:
	//data
	vector<Island>	island;
	Island			island_0;

	int					generation;
	int					maxGeneration;
	int					convGeneration;

	int					migrationInterval;
	int					numIsland;
	int					populationSize;
	int					totalPopulationSize;
	int					migrantSize;
	int					nativeSize;

	int 				fInitialPopulation;

	double				maxFitness;
	vector<double>		bestVariable;
	int					maxTrial;
	double				sumMaxFitness;
	int					sumGeneration;
	int					numConvergent;
	int					totalEvaluate;
	int					sumTotalEvaluate;

	vector<double>		sumFitness;

	Gene				elite;
	bool				fStop;
	bool				fConv;
	bool				fTotalPopSize;
	bool				fOutput;

	//method (constructor)
	Globe();
	virtual ~Globe();
	virtual void initialize();
	virtual void initialize(int argc, char *argv[]);
	virtual void reviseParameters();
	virtual void establish();

	//method (operator)
	virtual void emigrate();
	virtual void execute();
	virtual void showParameters() const;
	virtual void finalize(bool b);


  protected:
	//method
	void evaluateIslands();
	void dumpProgress();


  private:
	//data
	ofstream fLog;
	ofstream fFitness;
	ofstream fVariable;
};


#endif //_GA2K_H_
