//-----------------------------------------------------------------------------
// GA2K
//
// ISDL, Doshisha University Distributed Genetic Algorithm Software
// Version 1.4.1
//-----------------------------------------------------------------------------
// Parallel Genetic Algorithm Team of Intelligent Systems Design Laborato
// ry, Doshisha University.
//
// Copyright (C) 1999           Ikki Ohmukai,      All rights reserved.
// Copyright (C) 1999,2000,2001 Jun-ichi Yoshida,  All rights reserved.
// Copyright (C) 2001,2002,2003 Tomoyuki Hiroyasu, All rights reserved.
// Copyright (C) 2001,2002,2003 Mitsunori Miki,    All rights reserved.
// Copyright (C) 2002           Masaki Sano,       All rights reserved.
// Copyright (C) 2002,2003      Junji Sawada,      All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The names of the authors should not be used to endorse or promote
//    products derived from this software without specific prior
//    written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
// IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// ga2k_mpi.cpp
//-----------------------------------------------------------------------------

#include <cstdio>
#include <cstdlib>
#include <cfloat>
#include <vector>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>
using namespace std;

#include <mpi.h>
#include "ga2k_mpi.h"
#include "randring.h"

#define GENE_DATA     1000
#define FITNESS_DATA  2000
#define VARIABLE_DATA 3000


//=============================================================================
//functions
//=============================================================================

static int send_gene(const Gene& gene, int dest)
{
	void *buf = (void*)&(gene.bitString.front());
	int count = gene.bitString.size();

	return MPI_Send(buf, count, MPI_CHAR, dest, GENE_DATA, MPI_COMM_WORLD);
}

static int recv_gene(Gene& gene)
{
	void *buf = (void*)&(gene.bitString.front());
	int count = gene.bitString.size();
	MPI_Status stat;

	return MPI_Recv(buf, count, MPI_CHAR, MPI_ANY_SOURCE, GENE_DATA,
		MPI_COMM_WORLD, &stat);
}


//=============================================================================
//class MPIGlobe
//=============================================================================

MPIGlobe::MPIGlobe(int *pargc, char*** pargv) : destination(2)
{
	MPI_Init(pargc, pargv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
	MPI_Comm_size(MPI_COMM_WORLD, &nProcs);
	int namelen = 0;
	MPI_Get_processor_name(hostname, &namelen);

	fprintf(stderr, "Process %d on %s\n", myRank, hostname);
	root = 0;
	synchronize();
}

MPIGlobe::~MPIGlobe()
{
	MPI_Finalize();
}

void MPIGlobe::initialize()
{
// Default settings

	Globe::initialize();

	numIsland					= nProcs;
	island_0.populationSize  	= totalPopulationSize / numIsland;// 10;
}


void MPIGlobe::initialize(int argc, char *argv[])
{
	this->initialize();
	Globe::initialize(argc, argv);

	// revise numIsland
	numIsland = nProcs;
}

void MPIGlobe::establish()
{
	if(fTotalPopSize)
	{
		island_0.populationSize = totalPopulationSize / numIsland;
		island_0.numElite = island_0.populationSize / 2;
	}

	fTotalPopSize = false;
	numIsland = 1;
	Globe::establish();
	numIsland = nProcs;

	// broadcast seed for random ring
	int seed = genrand_int(INT_MAX);
	MPI_Bcast(&seed, 1, MPI_INT, root, MPI_COMM_WORLD);
	destination.srand(seed);
	destination.resize(numIsland);

	synchronize();
}

void MPIGlobe::emigrate()
{
	if(migrantSize == 0 || nProcs < 2)
		return;

	Island& island = (this->island)[0];

	// Ԥʰܽȥݥϥ󥰡
	destination.shuffle();

	// ܽΤϥ
	random_shuffle(island.gene.begin(), island.gene.end(), genrand_int);

	// Ԥ migrant ˸Τ򥳥ԡ
	int dest = destination[myRank];
	for(int j = 0; j < migrantSize; j++){
		send_gene(island.gene[nativeSize + j], dest);
		recv_gene(island.migrant[j]);
	}
}

void MPIGlobe::execute()
{
	Island& island = (this->island)[0];

	for(generation = 0; generation < maxGeneration; generation++)
	{
		if(generation % 10 == 0)
			if(myRank == root)
				cerr << "." << flush;

		island.reproduce();

		if(generation % migrationInterval == 0 && numIsland > 1){
			emigrate();
			island.immigrate();
		}

		island.crossover();
		island.mutate();

		eval->start();
		island.evaluate();
		eval->end();

		island.elitePreservation();

		setMaxFitness();

		if(!fConv && maxFitness > island_0.finValue)
			finalize(true);

		sumFitness[generation] += maxFitness;

		if(!fConv && myRank == root)
			dumpProgress();
	}
	finalize(false);
}


void MPIGlobe::finalize(bool b)
{
	if(b)
	{
		eval->finalize();

		maxFitness = 0.000000;
		sumMaxFitness += maxFitness;
		convGeneration = generation + 1;
		sumGeneration += convGeneration;
		numConvergent++;
		if(myRank == root) cerr << "*" << flush;
		fConv = true;

		if(fStop)
			for(; generation < maxGeneration; generation++){
				sumFitness[generation] += maxFitness;
				if(myRank == root)
					dumpProgress();
			}

		MPI_Reduce(&(eval->numEvaluate), &(totalEvaluate), 1, MPI_INT, MPI_SUM, root, MPI_COMM_WORLD);
		MPI_Bcast(&totalEvaluate, 1, MPI_INT, root, MPI_COMM_WORLD);

		sumTotalEvaluate += totalEvaluate;
	}
	else if(!b && !fConv)
	{
		eval->finalize();

		sumMaxFitness += maxFitness;
		MPI_Reduce(&(eval->numEvaluate), &(totalEvaluate), 1, MPI_INT, MPI_SUM, root, MPI_COMM_WORLD);
		MPI_Bcast(&totalEvaluate, 1, MPI_INT, root, MPI_COMM_WORLD);
		sumTotalEvaluate += totalEvaluate;
	}
}

//¹˥ѥ᡼cerrǽϤ
void MPIGlobe::showParameters()
{
	if(myRank == root){
		Globe::showParameters();
	}
	synchronize();
}

void MPIGlobe::synchronize() const
{
	MPI_Barrier(MPI_COMM_WORLD);
}

void MPIGlobe::setMaxFitness()
{
	Island& island = (this->island)[0];
	int bestProcess;
	MPI_Status stat;

	if(myRank == root){
		double fitness;

		maxFitness = island.bestFitness;
		bestProcess = 0;
		//ƥץ顤ɾͼ
		for(int i = 1; i < nProcs; i++){
			MPI_Recv(&fitness, 1, MPI_DOUBLE, i, FITNESS_DATA, MPI_COMM_WORLD, &stat);
			if(fitness > maxFitness){
				maxFitness = fitness;
				bestProcess = i;
			}
		}
	}
	else{
		//ɾͤ
		MPI_Send(&(island.bestFitness), 1, MPI_DOUBLE, root, FITNESS_DATA, MPI_COMM_WORLD);
	}

	MPI_Bcast(&maxFitness, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
	MPI_Bcast(&bestProcess, 1, MPI_INT, root, MPI_COMM_WORLD);

	if(myRank == root){
		bestVariable = island.bestVariable;
		//ɾͤΥץ顤߷ѿͼ(ʬʳä)
		if(bestProcess != root)
			MPI_Recv(&(bestVariable.front()), island.bestVariable.size(), MPI_DOUBLE, bestProcess, VARIABLE_DATA, MPI_COMM_WORLD, &stat);
	}
	else{
		//߷ѿͤ׵ᤵƤ顤
		if(bestProcess == myRank)
			MPI_Send(&(island.bestVariable.front()), island.bestVariable.size(), MPI_DOUBLE, root, VARIABLE_DATA, MPI_COMM_WORLD);
	}
}
