//-----------------------------------------------------------------------------
// GA2K
//
// ISDL, Doshisha University Distributed Genetic Algorithm Software
// Version 1.4.1
//-----------------------------------------------------------------------------
// Parallel Genetic Algorithm Team of Intelligent Systems Design Laborato
// ry, Doshisha University.
//
// Copyright (C) 1999           Ikki Ohmukai,      All rights reserved.
// Copyright (C) 1999,2000,2001 Jun-ichi Yoshida,  All rights reserved.
// Copyright (C) 2001,2002,2003 Tomoyuki Hiroyasu, All rights reserved.
// Copyright (C) 2001,2002,2003 Mitsunori Miki,    All rights reserved.
// Copyright (C) 2002           Masaki Sano,       All rights reserved.
// Copyright (C) 2002,2003      Junji Sawada,      All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The names of the authors should not be used to endorse or promote
//    products derived from this software without specific prior
//    written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
// IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// ga2k_ms.cpp
//-----------------------------------------------------------------------------

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cfloat>
#include <vector>
#include <iostream>
using namespace std;

#include <mpi.h>
#include "ga2k_ms.h"

///////////////////////////////////////////////////////////////////////////////
// function
///////////////////////////////////////////////////////////////////////////////

static void synchronize()
{
	MPI_Barrier(MPI_COMM_WORLD);
}


//=============================================================================
//class MasterSlaveGlobe
//=============================================================================

void MasterSlaveGlobe::initialize()
{
	Globe::initialize();

	MasterSlaveEvaluation* mse = dynamic_cast<MasterSlaveEvaluation*>(eval);
	root	= mse->root;
	nProcs	= mse->nProcs;
	myRank	= mse->myRank;
	strcpy(hostname, mse->hostname);
}

void MasterSlaveGlobe::initialize(int argc, char *argv[])
{
	this->initialize();
	Globe::initialize(argc, argv);
}

void MasterSlaveGlobe::establish()
{
	//MasterˤϡƤΥѥ᡼ꤹ
	if(myRank == root){
		Globe::establish();
	}
	//Slaveˤϡ˴ؤȡ˴ؤФ
	else{
		maxFitness		= -DBL_MAX;
		fConv			= false;

		eval->geneLength		= island_0.geneLength;
		eval->variableLength	= island_0.variableLength;
		eval->numVariable		= island_0.numVariable;
		eval->functionType		= island_0.functionType;
		eval->openMax			= island_0.openMax;
		eval->closedMin			= island_0.closedMin;
		eval->delta				= island_0.delta;
		eval->typeOfEncode		= island_0.typeOfEncode;

		island.resize(0);

		eval->initialize();
		evaluateIslands();
	}

	synchronize();
}

void MasterSlaveGlobe::execute()
{
	for(generation = 0; generation < maxGeneration; generation++)
	{
		if(generation % 10 == 0)
			if(myRank == root)
				cerr << "." << flush;

		if(myRank == root){
			// selection
			for(unsigned i = 0; i < island.size(); i++)
				island[i].reproduce();

			if(generation % migrationInterval == 0)
				emigrate();

			for(unsigned i = 0; i < island.size(); i++)
			{
				if(generation % migrationInterval == 0 && numIsland > 1)
					island[i].immigrate();

				island[i].crossover();
				island[i].mutate();
			}
		}

		evaluateIslands();

		if(myRank == root)
			for(unsigned i = 0; i < island.size(); i++)
			{
				island[i].elitePreservation();

				if(island[i].bestFitness > maxFitness){
					maxFitness = island[i].bestFitness;
					bestVariable = island[i].bestVariable;
				}
			}

		//SlaveǽλȽѤ뤿˺Ŭ٤ã
		MPI_Bcast(&maxFitness, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);

		if(!fConv && maxFitness > island_0.finValue)
			finalize(true);

		sumFitness[generation] += maxFitness;

		if(!fConv && myRank == root)
			dumpProgress();
	}
	finalize(false);
}

void MasterSlaveGlobe::finalize(bool b)
{
	if(b)
	{
		eval->finalize();

		maxFitness = 0.000000;
		sumMaxFitness += maxFitness;
		convGeneration = generation + 1;
		sumGeneration += convGeneration;
		numConvergent++;
		if(myRank == root) cerr << "*" << flush;
		fConv = true;

		if(fStop)
			for(; generation < maxGeneration; generation++){
				sumFitness[generation] += maxFitness;
				if(myRank == root)
					dumpProgress();
			}

		totalEvaluate = eval->numEvaluate;
		sumTotalEvaluate += totalEvaluate;
	}
	else if(!b && !fConv)
	{
		eval->finalize();

		sumMaxFitness += maxFitness;
		totalEvaluate = eval->numEvaluate;
		sumTotalEvaluate += totalEvaluate;
	}
}

void MasterSlaveGlobe::showParameters()
{
	if(myRank == root){
		Globe::showParameters();
	}
	synchronize();
}

//=============================================================================
//class MasterSlaveEvaluation
//=============================================================================

MasterSlaveEvaluation::MasterSlaveEvaluation(int* pargc, char*** pargv)
{
	MPI_Init(pargc, pargv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
	MPI_Comm_size(MPI_COMM_WORLD, &nProcs);
	int namelen = 0;
	MPI_Get_processor_name(hostname, &namelen);

	fprintf(stderr, "Process %d on %s\n", myRank, hostname);
	root = 0;
	numSlave = nProcs - 1;

	numSendMax.resize(numSlave);
	slaveTable.resize(numSlave);
	setNumSendMax(1);

	synchronize();
}

void MasterSlaveEvaluation::setNumSendMax(int num)
{
	for(int i = 0; i < numSlave; i++)
		numSendMax[i] = num;
	for(int i = 0; i < numSlave; i++)
		slaveTable[i].resize(numSendMax[i]);
}

MasterSlaveEvaluation::~MasterSlaveEvaluation()
{
	MPI_Finalize();
}

void MasterSlaveEvaluation::setting(int argc, char *argv[])
{
	Evaluation::setting(argc, argv);

	for(int i = 1; i < argc; i++)
	{
		const char *p = argv[i] + 1;
		switch(*argv[i])
		{
		case 'N':
			setNumSendMax(atoi(p));
			break;
		}
	}
}

void MasterSlaveEvaluation::initialize()
{
	Evaluation::initialize();
}

void MasterSlaveEvaluation::start()
{
	pool.clear();
}

void MasterSlaveEvaluation::fitness(Gene& gene)
{
	numEvaluate++;

	pool.push_back(&gene);
}

void MasterSlaveEvaluation::sendGenes(int dest)
{
	int numGene = pool.size();
	int numSend = 0;
	vector<int> sendList;

	//ꤲΤ
	while(currentIndex < numGene && numSend < numSendMax[dest-1]){
		sendList.push_back(currentIndex);
		slaveTable[dest-1][numSend] = currentIndex;
		numSend++;
		currentIndex++;
	}

	int count = geneLength * numSend;
	char* buf = new char[count];

	//Ҿchar˥ѥå
	for(int i = 0; i < numSend; i++)
		memcpy(buf + geneLength * i, &(pool[sendList[i]]->bitString.front()), geneLength);

	MPI_Send(buf, count, MPI_CHAR, dest, GENE_DATA, MPI_COMM_WORLD);

	delete [] buf;

	sendOut += numSend;
}

void MasterSlaveEvaluation::getResults(MPI_Status stat)
{
	int sender = stat.MPI_SOURCE;
	int num, count;
	double *fitness, *varSequence;

	//ƤȽ
	MPI_Get_count(&stat, MPI_DOUBLE, &num);
	//֤äƤɾͤΤ˳Ǽ
	fitness = new double[num];
	MPI_Recv(fitness, num, MPI_DOUBLE, sender, FITNESS_DATA, MPI_COMM_WORLD, &stat);
	for(int i = 0; i < num; i++)
		pool[slaveTable[sender-1][i]]->fitness = fitness[i];
	//֤äƤ߷ѿͤΤ˳Ǽ
	MPI_Probe(sender, MPI_ANY_TAG, MPI_COMM_WORLD, &stat);
	MPI_Get_count(&stat, MPI_DOUBLE, &count);
	varSequence = new double[count];
	numVariable = count / num;
	MPI_Recv(varSequence, count, MPI_DOUBLE, sender, VARIABLE_DATA, MPI_COMM_WORLD, &stat);
	for(int i = 0; i < num; i++){
		int index = slaveTable[sender-1][i];
		pool[index]->variable.resize(numVariable);
		for(int j = 0; j < numVariable; j++)
			pool[index]->variable[j] = varSequence[i * numVariable + j];
	}
	delete [] varSequence;
	delete [] fitness;
	sendOut -= num;
}

//Masterν ҤSlaveꤲ֤äƤ̤Ǽ
void MasterSlaveEvaluation::masterPart()
{
	int numGene = pool.size();
	int sig = -1;
	MPI_Status stat;

	currentIndex = 0;
	sendOut = 0;

	//ޤSlave˸Τ꿶
	for(int i = 1; i <= numSlave; i++)
		sendGenes(i);

	//sendOutä֤äƤƤʤĤĤäƤ롼
	while(sendOut > 0){
		//ɤSlaveƤΤȽ
		MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &stat);
		//̼
		getResults(stat);

		//ͤᤷƤSlave˼ΥǡĤ
		if(currentIndex < numGene){
			sendGenes(stat.MPI_SOURCE);
		}
	}

	//Slaveɾ׻λȤã
	for(int i = 1; i <= numSlave; i++)
		MPI_Send(&sig, 1, MPI_INT, i, EVAL_FINISH, MPI_COMM_WORLD);
}

void MasterSlaveEvaluation::getGenes(MPI_Status stat, vector<Gene>& gene)
{
	int count, numGene;

	MPI_Get_count(&stat, MPI_CHAR, &count);
	numGene = count / geneLength;
	gene.resize(numGene);
	for(int i = 0; i < numGene; i++)
		gene[i].initialize(geneLength);
	char* buf = new char[count];
	MPI_Recv(buf, count, MPI_CHAR, root, GENE_DATA, MPI_COMM_WORLD, &stat);
	for(int i = 0; i < numGene; i++)
		memcpy(&(gene[i].bitString.front()), &buf[geneLength * i], geneLength);
	delete [] buf;
}

void MasterSlaveEvaluation::sendResults(vector<Gene>& gene)
{
	int numGene = gene.size();
	double* fitness;
	double* varSequence;

	//ǡν
	fitness = new double[numGene];
	for(int i = 0; i < numGene; i++)
		fitness[i] = gene[i].fitness;
	numVariable = gene[0].variable.size();
	varSequence = new double[numVariable * numGene];
	for(int i = 0; i < numGene; i++)
		for(int j = 0; j < numVariable; j++)
			varSequence[i * numVariable + j] = gene[i].variable[j];
	//ɾ֤ͤ
	MPI_Send(fitness, numGene, MPI_DOUBLE, root, FITNESS_DATA, MPI_COMM_WORLD);
	//߷ѿ֤ͤ
	MPI_Send(varSequence, numGene * numVariable, MPI_DOUBLE, root, VARIABLE_DATA, MPI_COMM_WORLD);
	delete [] varSequence;
	delete [] fitness;
}

//Slaveν Masterꤲ줿Ҥɾơ̤֤
void MasterSlaveEvaluation::slavePart()
{
	vector<Gene> gene;
	MPI_Status stat;
	int sig;

	//Masterɾ׵᤬
	MPI_Probe(root, MPI_ANY_TAG, MPI_COMM_WORLD, &stat);
	while(stat.MPI_TAG == GENE_DATA){
		getGenes(stat, gene);
		//äΤɾ
		Evaluation::start();
		for(int i = 0; i < gene.size(); i++)
			Evaluation::fitness(gene[i]);
		Evaluation::end();
		sendResults(gene);

		//λ
		MPI_Probe(root, MPI_ANY_TAG, MPI_COMM_WORLD, &stat);
	}

	//Master顤ɾ׻äȤ
	MPI_Recv(&sig, 1, MPI_INT, root, EVAL_FINISH, MPI_COMM_WORLD, &stat);
}

void MasterSlaveEvaluation::end()
{
	if(myRank == root)
		masterPart();
	else
		slavePart();
}

void MasterSlaveEvaluation::finalize()
{
	Evaluation::finalize();
}
