//-----------------------------------------------------------------------------
// GA2K
//
// ISDL, Doshisha University Distributed Genetic Algorithm Software
// Version 1.4.1
//-----------------------------------------------------------------------------
// Parallel Genetic Algorithm Team of Intelligent Systems Design Laborato
// ry, Doshisha University.
//
// Copyright (C) 1999           Ikki Ohmukai,      All rights reserved.
// Copyright (C) 1999,2000,2001 Jun-ichi Yoshida,  All rights reserved.
// Copyright (C) 2001,2002,2003 Tomoyuki Hiroyasu, All rights reserved.
// Copyright (C) 2001,2002,2003 Mitsunori Miki,    All rights reserved.
// Copyright (C) 2002           Masaki Sano,       All rights reserved.
// Copyright (C) 2002,2003      Junji Sawada,      All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The names of the authors should not be used to endorse or promote
//    products derived from this software without specific prior
//    written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
// IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// ga2k_ns.cpp
//-----------------------------------------------------------------------------

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <iostream>
#include <fstream>
using namespace std;

extern "C" {
#include <netsolve.h>
}
#include "ga2k_ns.h"

//=============================================================================
//class NetSolveEvaluation
//=============================================================================

NetSolveEvaluation::NetSolveEvaluation()
{
	strcpy(problemName, "rastrigin()");

	method = Nonblocking;
	//method = Blocking;
	//method = Farming;
}

NetSolveEvaluation::~NetSolveEvaluation()
{

}

void NetSolveEvaluation::setting(int argc, char *argv[])
{
	Evaluation::setting(argc, argv);

	for(int i = 1; i < argc; i++)
	{
		const char *p = argv[i] + 1;
		switch(*argv[i])
		{
		case 'F':
			strcpy(problemName, p);
			strcat(problemName, "()");
			break;

		case 'N':
			switch(*p)
			{
			case 'b':
				method = Blocking;
				break;

			case 'n':
				method = Nonblocking;
				break;

			case 'f':
				method = Farming;
				break;
			}
		}
	}
}

void NetSolveEvaluation::initialize()
{
	Evaluation::initialize();
}

void NetSolveEvaluation::start()
{
	pool.clear();
}

void NetSolveEvaluation::fitness(Gene& gene)
{
	numEvaluate++;

	pool.push_back(&gene);
}

void NetSolveEvaluation::RPC_blocking()
{
	int numGene = pool.size();
	int status;

	for(int i = 0; i < numGene; i++){
		Gene& gene = *pool[i];
		status = netsl(problemName, &(gene.variable.front()), gene.variable.size(), &(gene.fitness));
		if(status != NetSolveOK){
			cerr << "NetSolve RPC failed" << endl;
			netslerr(status);
			cerr << "Exit" << endl;
			exit(1);
		}
	}
}

void NetSolveEvaluation::RPC_farming()
{
	int numGene = pool.size();
	vector<double*> var_arr;
	vector<int> dim_arr;
	vector<double*> fit_arr;
	char itr[16];

	var_arr.resize(numGene);
	dim_arr.resize(numGene);
	fit_arr.resize(numGene);
	for(int i = 0; i < numGene; i++){
		Gene& gene = *pool[i];
		var_arr[i] = &(gene.variable.front());
		dim_arr[i] = gene.variable.size();
		fit_arr[i] = &(gene.fitness);
	}
	sprintf(itr, "i=0,%d", numGene - 1);

	int* status = netsl_farm(itr, problemName, 
           ns_ptr_array((void**)&(var_arr.front()),"$i"), 
           ns_int_array((int*)&(dim_arr.front()),"$i"), 
           ns_ptr_array((void**)&(fit_arr.front()),"$i"));

	if(status[0] != 0){
		cerr << "NetSolve Farming failed" << endl;
		for(int i = 1; i <= numGene; i++){
			if(status[i] != NetSolveOK)
				cerr << "RPC " << i << ": " << netsolveErrorMessage(status[i]) << endl;
		}
		cerr << "Exit" << endl;
		exit(1);
	}
	free(status);
}

void NetSolveEvaluation::RPC_nonblocking()
{
  int i, j, retr, numGene = pool.size(), window_sz = 10, ready_idx;
  vector<double*> var_arr;
  vector<int> dim_arr;
  vector<double*> fit_arr;
  vector<int> status_arr;

  char itr[16];

  var_arr.resize(numGene);
  dim_arr.resize(numGene);
  fit_arr.resize(numGene);
  status_arr.resize(window_sz);

  for(i = 0; i < numGene; i++){
    Gene& gene = *pool[i];
    var_arr[i] = &(gene.variable.front());
    dim_arr[i] = gene.variable.size();
    fit_arr[i] = &(gene.fitness);
  }

  for(i=0; i < window_sz; i++) {
    status_arr[i] = netslnb(problemName, var_arr[i], dim_arr[i], fit_arr[i]);

    if(status_arr[i] < 0) {
      netslerr(status_arr[i]);
      return;
    }
  }

  retr = 0;

  while(retr < numGene) {
    ready_idx = -1;

    for(j=0;j<window_sz;j++) {
      if(status_arr[j] >= 0) {
        if(netslpr(status_arr[j]) == NetSolveOK) {
          ready_idx = j;
          break;
        }
      }
    }

    if(ready_idx >= 0) {
      if(netslwt(status_arr[ready_idx]) == NetSolveOK) {
        if(i < numGene) {
          status_arr[ready_idx] = netslnb(problemName, var_arr[i], dim_arr[i], fit_arr[i]);
          if(status_arr[ready_idx] < 0)
            netslerr(status_arr[i]);
          i++;
        }
        else
          status_arr[ready_idx] = -1;
      }
      else {
        status_arr[ready_idx] = -1;
      }
    
      retr++;
    }
  }
}

void NetSolveEvaluation::decodeAll()
{
	int numGene = pool.size();

	for(int i = 0; i < numGene; i++){
		Gene& gene = *pool[i];
		gene.variable = decode(gene);
	}
}

void NetSolveEvaluation::end()
{
	decodeAll();

	switch(method){
	case Nonblocking : RPC_nonblocking(); break;
	case Blocking    : RPC_blocking(); break;
	case Farming     : RPC_farming();  break;
	default :
		cerr << endl << "invalid RPC method" << endl << "Exit" << endl;
		exit(1);
	}
}

void NetSolveEvaluation::finalize()
{
	Evaluation::finalize();
}
