/**
 * @file
 *
 * This file contains structures and function prototypes for
 * the GridRPC client.
 *
 */
/* $Id: grpc.h,v 1.56 2008/12/11 02:00:18 tbrady Exp $ */
/* $UTK_Copyright: $ */

#ifndef _GRPC_H_
#define _GRPC_H_

#ifdef __cplusplus
extern "C" {
#endif


#define MAX_GRPC_REQUESTS 200
#define GRPC_USER_INFO_LEN 256
#define GS_ARG_SIZE_THRESH (64*1024)
#define AGENT_IP_EXPIRATION 3600

struct grpc_function_handle_s;

#include "portability.h"
#include "grpc_error.h"
#include "utility.h"
#include "proxylib.h"
#include "comm_basics.h"
#include "comm_data.h"
#include "comm_encode.h"
#include "general.h"
#include "problem.h"

#include "dsi.h"

#ifdef GS_SMART_GRIDSOLVE
#include "gs_smart_clib.h"
#endif

typedef int grpc_sessionid_t;

/** This structure represents a mapping between the function name
 * and a server that can execute it.  To support retry-based
 * fault tolerance, we actually store a list of servers capable
 * of executing the function, but it's transparent to the user.
 */
typedef struct grpc_function_handle_s {
  char *func_name;
  int num_servers;
  int srv_idx;
  gs_server_t **server_list;
  gs_problem_t *problem_desc;
  int bind_servers_at_call_time;
  char *criteria;
  int agent_taskid;
  int num_calls;
#ifdef GS_SMART_GRIDSOLVE
  /* if there are no servers to execute the task then this is -1 otherwise 0*/
   int valid_mapping;
#endif
} grpc_function_handle_t;

/** This structure represents an 'argument stack', which allows
 * building the arguments to a function dynamically at run-time
 * by pushing the arguments onto a stack.
 */
typedef struct {
  void **args;
  int max_size;
  int top;
} grpc_arg_stack;

/** This structure represents a previously submitted request.
 * We need to keep the handle and the problem structures so
 * that the function call can be resubmitted in case of
 * a failure in the original server or network.
 */
typedef struct grpc_request_s {
  grpc_function_handle_t *handle;
  gs_problem_t *problem;
  char *request_id;
  pid_t s_pid;
} grpc_request_t;

/**
 * This structure is used for storing the amount of time each part
 * of the transaction takes.
 **/
typedef struct {
  struct timeval gtv1, gtv2;

  double proxy_start;
  double object_init;
  double agent_comm;
  double send_input;
  double job_complete;
  double recv_output;
  double measure_comm;
} grpc_profile_t;

/**
 **  STRUCTURE DEFINITIONS for Task Farming
 **/

/* Iterator */

typedef struct {
  char *expression;
} grpc_int_iterator_t;

typedef struct {
  int *array;
  char *expression;
} grpc_int_array_iterator_t;

typedef struct {
  void **array;
  char *expression;
} grpc_ptr_array_iterator_t;

typedef struct {
  int type;
  union {
    grpc_int_iterator_t *int_iterator;
    grpc_int_array_iterator_t *int_array_iterator;
    grpc_ptr_array_iterator_t *ptr_array_iterator;
  } it;
} grpc_specific_iterator_t;

typedef struct {
  int returned_type;
  int i;
  grpc_specific_iterator_t *specific;
} grpc_iterator_t;

typedef union {
  int i;
  void *ptr;
} grpc_void_star_or_int_t;

typedef struct {
  struct timeval tv;
  double comm_bw;
} grpc_cache_entry_t;

typedef struct {
  char *hostname;
  int port;
  ipaddr_t ipaddr;
  struct timeval creation_time;
} agent_host_info_t;

#define GRPC_RETURN(err, minor_err) \
  { \
    grpc_errno = err; \
    grpc_minor_errno = minor_err; \
    return err; \
  }

/* Benchmarking macros */
#define GRPC_TIMER_CLEAR(p) \
  ((p) ? ((p)->gtv1.tv_sec = (p)->gtv1.tv_usec = \
    (p)->gtv2.tv_sec = (p)->gtv2.tv_usec = 0) : 0)
#define GRPC_TIMER_START(p) \
  ((p) ? gettimeofday(&((p)->gtv1), (struct timezone*)0) : 0)
#define GRPC_TIMER_ELAPSED(p) \
  ((p) ? ((((p)->gtv2.tv_usec-(p)->gtv1.tv_usec)+ \
     (((p)->gtv2.tv_sec-(p)->gtv1.tv_sec)*1000000))/1000000.0) : 0)
#define GRPC_TIMER_STOP(p) \
  ((p) ? gettimeofday(&((p)->gtv2), (struct timezone*)0) : 0)

/************************************************************************/
/************************ INITIALIZE, FINALIZE  *************************/
/************************************************************************/

grpc_error_t 
  grpc_initialize(char *),
  grpc_finalize();

/************************************************************************/
/*************************   HANDLE FUNCTIONS  **************************/
/************************************************************************/

grpc_error_t 
  grpc_function_handle_default(grpc_function_handle_t *, char *),
  grpc_function_handle_init(grpc_function_handle_t *, char *, char *),
  grpc_function_handle_destruct(grpc_function_handle_t *),
  grpc_get_handle(grpc_function_handle_t **, grpc_sessionid_t);

/************************************************************************/
/***********************   INVOCATION FUNCTIONS  ************************/
/************************************************************************/

grpc_error_t
  grpc_call(grpc_function_handle_t *, ...),
  grpc_call_async(grpc_function_handle_t *, grpc_sessionid_t *, ...),
  grpc_call_arg_stack(grpc_function_handle_t *, grpc_arg_stack *),
  grpc_call_arg_stack_async(grpc_function_handle_t *, grpc_sessionid_t *, 
    grpc_arg_stack *),
  grpc_call_arg_stack(grpc_function_handle_t *, grpc_arg_stack *),
  grpc_call_arg_stack_async(grpc_function_handle_t *, grpc_sessionid_t *,
    grpc_arg_stack *),
  grpc_call_arg_list_async(grpc_function_handle_t *handle,
    grpc_sessionid_t *sessionId, gs_va_list *list);


/************************************************************************/
/**************************   SESSION CONTROL  **************************/
/************************************************************************/

grpc_error_t 
  grpc_probe(grpc_sessionid_t sessionID),
  grpc_probe_or(grpc_sessionid_t *idArray, size_t length, 
    grpc_sessionid_t *idPtr),
  grpc_cancel(grpc_sessionid_t sessionID),
  grpc_cancel_all(void);


/************************************************************************/
/**************************   WAIT FUNCTIONS  ***************************/
/************************************************************************/

grpc_error_t
  grpc_wait(grpc_sessionid_t sessionID),
  grpc_wait_and(grpc_sessionid_t *idArray, size_t length),
  grpc_wait_or(grpc_sessionid_t *idArray, size_t length, 
    grpc_sessionid_t *idPtr),
  grpc_wait_all(void),
  grpc_wait_any(grpc_sessionid_t *idPtr);

/************************************************************************/
/***************************   ERROR_HANDLE *****************************/
/************************************************************************/

void grpc_perror(char *str);
char *grpc_error_string(grpc_error_t error_code);
char *grpc_minor_error_string(grpc_error_t error_code);
grpc_error_t grpc_get_error(grpc_sessionid_t sessionID);
grpc_error_t grpc_get_failed_sessionid(grpc_sessionid_t *);

/************************************************************************/
/***********   For the argument stack calling sequence. *****************/
/************************************************************************/

grpc_arg_stack *grpc_arg_stack_new(int);
int grpc_arg_stack_push_arg(grpc_arg_stack *, void *);
void *grpc_arg_stack_pop_arg(grpc_arg_stack *);
int grpc_arg_stack_destruct(grpc_arg_stack *);


/************************************************************************/
/********************   Extensions for GridSolve. ***********************/
/************************************************************************/

grpc_error_t 
  grpc_call_ft(grpc_function_handle_t *, ...),
  grpc_call_async_ft(grpc_function_handle_t *, grpc_sessionid_t *, ...),
  grpc_call_valist_ft(grpc_function_handle_t *, gs_va_list *),
  grpc_call_valist_async_ft(grpc_function_handle_t *, grpc_sessionid_t *,
    gs_va_list *),
  grpc_call_arg_stack_ft(grpc_function_handle_t *, grpc_arg_stack *),
  grpc_call_arg_stack_async_ft(grpc_function_handle_t *, 
    grpc_sessionid_t *,grpc_arg_stack *),
  grpc_serialize_request(grpc_sessionid_t, char **),
  grpc_deserialize_request(char *str, grpc_request_t *req),
  grpc_wait_ft(grpc_sessionid_t),
  gs_wait_common(grpc_request_t *),
  grpc_profile(grpc_profile_t *),
  grpc_set_client_major(char *),
  grpc_set_default_major(char *),
  grpc_set_client_language(int),
  grpc_probe_ft(grpc_sessionid_t),
  grpc_retrieve(char *, ...),
  grpc_function_handle_default_ns(grpc_function_handle_t *, char *),
  grpc_set_criteria(grpc_function_handle_t *, char *);

grpc_request_t 
  *grpc_get_request(grpc_sessionid_t);

int 
  grpc_request_destruct_free_clear(grpc_sessionid_t);

/* Functions returning information about servers and problems */
grpc_error_t grpc_get_servers(char **info);
grpc_error_t grpc_get_problems(char **info);

/************************************************************************/
/*****************************   FARMING  *******************************/
/************************************************************************/

/**
 ** FARMING MACRO DEFINITIONS
 **/

#define MIN_WINDOW_SIZE 5

#define INT_ITERATOR 0
#define INT_ARRAY_ITERATOR 1
#define PTR_ARRAY_ITERATOR 2

#define IGNORE  0
#define INTEGER 1
#define POINTER 2

/**
 **  FARMING FUNCTION PROTOTYPES
 **/

grpc_error_t
 *grpc_farming(int, int, char *, va_list),
  grpc_setup_farming_args(gs_problem_t *, grpc_iterator_t **, int),
  grpc_generate_array_from_iterators(gs_problem_t *, 
    grpc_void_star_or_int_t **, grpc_iterator_t **, int),
  grpc_construct_iterator_array(va_list, int, grpc_iterator_t ***, int),
  grpc_eval_integer(char *, int),
  grpc_send_farming_request(grpc_function_handle_t *, grpc_sessionid_t *,
     grpc_iterator_t **, int),
  *grpc_farm(char *, char *, ...);

void
  grpc_free_iterator(grpc_iterator_t *),
  grpc_free_specific_iterator(grpc_specific_iterator_t *);

grpc_iterator_t
 *grpc_int(char *),
 *grpc_int_array(int *, char *),
 *grpc_ptr_array(void **, char *);

grpc_void_star_or_int_t
  grpc_get_next(grpc_iterator_t *, int arg_type),
  grpc_get_value(grpc_specific_iterator_t *, int, int arg_type);

/**
 ** DSI
 **/

grpc_error_t
  grpc_dsi_open(DSI_FILE **, char*, int, int, int, dsi_type),
  grpc_dsi_read_vector(DSI_OBJECT *, void *, int, enum datatype, int *),
  grpc_dsi_read_matrix(DSI_OBJECT *, void *, int, int, enum datatype, int *),
  grpc_dsi_write_vector(DSI_OBJECT **, DSI_FILE *, void *, int, enum datatype),
  grpc_dsi_write_matrix(DSI_OBJECT **, DSI_FILE *, void *, int, int, enum datatype);

int
  grpc_dsi_close(DSI_FILE *);

#ifdef __cplusplus
}
#endif

#endif
