/********************************************************************/
/*                             gs_dag.h                             */
/*     the header file for GridSolve request sequencing DAG API     */
/*                       Yinan Li, 06/17/2008                       */
/********************************************************************/


#ifndef GS_DAG_H_
#define GS_DAG_H_


#include <stdio.h>
#include <stdlib.h>

#include "grpc.h"
#include "problem.h"
#include "server.h"


/********************************************************************/
/* ----- Core GridSolve Request Sequencing DAG Data Structures ---- */
/********************************************************************/
/* enumeration of dependency types */
enum GS_DEP_TYPE {
    INPUT_AFTER_OUTPUT_DEPENDENCY,
    OUTPUT_AFTER_INPUT_DEPENDENCY,
    OUTPUT_AFTER_OUTPUT_DEPENDENCY,
    CONSERVATIVE_SCALAR_DEPENDENCY
};


/* enumeration of scheduling states */
enum GS_NODE_SCHED_STATE {
    INITIAL,     /* initial state */
    SUBMITTED,   /* reserved for gs_call_async */
    RUNNING,
    WAITING,     /* due to dependency */
    COMPLETED
};


/* types of DAG scheduler */
enum GS_SEQ_SCHEDULER_TYPE {
	ROUND_ROBIN,
	AVERAGE,
	OPT_COMM,
	OPT_COMP,
	AGENT
};


typedef struct GS_DAG_Node GS_DAG_Node_t;
typedef struct GS_DAG_Dep GS_DAG_Dep_t;
typedef struct GS_DAG GS_DAG_t;


/* DAG node structure */
struct GS_DAG_Node {
    char *func_name;                  /* the function name of the node */
    grpc_function_handle_t *handle;   /* grpc function handle */
    gs_va_list *arg_list;             /* argument list */
    grpc_arg_stack *arg_stack;        /* argument stack */
                                      /* either argument list or 
                                         argument stack can be NULL,
                                         but not both */
    gs_server_t *server;              /* ultimately assigned server */
    GS_DAG_Node_t *next;              /* next pointer */
    GS_DAG_Node_t *prev;              /* prev pointer */
    int state;                        /* scheduling state */
    int sched_level;                  /* scheduling level */
	int group_id;                     /* the scheduling group id */
};


/* DAG dependency structure */
struct GS_DAG_Dep {
    GS_DAG_Node_t *pnode;   /* pointer to the parent node */
    GS_DAG_Node_t *cnode;   /* pointer to the child node */
    int dep_type;           /* depedency type */
    gs_argument_t *largp;   /* depended argument */
    gs_argument_t *rargp;   /* depending argument */
    int data_size;          /* the size of the data to be transferred */
                            /* when pushing data from pnode to cnode */
                            /* this will be used when mapping servers */
    GS_DAG_Dep_t *next;     /* next pointer */
    GS_DAG_Dep_t *prev;     /* prev pointer */
};


/* DAG structure */
struct GS_DAG {
    GS_DAG_Node_t *head_node;   /* the head of the list of nodes */
    GS_DAG_Node_t *tail_node;   /* the tail of the list of nodes */
                                /* the list is double-linked */
    GS_DAG_Dep_t *head_dep;     /* the head of the list of dependencies */
    GS_DAG_Dep_t *tail_dep;     /* the tail of the list of dependencies */
                                /* the list is double-linked */
    int num_nodes;              /* number of nodes */
    int num_deps;               /* number of dependencies */
    int max_sched_level;        /* maximum scheduling level */
    int analyzed;               /* flag indicating if the DAG is analyzed */
};


/********************************************************************/
/* ---- General GridSolve Request Sequencing DAG API Functions ---- */
/* ------------------------- gs_seq_dag.c ------------------------- */
/********************************************************************/

/* makes a new DAG structure */
extern GS_DAG_t *make_new_GS_DAG();

/* inserts a node into the DAG */
extern int insert_node_GS_DAG(GS_DAG_t *dag, char *func_name,
    gs_va_list *arg_list, grpc_arg_stack *arg_stack);

/* deletes a node from the DAG */
extern int delete_node_GS_DAG(GS_DAG_t *dag, GS_DAG_Node_t *node);

/* inserts a new dependency between nodes into the DAG */
extern int insert_dep_GS_DAG(GS_DAG_t *dag, GS_DAG_Node_t *pnode,
    GS_DAG_Node_t *cnode, int type,
    gs_argument_t *largp, gs_argument_t *rargp);

/* deletes a dependency from the DAG */
extern int delete_dep_GS_DAG(GS_DAG_t *dag, GS_DAG_Dep_t *dep);

/* find the first node with specified scheduling level */
extern GS_DAG_Node_t *find_node_sched_GS_DAG(
    GS_DAG_t * dag, int sched_level);

/* find the first node with specified parent node */
extern GS_DAG_Node_t *find_node_pnode_GS_DAG(
    GS_DAG_t *dag, GS_DAG_Node_t *pnode);

/* find the first dependency with specified parent node */
extern GS_DAG_Dep_t *find_dep_pnode_GS_DAG(
    GS_DAG_t *dag, GS_DAG_Node_t *pnode);

/* count the number of nodes with specified scheduling level */
extern int count_node_GS_DAG(GS_DAG_t *dag, int sched_level);

/* frees the memory of a DAG */
extern int free_GS_DAG(GS_DAG_t *dag);


/********************************************************************/
/* ----- GridSolve Request Sequencing DAG Scheduling Functions ---- */
/* ---------------------- gs_seq_scheduler.c ---------------------- */
/********************************************************************/

/* analyzes the dependencies among nodes in the DAG */
extern int analyze_dep_GS_DAG(GS_DAG_t *dag);

/* maps a DAG to a list of GridSolve servers */
extern int map_to_servers_GS_DAG(GS_DAG_t *dag, int scheduler);

/* Do some post analysis stuffs to optimize
   the DAG after server mapping */
extern int post_analysis_GS_DAG(GS_DAG_t *dag, va_list arg_list, int n);

/* schedules DAG nodes with specified scheduling level */
extern int execute_GS_DAG(GS_DAG_t *dag);


/********************************************************************/
/* ----------------------- Utility Functions ---------------------- */
/* ------------------------ gs_seq_utils.c ------------------------ */
/********************************************************************/

/* verify that the object has valid-for-comparing 
   type under certain sequencing mode */
extern int verify_object_type(gs_argument_t *argp);

/* compare two objects to make sure that
   they reference to the same data */
extern int compare_object(gs_argument_t *largp, gs_argument_t *rargp);

/* outputs a DAG to corresponding dot graph description */
extern int output_GS_DAG_to_dot(GS_DAG_t *dag, char *fname);


#endif
