
/**
 * @file
 *
 * This file contains structures and functions for the
 * application performance model.  
 * Before this structure is sent to the agent, it only represents
 * the information gathered from the stored handles of the group of tasks.
 * After it is sent to agent it also represent any of the information that 
 * is available to the agent (i.e. the information in the tasks idl file).
 * This includes whether the arguments are inputs/outputs, are vectors
 * ,matrixes,..,.  and the functions required to calculate the complexity of
 * the tasks.  This structure is used to generate the task graph.
 *
 */


#include "config.h"
#ifdef GS_SMART_GRIDSOLVE
#ifndef GS_SMART_APP_PM
#define GS_SMART_APP_PM
#include "problem.h"
typedef struct gs_smart_app_pm_t		 gs_smart_app_pm;
typedef struct gs_smart_app_pm_task_t		 gs_smart_app_pm_task;
typedef struct gs_smart_app_pm_args_t		gs_smart_app_pm_args;
typedef struct gs_smart_t_dep_anayl_t           gs_smart_t_dep_anayl;
typedef struct gs_smart_a_dep_anayl_t           gs_smart_a_dep_anayl;


typedef struct gs_smart_app_pm_wait_t		gs_smart_app_pm_wait;


/*
 * This structure represents the application performance model
 */

struct gs_smart_app_pm_t{
  int nb_tasks;                               /*nb of tasks in the group*/
  gs_smart_app_pm_task ** tasks;
};

/*
 * This structure represents each task in the group.
 * The information gathered from the list of handles is stored here,
 * such as the name of the tasks, whether they are blocking and the pointers
 * to the argument.
 * In addition the idl information on the agent 
 * is stored which is used when generating the task graph
 * to determine the complexity of tasks  
 * and size of arguments in the task graph.
 * The idl information and the pointers to the arguments are analyzed
 * and  the results are stored in the gs_smart_t_dep_anayl. 
 * This information is used to create dependencies between tasks in 
 * the task graph.
 */

struct gs_smart_app_pm_task_t{
   int local;
   int task_id;
   int blocking;
   char * nickname;
   int nb_args;

   /*
    * structure contains idl information for this task
    */
    gs_problem_t * problem;


   /*
    * results of analysis of dependencies and idl information of arguments
    */

  gs_smart_t_dep_anayl * t_anayl;


 /*
  * list of arguments of the task
  */ 

  gs_smart_app_pm_args ** args;
};



/*
 * This structure represents the arguments of
 * the task. It contains the pointer value
 * of each argument, used to determine dependencies.
 * When the task graph is being generated the results of 
 * dependency analysis is stored in a_anayl.
 */

struct gs_smart_app_pm_args_t{
   int id;
   int scalar;

  int type;
  union { 
    int i; 
    double d; 
    char c;
  } uvalue;


  /*
   * value determined from analysis of dependencies of arguments
   */  
  gs_smart_a_dep_anayl * a_anayl;
  


};


/*
 * This strucuture is used to store information on 
 * the task dependencies. It stores the results of
 * the analysis of dependencies, which outlines 
 * the link structure of the task graph. Specifically,
 * those links whose edges start from task and client nodes
 * of the task graph.
 */


struct gs_smart_t_dep_anayl_t{
  int client_node_required;
  int nb_task_for_links;
  int nb_prev_client_for_links;
  int nb_task_back_links;
  int nb_next_client_back_links;
};


/*
 * This strucuture is used to store information on 
 * the task dependencies. It stores the results of
 * the analysis of dependencies, which outlines 
 * the link structure of the task graph. Specifically,
 * those links whose edges start from object nodes of 
 * the task graph.
 */
struct gs_smart_a_dep_anayl_t{
  int is_sent_remotely;
  int is_received_remotely;

  int nb_obj_for_links;
  int nb_clnt_obj_for_links;

  int nb_prev_client_obj_for_links;

  int out_task_dep;
  int out_task_dep_link;
  int out_obj_dep_link;

  int in_task_dep;
  int in_task_dep_link;
  int in_obj_dep_link;
};

/*
 * TODO: Introduce the wait node to application 
 * performance model. This will represent wait functions
 * in the GridRPC enable applications.
 */


struct gs_smart_app_pm_wait_t{
  int task_id;
};


/*
 * Frees application performance model
 */

int gs_smart_app_pm_free(gs_smart_app_pm *);

int gs_smart_free_handle_list();

#endif 
#endif
