/*
 * @file
 *
 * SmartGridSolve client library
 */

#include "config.h"
#ifndef GS_SMART_CLIB
#define GS_SMART_CLIB


/*
 * The gs_smart_map "function" is a macro that inserts a while loop around
 * the code block specified by the parenthesis. When the gs_smart_map 
 * function is called the code within the parenthesis will be iterated through
 * twice and the function gs_smart_map_x will be called twice. On the first iteration each grpc_call and grpc_call_asynch is discovered
 * but not executed.  From these discoverd calls a task graph is generated.
 * At the beginning of the second iteration the mapping heuristic specified by
 * the gs_smart_map parameter will generate a mapping solution based on the 
 * task graph and the performance model of the network.  The mapping solution
 * outlines a task to server mapping and also the communication scheme between
 * tasks.
 */

#define grpc_map(heuristic) while(gs_smart_map_X(heuristic))

/*
 * The gs_smart_map_ft function is the same as the gs_smart_map function, except that the mapping 
 * soluction generated does not implement server-server communcation. This means that if
 * any task fails execution the system can recover by sending the same task to the next best
 * available server.
 */
#define grpc_map_ft(heuristic) while(gs_smart_map_ft_X(heuristic))


/*
 * These are the functions the macro wraps the while loop around.
 */

int gs_smart_map_X(char * mapper_name);
int gs_smart_map_ft_X(char * mapper_name);




/*
 * Function maps a group of tasks by iterating through
 * the group of tasks twice.  First cycle the 
 * task graph is generated. At the beginning of the second
 * cycle the group of tasks is mapped. On the second cycle
 * the group of tasks is mapping according to the mapping 
 * solution
 */
int gs_smart_map_two_cycles(char * mapper_name, char * comm_type);

/*
 * This function is used when there is section of code that should 
 * be ignored during the first cycle when the the task graph is in 
 * the process of being built.
 *
 * USAGE:
 *
 * grpc_local_region(){
 *  ..
 *  //section of code to ignore during task graph building
 *  ..
 * }
 */

#define grpc_local() if(grpc_local_X())

int grpc_local_X();

#endif 
