
/**
 * @file
 *
 * This file contains structures and functions for
 * the mapping graph.
 * The mapping graph is generated by the mapping heuristic. It outlines
 * the task to server mapping of each task in the group of tasks. 
 * In addition it outlines the communication between tasks.
 * This mapping graph is embedded into the task at handles when each
 * task is subsequently called.
 *
 */



#include "config.h"
#ifdef GS_SMART_GRIDSOLVE
#ifndef GS_SMART_MAPPING_INTERFACE
#define GS_SMART_MAPPING_INTERFACE
#include "grpc.h"
#include "gs_smart_app_pm.h"
#include "server.h"
#include "problem.h"
#include "comm_protocol.h"
#include "gs_smart_task_graph.h"
#include "comm_basics.h"
#include "comm_data.h"
#include "gs_smart_netpm.h"
#define GS_MAX_PINGS 5
#define GS_SMART_ARG_FILE_PREFIX "gs_smart_arg"

typedef struct gs_smart_mg_t       gs_smart_mg;
typedef struct gs_smart_mg_node_t       gs_smart_mg_node;
typedef struct gs_smart_mg_task_node_t       gs_smart_mg_task_node;
typedef struct gs_smart_mg_client_node_t       gs_smart_mg_client_node;
typedef struct gs_smart_mg_obj_node_t       gs_smart_mg_obj_node;
typedef struct gs_smart_mg_link_t       gs_smart_mg_link;
typedef struct gs_smart_mg_node_type_t          gs_smart_mg_node_type;
typedef struct gs_smart_mg_in_arg_map_info_t          gs_smart_mg_in_arg_map_info;
typedef struct gs_smart_mg_out_arg_map_info_t          gs_smart_mg_out_arg_map_info;



/*
 * This structure represents the mapping graph
 */


struct gs_smart_mg_t{
  int nb_mg_nodes;
  int nb_mg_rem_nodes;
  gs_smart_mg_node ** mg_nodes;
  double total_time;
};

/*
 * This structure represents each node in the 
 * mapping graph. A mapping node can either
 * be a client node or a task node. 
 */

struct gs_smart_mg_node_t{
  int node_type;
  int id;
  gs_smart_mg_task_node * mg_task_node;
  gs_smart_mg_client_node * mg_client_node;
};

/*
 * This structure represents a client mapping
 * node in the mapping graph. This structure
 * is used to outline communication 
 * between task nodes and the client.
 */


struct gs_smart_mg_client_node_t{
  gs_smart_tg_client_node * mapped_client_node;
  int nb_for_links;
  gs_smart_mg_link ** mg_node_for_links;
};



/*
 * This structure represents a task mapping node
 * Each task mapping node is assigned a task node
 * from the task graph and a server node from 
 * the netpm.  Each mapping task node is timed
 * based on the flops/sec performance of the server
 * and the nb flops computation of the task. 
 */
struct gs_smart_mg_task_node_t{
  double time;
  gs_smart_tg_remote_task_node * mapped_task_node;
  gs_smart_netpm_node * mapped_server_node;
  
  int * back_link_set;
  int nb_for_links; 
  gs_smart_mg_link ** mg_node_for_links;
  int nb_back_links; 
  gs_smart_mg_link ** mg_node_back_links;
};


/*
 * This structure represents an argument which is
 * sent remotely.  The back links are links
 * which specify whether the source of the argument 
 * is another task-server mapping node or a client
 * mapping node. The forward links are links which
 * specify the destinations of the argument, whether
 * they are other task-server mapping nodes or a client
 * mapping node.  
 * 
 * The smart_arg_file attribute is the
 * file location where the argument
 * will be stored remotely.
 * The obj_node_set attribute is used 
 * when generating mapping graph
 * to determine whether obj_node_set
 * has already been created
  */


struct gs_smart_mg_obj_node_t{
  int sent_remotely;
  gs_smart_tg_obj_node * mapped_obj_node;
  int nb_for_links; 
  int nb_dest_servers;
  gs_smart_mg_link ** mg_obj_for_links;
  gs_smart_mg_link * mg_obj_back_link;
  int obj_node_set;    
  char * smart_arg_file;
};



/*
 * This structure is used to specifies a two way
 * dependency from an mapping object node to a task-server
 * mapping node or from a task-server/client mapping node 
 * to an object node.
 * The for_link_pos indicates its link position
 * from source node and the back_link_pos 
 * indicate its link position on the destination node.
 * 
 * The time attribue is calculated based on the number of 
 * bytes of its connecting object mapping node and the 
 * bandwidth of the assigned network pm link 
 * (mapped_netpm_link). 
 *
 * The caching attribute specifies whether the arg is 
 * cached. The store_arg attribute indicates the arg
 * should be stored after it is used for future tasks.
 * The delete arg attribute indicates whether this arg
 * should be deleted after it is used.
 */


struct gs_smart_mg_link_t{
  double time;
  int caching;
  int for_link_pos;
  int back_link_pos;

  int already_on_server;
  int store_arg;
  int delete_arg;

  gs_smart_mg_node * mg_node_ptr;
  gs_smart_mg_obj_node * mg_obj_node_ptr;
  gs_smart_netpm_link * mapped_netpm_link;
};




typedef enum {
  GS_SMART_MG_START_NODE = 0,
  GS_SMART_MG_CLIENT_NODE,
  GS_SMART_MG_TASK_NODE,
  GS_SMART_MG_END_NODE
} gs_smart_mg_node_type_t;



struct gs_smart_mg_in_arg_map_info{
  int input_arg_stored;             
  int input_arg_sent_remotely;      
  int input_nb_dest_servers;        
  char * input_arg_file;
  gs_server_t ** input_dest_servers;
  char ** enc_input_dest_servers;   
};





typedef enum {
  GS_SMART_STANDARD_EXEC = 0,
  GS_SMART_TASK_DISCOVERY,
  GS_SMART_MAPPING,
  GS_SMART_EXEC_CALLED_HANDLES,
  GS_SMART_EXEC_STORED_HANDLES,
  GS_SMART_EXEC_TASK_FAIL,
  GS_SMART_MAP_FAIL
} gs_smart_phase_enum_t;



/*
 *   If set to exlicit, the number of tasks in the group
 *   has been explicitly specified by the user in a paramater
 *   of the gs_smart_map function.
 *  
 *   If set to implicit the number of task has been implicitly
 *   specified either by paranthesis or gs_smart_end_map 
 *   to indicate the scope of the group of tasks
 */

typedef enum {
  GS_SMART_EXPLICIT_GROUP = 0,
  GS_SMART_IMPLICIT_GROUP
} gs_smart_group_type_t;



typedef enum {
  GS_SMART_ONE_PASS_CALL = 0,
  GS_SMART_TWO_PASS_CALL
} gs_smart_call_type_t;







/*
 * Function which invokes the mapping heuristic functions 
 */

int gs_smart_map_common(char * mapper_type,   gs_smart_tg * tg, gs_smart_netpm * netpm, int *);


/***********************************************************/
/*******  FUNCTIONS FOR CONSTRUCTING MAPPING GRAPH *********/
/***********************************************************/


int 
  gs_smart_generate_mapping_graph(gs_smart_tg * , gs_smart_netpm * ,int * ,
                                                       gs_smart_mg *, int *),
  gs_smart_mg_construct_frame(gs_smart_tg * , gs_smart_netpm * , int * , 
                                                       gs_smart_mg *, int *),
  gs_smart_mg_count_c_comm_for_obj_links(gs_smart_tg * , gs_smart_netpm * ,
                                    gs_smart_mg *, int ,int , int * , int * ),
  gs_smart_mg_construct_c_comm_obj_back_links(gs_smart_tg * , gs_smart_netpm *,
                                               gs_smart_mg *, int , int , int),
  gs_smart_mg_set_c_comm_dep(gs_smart_tg * , gs_smart_netpm * , gs_smart_mg *, 
                                                   int, int ,int , int , int *),
  gs_smart_mg_count_s_comm_for_obj_links(gs_smart_tg * , gs_smart_netpm * , 
                                    gs_smart_mg *, int ,  int , int * , int *),
  gs_smart_mg_construct_s_comm_obj_for_links(gs_smart_tg * , gs_smart_netpm *,
                         gs_smart_mg *, int , int , int ),
  gs_smart_mg_set_s_comm(gs_smart_tg *, gs_smart_netpm *, gs_smart_mg *, 
                                              int , int , int , int , int *),
  gs_smart_check_if_valid_assignment(gs_smart_tg_remote_task_node * ,
				 gs_server_t * , int *),
  gs_smart_mg_create_argument_file_names(gs_smart_mg *),
  gs_smart_time_mg_task_node(gs_smart_mg_task_node * , double * ),
  gs_smart_time_tg_task_node(gs_smart_tg_remote_task_node *, gs_server_t * ,
                                                                 double * ),
  gs_smart_get_nb_auto_tasks_left_to_build(),
  gs_smart_free_mg(gs_smart_mg * ),
  gs_smart_mg_print_to_dotgraph(gs_smart_mg * , char * );


#endif 
#endif
