
/**
 * @file
 *
 * This file contains structures and functions for
 * network performance model.  
 * The network performance model is a representation of the 
 * performance of the servers and communication links of the fully or
 * partially connected network.
 * If SmartGridSolve is installed only on the client side, this structure will
 * represent a network where no servers can communicate (in this case the 
 * application programmer will only benefit from improved mapping).
 * If some of the servers are SmartGridSolve enabled then this structure
 * will represent a partially connected network (the AP will also benefit from 
 * remote communication between servers). If all servers are SmartGridSolve
 * enabled the network will be fully connected. 
 * This structure specifies the preformance (flops/sec) of each server and also 
 * each link (bandwidth) connecting it will the client. Where two servers in 
 * the network are SmartGridSolve enabled the performance (bandwidth) of the 
 * link between these servers will also be specified.
 *
 */



#include "config.h"
#ifdef GS_SMART_GRIDSOLVE
#ifndef GS_SMART_NETPM
#define GS_SMART_NETPM
#include "server.h"




typedef struct gs_smart_netpm_t			        gs_smart_netpm;
typedef struct gs_smart_netpm_node_t			gs_smart_netpm_node;
typedef struct gs_smart_netpm_link_t			gs_smart_netpm_link;
typedef struct gs_smart_netpm_node_type_t		gs_smart_netpm_node_type;


/*
 * This structure represents the network performance model. 
 */

struct gs_smart_netpm_t{
   int nb_nodes;
  gs_smart_netpm_node * netpm_clnt_node;  /* client node */
  gs_smart_netpm_node ** netpm_nodes;   /* Array of all task nodes */
  double avg_c_bw;
  double avg_s_bw;
};


/*
 * This structure represents the network node which can
 * either be a client node or a server node. 
 */

struct gs_smart_netpm_node_t{
  int node_type;   
  int id;
  int nb_links;
  gs_server_t * server;
  gs_smart_netpm_link ** netpm_links;
};



/*
 * This structure represents a two way link between
 * two network nodes, either between the client and server or
 * between two server nodes. This structure also specifies
 * the bandwidth of the link.
 */

struct gs_smart_netpm_link_t{
  double bw;
  gs_smart_netpm_node * linked_nodeA;
  gs_smart_netpm_node * linked_nodeB;
};

/*
 * This tag identifies whether the task node is either a 
 * server node or a client node.
 */

typedef enum {
  GS_SMART_NETPM_SERVER_NODE = 0,
  GS_SMART_NETPM_CLIENT_NODE
} gs_smart_netpm_node_type_t;


/************************************************************************/
/******  FUNCTIONS FOR CONSTRUCTING THE NETWORK PERFORMANCE MODEL *******/
/************************************************************************/
int
  gs_smart_netpm_construct_netpm(gs_smart_netpm * , gs_server_t ** , int ),
  gs_smart_netpm_construct_nodes(gs_smart_netpm *, gs_server_t ** , int ),
  gs_smart_netpm_construct_links(gs_smart_netpm * , gs_server_t **, int ),
  gs_smart_netpm_add_bw_to_links(gs_smart_netpm * , gs_server_t ** , int ),
  gs_smart_netpm_measure_client_links_perf(gs_server_t ** , int ),
  gs_smart_netpm_aggregate_ping_info(gs_server_t ** , int ),
  gs_smart_net_pm_print_to_dotgraph(gs_smart_netpm *, char * ),
  gs_smart_netpm_free(gs_smart_netpm * );


  
#endif 
#endif
