/**
 * @file
 *
 * This file contains structures and function prototypes for
 * application task graph. The task graph is the full representation
 * of a group of tasks.  This representation is based on the 
 * discovery of the group of tasks and the idl of information of
 * each task in the group. 
 * The task graph highlights the order of tasks, their synchronisation 
 * (whether they are executed in sequence or parallel), the dependencies 
 * between tasks, the load of computation and communication of 
 * each task in the group.
 *
 */

#include "config.h"
#ifdef GS_SMART_GRIDSOLVE
#ifndef GS_SMART_TASK_GRAPH
#define GS_SMART_TASK_GRAPH
#include "gs_smart_app_pm.h"
#include "gs_smart_netpm.h"

typedef struct gs_smart_tg_t			        gs_smart_tg;
typedef struct gs_smart_tg_task_node_t			gs_smart_tg_task_node;
typedef struct gs_smart_tg_client_node_t		gs_smart_tg_client_node;
typedef struct gs_smart_tg_remote_task_node_t		gs_smart_tg_remote_task_node;
typedef struct gs_smart_tg_obj_node_t			gs_smart_tg_obj_node;
typedef struct gs_smart_tg_dep_link_t			gs_smart_tg_dep_link;
typedef struct gs_smart_tg_node_type_t			gs_smart_tg_node_type;
typedef struct gs_smart_tg_graph_type_t			gs_smart_tg_graph_type;
typedef struct gs_smart_graph_indexes_t 	gs_smart_graph_indexes;

/*
 * This structure represents the task graph
 */
struct gs_smart_tg_t{
   int nb_nodes;                         /* total nb of nodes including 
                                                client and task nodes */
  gs_smart_tg_task_node ** task_nodes;   /* Array of all task nodes */
  int nb_rem_nodes;                      /* nb of tasks in the group */
};

/*
 * This structure represents the task nodes in the task graph.
 */

struct gs_smart_tg_task_node_t{
  int node_type;                          /* specifies whether its a 
                                          start,client, task or end node */
  int id;                                 /* specifies its position in task graph */
  gs_smart_tg_remote_task_node * tg_rem_task_node;  
  gs_smart_tg_client_node * tg_client_node;  
};

/*
 * The client node structure is used to represent any communication from
 * the client to remote task nodes. 
 */

struct gs_smart_tg_client_node_t{
  int nb_node_frwd_links;
  int nb_node_bkwd_links;
  gs_smart_tg_dep_link ** tg_node_frwrd_link;
  gs_smart_tg_dep_link ** tg_node_bkwrd_link;
};

/*
 * The remote task node structure represents tasks which are executed 
 * remotely. It specifies the computation of the task (nb_flops) and
 * the dependencies between tasks and client node are specified
 * with the forward and backward links..
 */

struct gs_smart_tg_remote_task_node_t{
  int rem_id;
  gs_problem_t * problem;
  int blocking;
  int nb_servers;
  gs_server_t **avail_servers;  //servers which can perform task
  double nb_flops;
  int nb_node_frwd_links;
  int nb_node_bkwd_links;
  gs_smart_tg_dep_link ** tg_node_bkwrd_link;
  gs_smart_tg_dep_link ** tg_node_frwrd_link;
};



/*
 * The object node structure represents arguments which are sent 
 * remotely. It specifies the size of the arguments of the task 
 * (byte_size). The backward link specifies the source of the argument 
 * which can either be a client node or another remote task node.
 * The forward links specify the destinations of the argument, 
 * whether its the client or a server or list of servers 
 * (where multiple dependencies exist).
 */


struct gs_smart_tg_obj_node_t{

  double time;
  int data_ptr;
  int in_call_seq_pos; 
  int out_call_seq_pos; 
  int rem_dep_id;

  int byte_size;    

  char * name;
 
  gs_argument_t *argptr;


  int nb_obj_frwd_links;
  gs_smart_tg_dep_link * tg_obj_bkwrd_link;
  gs_smart_tg_dep_link ** tg_obj_frwrd_link;
};



/*
 * This structure is used to specifies a two way
 * dependency from an object node to a task or
 * from a task/client node to an object node.
 * The for_link_pos indicates its link position
 * from source node and the back_link_pos 
 * indicate its link position on the destination node.
 *
 */


struct gs_smart_tg_dep_link_t{
  int for_link_pos;
  int back_link_pos;
  gs_smart_tg_task_node * task_node_ptr;
  gs_smart_tg_obj_node * obj_node_ptr;
};


/*
 * This structure is used when generating the task graph
 * to keep track of the current position on both the 
 * application performance model graph and the task graph.
 * When analyzing dependencies,  the position on the source
 * node and destination node of both the task graph and application
 * performance model graph need to be stored.
 */

struct gs_smart_graph_indexes_t{
  int cur_arg_id;
  int rem_dep_id;

  int tg_ti;
  int tg_sti; /* task node index */
  int tg_dti;

  int tg_fli; /* task graph forward_link index */
  int tg_clnt_fli; /* task graph forward_link index */
  int tg_bli; /* task graph backward_link index */
  int tg_lnai; /* task graph local node arg index */
  int tg_rnai;
  int tg_drnai; /* destination task node arg index */

  int ap_ti;
  int ap_ai;
  int ap_sti;  /* src handle index */
  int ap_sai; /* src handle arg index */
  int ap_dti;  /* dest handle index */
  int ap_dai; /* dest handle arg index */
  int ap_ddti;  /* dest handle index */
  int ap_ddai;  /* dest handle index */

};

/*
 * This tag identifies the node types of
 * each task node in the task graph
 */

typedef enum {
  GS_SMART_TG_START_NODE = 0,
  GS_SMART_TG_CLIENT_NODE,
  GS_SMART_TG_REM_TASK_NODE,
  GS_SMART_TG_END_NODE
} gs_smart_tg_node_type_t;


/*
 * This tag identifies which dependencies
 * to analyze when generating the task graph.
 */

typedef enum {
  GS_SMART_NO_DEP_GRAPH = 0,
  GS_SMART_P2P_GRAPH ,
  GS_SMART_SRV_BRDCST_GRAPH,
  GS_SMART_CLNT_BRDCAST_COMM,
  GS_SMART_CL_SRV_BRDCST_GRAPH
}gs_smart_tg_graph_type_t;




/************************************************************************/
/**************  FUNCTIONS FOR CONSTRUCTING TASK GRAPH ******************/
/************************************************************************/


int
  gs_smart_generate_netpm_and_task_graph(gs_smart_app_pm * , gs_smart_tg * ,
                                                     gs_smart_netpm * , char *),
  gs_smart_tg_construct_tg(gs_smart_app_pm * ,  gs_smart_tg * , int),
  gs_smart_tg_analyse_dependencies(gs_smart_app_pm * , gs_smart_tg * ,  int),
  gs_smart_tg_analyse_no_dependencies(gs_smart_app_pm *,  gs_smart_tg *),
  gs_smart_tg_analyse_p2p_dependencies(gs_smart_app_pm *,  gs_smart_tg *),
  gs_smart_tg_analyse_srvr_brdcst_dependencies(gs_smart_app_pm *,  gs_smart_tg *),
  gs_smart_tg_analyse_clnt_brdcst_dependencies(gs_smart_app_pm * , gs_smart_tg * ),
  gs_smart_tg_analyse_clnt_srvr_brdcst_dependencies(gs_smart_app_pm *,  gs_smart_tg *),
  gs_smart_tg_construct_frame(gs_smart_app_pm *,  gs_smart_tg *),
  gs_smart_tg_construct_start_node(gs_smart_tg *, gs_smart_graph_indexes *),
  gs_smart_tg_construct_client_node(gs_smart_app_pm *,
      gs_smart_tg *, gs_smart_graph_indexes *),
  gs_smart_tg_construct_remote_task_node(gs_smart_app_pm * ,
      gs_smart_tg *, gs_smart_graph_indexes *),  
  gs_smart_tg_construct_forward_links(gs_smart_app_pm * ,
    gs_smart_tg * , gs_smart_graph_indexes * ),
  gs_smart_tg_construct_backward_links(gs_smart_app_pm * ,
    gs_smart_tg * , gs_smart_graph_indexes * ),
  gs_smart_tg_construct_end_node(gs_smart_tg * ,
      gs_smart_graph_indexes * ),
  gs_smart_tg_dependency_search(gs_smart_app_pm *, gs_argument_t * , gs_argument_t **,  
       gs_smart_graph_indexes * , int * ),
  gs_smart_tg_set_scalar_arg_values(gs_smart_app_pm * , icl_hash_t *, gs_smart_graph_indexes *),
  gs_smart_tg_get_problem_size(gs_problem_t * , double * ),
  gs_smart_tg_rem_for_links_to_dotgraph(gs_smart_tg_remote_task_node *,int , FILE * , int * ),
  gs_smart_tg_client_for_links_to_dotgraph(gs_smart_tg_client_node * , int , FILE *),
  gs_smart_tg_dep_anayl_set_remote_dep(gs_smart_t_dep_anayl * , 
      gs_smart_t_dep_anayl * , gs_smart_a_dep_anayl * ,   gs_smart_a_dep_anayl * ,
      gs_smart_graph_indexes * , int ),
  gs_smart_tg_alloc_backward_links(gs_smart_app_pm *, gs_smart_tg * , 
      gs_smart_graph_indexes * gi),
  gs_smart_tg_join_for_back_links(gs_smart_app_pm * , gs_smart_tg *, 
      gs_smart_graph_indexes * ),
  gs_smart_tg_rem_task_join_for_back_links(gs_smart_app_pm *, gs_smart_tg *, 
      gs_smart_graph_indexes *),
  gs_smart_tg_client_node_join_for_back_links(gs_smart_app_pm *,gs_smart_tg * , 
      gs_smart_graph_indexes *),
  gs_smart_get_argument_size(gs_argument_t * ),
  gs_smart_init_graph_indexes(gs_smart_graph_indexes **),
  gs_smart_tg_free(gs_smart_tg *),
  gs_smart_tg_print(gs_smart_tg *),
  gs_smart_tg_print_to_dotgraph(gs_smart_tg * , char *);

#endif 
#endif
