/* 
 * @file
 *
 * Functions for accessing agent storage database. These
 * functions are abstracted so that we can place various database
 * backends in place w/o changing the interface.
 */
/* $Id: gs_storage.h,v 1.28 2008/12/15 21:01:52 seymour Exp $ */
/* $UTK_Copyright: $ */

#ifndef GS_STORAGE_H
#define GS_STORAGE_H

#include "server.h"
#include "problem.h"
#include "agent.h"
#include "gs_htm.h"

#define GS_SUCCESS       0
#define GS_BADPROBDESC  -1
#define GS_NOSUCHSERVER -2
#define GS_DBERROR      -3

#define GS_CRIT_PARSE_SQLITE 1
#define GS_CRIT_PARSE_MYSQL  2

int
  gs_update_perf_expr(char *, char *, char *);

int
  gs_mysql_init_db(gs_agent_t *),
  gs_spawn_db_manager(gs_agent_t *),
  gs_get_task(char *, char *, gs_htm_task *),
  gs_get_task_by_agent_taskid(int, gs_htm_task *),
  gs_get_tasks_for_server(char *, gs_htm_task ***, int *, int),
  gs_update_task(char *, char *, char *, int, double, double, double, double, int, int),
  gs_insert_submitted_task(char *, char *, int, double, double, double, double,
    int, int),
  gs_insert_submitted_task_guess(char *, char *, int, double, double, double, double,
    int, int),
  gs_insert_completed_task(char *, char *, int, double, double, double, double,
    int, int);

/** \brief Open the storage db
 *
 * \param gs_agent The agent structure.
 * \return 0 on success, -1 on error.
 */
int gs_storage_init(gs_agent_t * gs_agent);

/** \brief Close the storage db
 *
 * \param gs_agent The agent structure.
 */
void gs_storage_finalize(gs_agent_t * gs_agent);

/** \brief Add a server to the database.
 *
 * This is the first step of registering a server.   The problems
 * must still be registered.
 *
 * \param gs_agent The agent structure.
 * \param gs_server The structure of the server being added.
 * \return 0 on success, -1 on error.
 */
int gs_add_server(gs_agent_t * gs_agent, gs_server_t * gs_server);

/** \brief Removes a server and all problem associations from the database.
 *
 * When a server is removed, the database automatically handles 
 * unregistering the related problems and removing problems that
 * no longer have servers.
 * 
 * \param gs_agent The agent structure.
 * \param gs_server The server to remove.  Only componentid needs
 *                  to be filled in.
 * \return 0 on success, -1 on error.
 */
int gs_delete_server(gs_agent_t * gs_agent, gs_server_t * gs_server);

/** \brief Associated a problem w/ a server, adding the problem if it is 
 *         unknown.
 *  \param gs_agent The agent structure.
 *  \param gs_problem array of complete problem structures.
 *  \param num_services number of elements in gs_problem
 *  \param models array of performance models
 *  \param rm_prob array of problem names to be removed
 *  \param num_removed number of elements in rm_prob
 *  \param cid_string The component id of the server.
 *  \return 0 on success, -1 on error.
 */
int gs_register_problem_changes(gs_agent_t * gs_agent, 
  gs_problem_t ** gs_problem, int num_services, char **models,
  char **rm_prob, int num_removed, char *cid_string);

/** \brief Disassociate a problem w/ a server.
 *
 * When a problem is disassociated from its last server, it is automatically
 * removed by the database.
 *
 * \param gs_agent The agent structure.
 * \param gs_problem The name of the problem to delete.
 * \param cid_string The server component id to disassociate with.
 * \return 0 on success, -1 on error.
 */
int gs_delete_problem(gs_agent_t * gs_agent, char * probname,
                      char *cid_string);

/** \brief Update a server workload.
 *
 * At least gs_server->componentid and gs_server->workload must be
 * filled in.
 *
 * \param gs_agent The agent structure.
 * \param gs_server The server structure.  At least componentid and workload
 *                  must be filled in.
 * \return 0 on success, -1 on error.
 */
int gs_update_workload(gs_agent_t * gs_agent, gs_server_t * gs_server);
int gs_update_ping_list(gs_agent_t * gs_agent, char *cid, char *pings);

/** Checks whether a server is registered.
 *
 * At least gs_server->componentid must be filled in.
 *
 * \param gs_agent The agent structure.
 * \param gs_server The server structure.  At least componentid must be 
 *                  filled in.
 * \return 0 if server is not registered, 1 if it is, -1 on error.
 */
int gs_server_exists(gs_agent_t * gs_agent, gs_server_t * gs_server);

/** \brief Expires all servers that have not updated recently.
 *
 * Removes all servers and their associated problems if they haven't 
 * sent a workload report in more than timeout seconds.
 *
 * \param gs_agent The agent structure.
 * \param timeout The number of seconds for the timeout value.
 * \return 0 on success, -1 on error.
 */
int gs_server_expire(gs_agent_t * gs_agent, int timeout);

/** \brief Returns a list of servers tha can solve a given problem.
 *
 * At least gs_problem->problemname needs to be filled in, the rest
 * will be filled in by the function.
 *
 * \param gs_agent The agent structure.
 * \param gs_problem The problem structure.  Just problemname is needed, the
 *                   rest will be filled in.
 * \param sa_list Server attributes to require.  If sa_list is NULL, it
 *                is ignored.
 * \param servers An array of server structures of servers that can solve
 *                the requested problem. This is allocated in the function.
 * \param count The number of servers in servers list.
 * \return number of servers on success, -1 on error.
 */
int gs_get_server_list(gs_agent_t * gs_agent, gs_problem_t * gs_problem,
                       char *client_criteria, gs_server_t *** gs_servers,
                       int *);

/** \brief Returns a list of all problems the a given server can solve.
 *
 * At least gs_server->componentid must be filled in for this function.
 *
 * \param gs_agent The agent structure.
 * \param gs_server The server structure.  At least componentid is needed.
 * \param problems A list of problem structures known by this server. This 
 *                 is allocated by the function.
 * \param count The number of problems in the list.
 * \return The number on problems or -1 on error.
 */
int gs_get_problem_list(gs_agent_t * gs_agent, gs_server_t * gs_server,
                        gs_problem_t *** gs_problems, int *);

/** \brief Returns a list of all known servers.
 *
 * \param gs_agent The agent structure.
 * \param servers A list of all server structures, allocated by the function.
 * \param count The number of servers in servers list.
 * \return number of servers or -1 on error.
 */
int gs_get_all_servers(gs_agent_t * gs_agent, gs_server_t *** gs_servers,
                       int *count);
int gs_get_all_smart_servers(gs_agent_t * gs_agent, gs_server_t *** gs_servers,
                       int *count);
int gs_get_server_ping_list(gs_agent_t * gs_agent, gs_server_t *** servers, 
                       char *cid, int *count);

/** \brief Returns a list of all problems the a given server can solve.
 *
 * \param gs_agent The agent structure.
 * \param gs_server The server structure.  At least componentid is needed.
 * \param problems A list of problem structures known by this server. This 
 *                 is allocated by the function.
 * \param count The number of problems in the list.
 * \return The number on problems or -1 on error.
 */
int gs_get_all_problems(gs_agent_t * gs_agent, gs_problem_t *** gs_problems,
                        int *count);
int gs_get_problem_info(gs_agent_t * gs_agent, gs_problem_t *** problems, 
                        int *count, char *name);

/** \brief Returns one server structure based on the hostname provided
 * 
 * \param gs_agent The agent structure.
 * \param name The hostname of the desired server.
 * \param gs_server The server structure.
 * \return The 0 or -1 on error.
 */
int gs_get_server(gs_agent_t * gs_agent, char *name, gs_server_t *server);
int gs_get_server_by_cid(gs_agent_t * gs_agent, char *cid, gs_server_t *server);
int gs_get_all_servers_by_hostname(gs_agent_t * gs_agent, char *hostname,
        gs_server_t *** servers, int *count);

#endif
