/**
 * @file
 *
 * Header file for logging routines and macros.
 *
 */
/* $Id: log.h,v 1.20 2007/09/11 15:46:37 yarkhan Exp $ */
/* $UTK_Copyright: $ */

#ifndef _GS_LOG_H
#define _GS_LOG_H

#include <stdio.h>
#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h" 
#include "gridsolve-config.h" 
#endif /* HAVE_CONFIG_H */

#include "portability.h"
#include "utility.h"

/* -------------------------------------------------- */

#define ASSERT_EXPR(assertion, expr) do {if (!(assertion)) { ERRPRINTF("FAILED ASSERTION: %s\n",#assertion); expr; }} while (0);
#define DO_EXPR_IF_FAIL(assertion, expr) do {if (!(assertion)) { ERRPRINTF("FAILED ASSERTION: %s\n",#assertion); expr; }} while (0);

/* -------------------------------------------------- */
/* These functions defined for all loggers */

/* Handle initialization for the logger, e.g., setting up netlogger */
int gs_log_init(const char *ident, int option, int facility, const char *default_event);

/* This is compatible with the NetLogger write format, the
   NetLoggerHandle has to be prepended to the parameters */
int gs_log_write(const char* file, const int line, const char* function, int level, const char* event, const char* format, ...);

/* Handle finalization of the logger, e.g., flushing netlogger */
int gs_log_finalize();

/* -------------------------------------------------- */

#ifdef HAVE_NETLOGGER

#ifdef GS_DEBUG
#define DBGPRINTF(...) do { gs_log_write( __FILE__, __LINE__, __FUNCTION__, LOG_DEBUG, NULL, __VA_ARGS__); } while (0)
#endif /* GS_DEBUG */
#define ERRPRINTF(...) do { gs_log_write( __FILE__, __LINE__, __FUNCTION__, LOG_ERR, NULL, __VA_ARGS__); } while (0)
#define LOGPRINTF(...) do { gs_log_write( __FILE__, __LINE__, __FUNCTION__, LOG_INFO, NULL, __VA_ARGS__); } while (0)

#endif /* HAVE_NETLOGGER */

/* -------------------------------------------------- */

/* If debugging/output macros are not yet defined, define them now */

#ifndef DBGPRINTF
#  ifdef GS_DEBUG 
#    ifdef HAVE_ISO_VARARGS
#       define DBGPRINTF(...) do {fprintf(stderr,"%s:%d [%s] %d %g DBG: ",__FILE__,__LINE__,__FUNCNAME__, (int)getpid(), walltime());fprintf(stderr, __VA_ARGS__); fflush(stderr); } while (0)
#    else  /* does not know ISO style varargs */
#       define DBGPRINTF printf("%s: %d [] %d %g DBG: ", __FILE__ , __LINE__ , -1, walltime() )+printf
#    endif  /* #ifdef HAVE_ISO_VARARGS */
#  else  /* dont want debugging */
#    ifdef HAVE_ISO_VARARGS
#      define DBGPRINTF(...)
#    else  /* does not know ISO style varargs */
static void DBGPRINTF(const char *format, ...) {}
#    endif /* #ifdef HAVE_ISO_VARARGS */
#  endif /* #ifdef GS_DEBUG  */
#endif /* #ifndef DBGPRINTF */


#ifndef ERRPRINTF
#  ifdef HAVE_ISO_VARARGS
#    define ERRPRINTF(...) do { time_t clock; struct tm *now;\
                             clock = time(0); now = localtime(&clock); \
                             fprintf(stderr,"[(%02d/%02d/%04d) %02d:%02d:%02d] [%s:%d] [%s] %d %g ERR: ",\
                             now->tm_mon+1,now->tm_mday,now->tm_year+1900,\
                             now->tm_hour,now->tm_min,now->tm_sec,\
                             __FILE__,__LINE__,__FUNCNAME__, (int)getpid(), \
                             walltime());fprintf(stderr, __VA_ARGS__); \
                             fflush(stderr); } while (0)
#  else
#       define ERRPRINTF printf
#  endif /* ifdef HAVE_ISO_VARARGS */
#endif /* ERRPRINTF */


#ifndef LOGPRINTF
#  ifdef HAVE_ISO_VARARGS
#    ifdef GS_DEBUG
#      define LOGPRINTF(...) do { time_t clock; struct tm *now;\
                             clock = time(0); now = localtime(&clock); \
                             fprintf(stderr,"[(%02d/%02d/%04d) %02d:%02d:%02d] [%s:%d] [%s] %d %g LOG: ",\
                             now->tm_mon+1,now->tm_mday,now->tm_year+1900,\
                             now->tm_hour,now->tm_min,now->tm_sec,\
                             __FILE__,__LINE__,__FUNCNAME__,\
                             (int)getpid(), walltime());\
                             fprintf(stderr, __VA_ARGS__); \
                             fflush(stderr); } while (0)
#    else
#      define LOGPRINTF(...) do { time_t clock; struct tm *now;\
                             clock = time(0); now = localtime(&clock); \
                             fprintf(stderr,"[(%02d/%02d/%04d) %02d:%02d:%02d] [%s] %d %g LOG: ",\
                             now->tm_mon+1,now->tm_mday,now->tm_year+1900,\
                             now->tm_hour,now->tm_min,now->tm_sec,__FUNCNAME__,\
                             (int)getpid(), walltime());\
                             fprintf(stderr, __VA_ARGS__); \
                             fflush(stderr); } while (0)
#    endif
#  else
#       define LOGPRINTF printf
#  endif /* ifdef HAVE_ISO_VARARGS */
#endif /* LOGPRINTF */

#ifndef SENSORPRINTF
extern FILE *GS_SENSORFILE;
#  ifdef HAVE_ISO_VARARGS
#    define SENSORPRINTF(...) do { fprintf(GS_SENSORFILE, __VA_ARGS__); fflush(GS_SENSORFILE); } while (0)
#  else
#       define SENSORPRINTF printf
#  endif /* ifdef HAVE_ISO_VARARGS */
#endif /* SENSORPRINTF */

/* -------------------------------------------------- */

#endif /* GS_LOG_H */


