/**
 * @file
 * Header file for various problem specific definitions.
 *
 */
/* $Id: problem.h,v 1.62 2008/11/07 20:46:48 tbrady Exp $ */
/* $UTK_Copyright: $ */

#ifndef PROBLEM_H
#define PROBLEM_H

#include "proxylib.h"
#include "utility.h"
#include "dsi.h"

#define GS_END_PROB_REG "<end_of_problems>"
#define GS_BATCH_XML "batchp.xml"
#define GS_SERVER_XML_TEMPLATE "%s/server.xml"

extern char *gs_inout[];
extern char *gs_c_datatype[];
extern char *gs_objecttype[];
extern char *gs_probtype[];
extern char *gs_const[];
extern char *gs_service_error[];

typedef struct {
  float r;
  float i;
} gs_scomplex;

typedef struct {
  double r;
  double i;
} gs_dcomplex;

enum inout {
  GS_IN,
  GS_INOUT,
  GS_OUT,
  GS_WORKSPACE,
  GS_VAROUT,
  GS_BAD_INOUT
};

enum datatype {
  GS_INT,
  GS_FLOAT,
  GS_DOUBLE,
  GS_SCOMPLEX,
  GS_DCOMPLEX,
  GS_CHAR,
  GS_BAD_DTYPE
};

enum objecttype {
  GS_SCALAR,
  GS_VECTOR,
  GS_MATRIX,
  GS_SPARSEMATRIX,
  GS_FILE,
  GS_PACKEDFILE,
  GS_BAD_OTYPE
};

enum probtype {
  GS_SUBROUTINE,
  GS_FUNCTION,
  GS_BAD_PTYPE
};

typedef struct {
  /* GridSolve expects compressed sparse row CSR storage */
  int nnz;
  char *nnzexp;
  char *indices;		/* indices has size nnz */
  char *pointer;		/* pointer rows+1 */
  int rows_val_saved; /* saved while evaluating scalar expressions  */
  int cols_val_saved; /* saved while evaluating scalar expressions  */
} gs_sparse_attr;
  
typedef struct gs_arg_enum_s {
  char *val;
  struct gs_arg_enum_s *next;
} gs_arg_enum_t;

typedef struct gs_argument_s {
  char *name;
  char *description;
  enum inout inout;
  enum datatype datatype;
  enum objecttype objecttype;
  int dsi;
  DSI_OBJECT *dsi_obj;
  char *rowexp;
  int rows;
  char *colexp;
  int cols;
  gs_sparse_attr sparse_attr;
  void *data;
  double expr_val;              /* used in expression evaluation.  */
  union {
    int int_val;
    char char_val;
    float float_val;
    double double_val;
    gs_scomplex scomplex_val;
    gs_dcomplex dcomplex_val;
  } scalar_val;
  struct gs_problem_s *prob;
  struct gs_arg_enum_s *arg_enum;
  struct gs_argument_s *next;
  /* new fields added for GridSolve request sequencing */
  int data_handle; /* if 1, the data sent is a handle; one the server side,
					the server has to retrieve the argument from the server
		            that produced the argument via direct data transfer - 
				    GS_IN || GS_INOUT; (void *data) points to the data
			        handle, which is a LFS_DSI_OBJECT */
  char *hostname;  /* the host that produces the argument */
  /* a list of target servers that need the argument 
	 - GS_OUT -> GS_IN || GS_INOUT */
  struct gs_server_s **target_server_list;
  /* the server that output the argument - GS_INOUT || GS_OUT */
  struct gs_server_s *source;
  int num_targets; /* number of target servers */
  int pass_back; /* whether the argument should be passed back or not */
  int index; /* the index of the argument in the argument list */			  
#ifdef GS_SMART_GRIDSOLVE 
   /* num bytes of this arg */
  int byte_size;   
  /* specifies whether arg has been stored in memory or file */
  int arg_data_location;
  /* set to 1 if remote input arg should be deleted once used by remote task */
  int input_arg_deleted;      
  /* set to 1 if input arg is stored remotely for a subsequent remote task */
  int input_arg_stored;              
  /* set to 1 if this input arg should be sent to another server or cached */
  int input_arg_sent_remotely;      
  /* number of destination servers to send input arg */
  int input_nb_dest_servers;       
  /* if arg is located locally on server either in memory or store in a file */
  int input_arg_received_remotely;  
  /* specifies the file which to store the input argument */
  char * input_arg_file;           
  /* set to 1 if the input arg is stored locally on the same server */
  int input_cached;
  /* string encoding of the list of destination servers to send input arg */
  char ** enc_input_dest_servers;  

  /* set to 1 if this output arg is sent remotely or cached */
  int output_arg_sent_remotely;
  /* set to GS_CLIENT_SIDE if output sent on client side*/
  int send_output_side;
  /* number of destination servers to send output arg */
  int output_nb_dest_servers;
  /* set to one if output arg is stored locally on the same server */ 
  int output_cached;
  /* specifies the file which to store the output argument */
  char * output_arg_file;
  /* string encoding of the list of destination servers to send output arg */
  char ** enc_output_dest_servers;

#endif
} gs_argument_t;


typedef struct gs_problem_s {
  char *name;
  char *description;
  char *callsig;
  enum probtype type;
  int *work;
  int worksize;
  int major;
  gs_argument_t *arglist;
  gs_info_t *infolist;
  struct gs_problem_s *next;
  /* fields added for GridSolve request sequencing */ 
  int size_computed;
  int seq_id;
#ifdef GS_SMART_GRIDSOLVE 
  int has_smart_arg_comm;    /* is set to 1 if this problem has remote comm */
#endif

} gs_problem_t;

/* struct for passing back sparse matrix copy */

typedef struct {
  void *m;
  void *ind;
  void *ptr;
} gs_sparse_dup_t;

void 
  idl_lexer_free_memory(),
  gs_problemlist_dump(gs_problem_t *),
  gs_problem_dump(gs_problem_t *),
  gs_arglist_dump(gs_argument_t *),
  gs_arg_dump(gs_argument_t *), gs_infolist_dump(gs_info_t *);

enum inout 
  gs_parse_inout(const char *);

enum datatype 
  gs_parse_datatype(const char *);

enum objecttype 
  gs_parse_objecttype(const char *);

enum probtype 
  gs_parse_probtype(const char *);

gs_sparse_dup_t *
  gs_dup_sparse_mat(int, int, size_t, void *, int *, int *);

int 
  gs_expr_d(const char *, double *, icl_hash_t *),
  gs_expr_i(const char *, int *, icl_hash_t *),
  gs_get_var_list_from_expr(char *, icl_list_t *),
  gs_dup_problem(gs_problem_t *, gs_problem_t *),
  gs_gen_sparse_mat_double(int, int, int, double **, int **, int **),
  gs_gen_sparse_mat_float(int, int, int, float **, int **, int **),
  gs_gen_sparse_mat_int(int, int, int, int **, int **, int **),
  gs_gen_sparse_mat_char(int, int, int, char **, int **, int **),
  gs_create_error_file(char *, int),
  gs_create_timestamp_file(char *, char *, double),
  gs_free_problem(gs_problem_t *),
  gs_free_problem_and_data(gs_problem_t *),
  gs_create_request_id(char *),
  gs_read_problem_from_file(char *, gs_problem_t *),
  gs_free_infolist(gs_info_t *), 
  gs_idl_dump_info(gs_problem_t *, char *),
#ifdef GS_SMART_GRIDSOLVE
  gs_smart_free_arg(gs_argument_t * , int ),
#endif
 gs_free_arglist(gs_argument_t *, int);

char 
  *gs_problem_getinfo(gs_problem_t *, char *, char *);

int gs_arg_is_sparse_attr(char *name, gs_argument_t *arglist);
gs_argument_t *gs_arg_lookup_by_name(char *name, gs_argument_t *arglist);
int gs_datatype_sizeof(int datatype);
char *gs_problem_prototype(gs_problem_t *);

#endif                          /* PROBLEM_H */
