/**
 * @file
 *
 * Header file for the GridSolve server routines.
 *
*/
/* $Id: server.h,v 1.57 2008/12/11 19:03:38 tbrady Exp $ */
/* $UTK_Copyright: $ */

#ifndef SERVER_H
#define SERVER_H

#include "proxylib.h"
#include "problem.h"

/* filename length for various temp files */
#define FN_LEN 1024

#define GRIDSOLVE_SERVER_PORT_DEFAULT 9000

/* workload update frequency in seconds: */
#define GS_UPDATE_FREQUENCY 10

/* server ping frequency in seconds: */
#define GS_SERVER_PING_FREQUENCY 30

/* number of bytes to use for server pings */
#define GS_SERVER_PING_SIZE (32*1024)

/* maximum time to sleep between polling for job completion */
#define GS_MAX_COMPLETION_POLL_FREQ  30

/* prefix location of file to store current job count */
#ifdef __INTERIX
#define GS_SERVER_JOB_COUNT_FILE_PREFIX "gs_jc"
#else
#define GS_SERVER_JOB_COUNT_FILE_PREFIX "gs_jobcount"
#endif


/*SMARTGRIDSOLVE */

#ifdef GS_SMART_GRIDSOLVE
#define GS_SMART_OBJ_FILE_PREFIX "gs_smart_obj"
#endif

/* prefix location for info subdirectories */
#define GS_INFODIR_PATH   "/tmp"
#define GS_INFODIR_PREFIX "gs_"

/* message for indicating that there's no model to be sent */
#define GS_NO_MODEL_UPDATE "<no_model>"

/* message for indicating that the problem wasn't in the database */
#define GS_UNKNOWN_PROB "##bad#prob##"

/* message for indicating that there's no server ping information */
#define GS_NO_SERVER_PING_UPDATE "<no_pings>"

extern char GS_SERVER_JOB_COUNT_FILE[FN_LEN];

enum gs_restriction_enum {
  GS_JOB_RESTRICT=0,
  GS_TIME_RESTRICT
};

/* 
 * for the type-specific restrictions in the union, it's easier
 * not to point to dynamically allocated memory since it would
 * require separate free() routines for each type.
 */

typedef struct {
  int max_jobs;
} gs_job_restriction_t;

typedef struct {
  struct tm start, end;
} gs_time_restriction_t;

typedef struct gs_restriction_s {
  enum gs_restriction_enum restriction_type;

  union {
    gs_job_restriction_t job_restriction;
    gs_time_restriction_t time_restriction;
  } restriction;

  /* first arg is gs_server_t, but can't do forward reference, so use void */
  int (*check_func)(void *, struct gs_restriction_s *, char *, int *);

  struct gs_restriction_s *next;
} gs_restriction_t;

typedef struct gs_server_pings_s {
  char componentid[CID_LEN];
  double comm_time;
} gs_server_pings_t;

typedef struct gs_server_s {
  char *hostname;		/* hostname */
  ipaddr_t ipaddress;
  in_port_t port;
  ipaddr_t proxyip;
  in_port_t proxyport;
  char componentid[CID_LEN];
  char *arch;
  int data_format;		/* ieee, endian, etc */

  int kflops;
  int workload;
  int ncpu;
  int status;
  double availcpu;
  double availmem;
  
  double score;			/* used in comparing complexity */
  double comm_bw;
  double comm_time;

  int nproblems;
  gs_problem_t *problemlist;

  char *agenthost;
  int agentport;

  char *infodir;

  long output_ttl;              /* how many seconds to allow data to 
                                 * sit without being retrieved */
  time_t last_update;

  int smart;       /* TRUE if this server is SmartGridSolve enabled */

  gs_info_t *sa_list; /* Server Attributes parsed from server_config file */

  gs_restriction_t *restrictions;   /* list of server restrictions */

  char *server_pings;  /* server-to-server pings (for SmartGridSolve) */


  char *my_ping_str;  /* server-to-server pings (for SmartGridSolve) */
#ifdef GS_SMART_GRIDSOLVE
  int smart_id;          /* unique id given to each smart server */
  int nb_my_pings;       /* number of pings performed by this server */
  int nb_smart_servers;    /* number of smart servers in the network */
  gs_server_pings_t ** my_pings;  /* pings performed by this server */
  gs_server_pings_t ** all_pings; /* pings performed by all servers */
  double task_workload; /* workload of assigned tasks (used when mapping) */
#endif


  char *perf_expr;  /* for use only in the agent */
} gs_server_t;

typedef struct {
  unsigned char msgtype;
  char *agent_host;
  int agent_port;
  char agent_cid[CID_LEN];
  char server_cid[CID_LEN];
  int server_workload;
  int nproblems;
  char *coeff_update;
} gs_workload_packet;

/*** Function prototypes ***/

int 
  gs_free_restrictions(gs_restriction_t *),
  gs_do_ping(gs_server_t *, char *, int, double *),
  gs_notify_agent_problem_solve(char *, in_port_t,
    char *, double, char *, char *, char *, char *, char *, int, double),
  gs_notify_agent_problem_complete(char *, in_port_t,
    char *, char *, char *, char *, char *, char *, int, double),
  gs_is_current_time_in_range(struct tm *, struct tm *),
  gs_parse_time_range(char *, struct tm *, struct tm *),
  gs_server_process_message(gs_server_t *, int),
  gs_server_listen_and_process_messages(gs_server_t *, int),
  gs_server_register_problems(gs_server_t *),
  gs_server_register(gs_server_t *),
  gs_increment_job_count(char *),
  gs_decrement_job_count(char *),
  gs_init_job_count(char *),
  gs_get_job_count(char *, int *),
  gs_clean_up_old_temp_files(char *),
  gs_parse_pid_from_requestid(char *),
  gs_parse_pid_from_tempname(char *),
  gs_find_coefficient_files(gs_server_t *, icl_list_t *),
  gs_read_server_from_file(char *, gs_server_t *),
  gs_create_info_dir(char [CID_LEN], ipaddr_t, in_port_t, char **),
  gs_server_dump(gs_server_t *),
#ifdef GS_SMART_GRIDSOLVE
  gs_write_to_smart_obj_file(char *, char *, int ),
  gs_smart_read_update_app_no_file(char * , int * ),
  gs_smart_read_app_no_file(char * , int * ),
#endif 
  gs_server_free(gs_server_t *);

gs_server_t
  *gs_server_init(char *);

void 
  gs_workload_report(void **args),
  gs_server_ping(void **args);

#endif /* SERVER_H */
