/** 
 * @file
 *
 * This is a Yacc grammar representing the GridSolve client-specified
 * criteria.
 *
 * To use, call crit_parse() and upon return you'll find a 
 * SQL query in the global string crit_query.
 *
 */
/* $Id: criteria_parser.y,v 1.4 2007/02/22 02:54:06 seymour Exp $ */
/* $UTK_Copyright: $ */

%{
#include "criteria_parsing_defs.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "utility.h"
#include "gs_storage.h"

int crit_lex(void);
void crit_error(const char *);
void crit_parse_init(void);
int is_criteria_keyword(char *);

/* A global variable */
int parse_mode = GS_CRIT_PARSE_SQLITE;

char *crit_query = NULL;

icl_hash_t *table;

char *crit_keywords[] = {
  "HOSTNAME",
  "ARCH",
  "WORKLOAD",
  0
};
  
%}

%union {
  char *strp;
}

%token AND_OP OR_OP LE_OP GE_OP EQ_OP NE_OP

%token <strp> IDENT CONSTANT STRING_LITERAL

%type <strp> Start expression or_expr and_expr eq_expr relop_expr primary

%%

Start : { crit_parse_init(); } expression
          {
            char *kp, *vlist, *olist;
            icl_entry_t *tmpent;
            int tmpint;
            void *dp;

            vlist = dstring_sprintf("servers,");
            icl_hash_foreach(table, tmpint, tmpent, kp, dp) {
              vlist = dstring_append(vlist, tmpent->key);
              vlist = dstring_append(vlist, ",");
            }

            if(vlist[strlen(vlist)-1] == ',')
              vlist[strlen(vlist)-1] = '\0';

            olist = dstring_sprintf(" ON problem_server.componentid=servers.componentid");
            icl_hash_foreach(table, tmpint, tmpent, kp, dp) {
              char *tmp;

              tmp = dstring_sprintf(" AND problem_server.componentid=%s.componentid",
                tmpent->key);
                 
              olist = dstring_append(olist, tmp);

              free(tmp);
            }

            if(parse_mode == GS_CRIT_PARSE_SQLITE)
              crit_query = dstring_sprintf("SELECT servers.*,problem_server.perf_model_expr FROM problem_server JOIN %s %s WHERE problemname=%%Q AND %s;", vlist, olist, $2);
            else
              crit_query = dstring_sprintf("SELECT servers.*,problem_server.perf_model_expr FROM problem_server JOIN (%s) %s WHERE problemname='%%s' AND %s;", vlist, olist, $2);

            free($2);
            free(vlist);
            free(olist);
            icl_hash_destroy(table, NULL, free);
          }
;

expression: or_expr
          {
            $$ = $1;
          }
;

or_expr: and_expr
          {
            $$ = $1;
          }
        | or_expr OR_OP and_expr
          {
            $$ = dstring_sprintf(" %s OR %s ", $1, $3);
            free($1);
            free($3);
          }
;

and_expr: eq_expr
          {
            $$ = $1;
          }
        | and_expr AND_OP eq_expr 
          {
            $$ = dstring_sprintf(" %s AND %s ", $1, $3);
            free($1);
            free($3);
          }
;

eq_expr: relop_expr
          {
            $$ = $1;
          }
        | eq_expr EQ_OP relop_expr
          {
            $$ = dstring_sprintf(" %s = %s ", $1, $3);
            free($1);
            free($3);
          }
        | eq_expr NE_OP relop_expr
          {
            $$ = dstring_sprintf(" %s != %s ", $1, $3);
            free($1);
            free($3);
          }
;

relop_expr: primary
          {
            $$ = $1;
          }
        | relop_expr '<' primary
          {
            $$ = dstring_sprintf(" %s < %s ", $1, $3);
            free($1);
            free($3);
          }
        | relop_expr '>' primary
          {
            $$ = dstring_sprintf(" %s > %s ", $1, $3);
            free($1);
            free($3);
          }
        | relop_expr LE_OP primary
          {
            $$ = dstring_sprintf(" %s <= %s ", $1, $3);
            free($1);
            free($3);
          }
        | relop_expr GE_OP primary
          {
            $$ = dstring_sprintf(" %s >= %s ", $1, $3);
            free($1);
            free($3);
          }
;

primary:  IDENT
          {
            if(!is_criteria_keyword(crit_lval.strp)) {
              if(!icl_hash_find(table, crit_lval.strp)) {
                char *keydup;
                keydup = strdup(crit_lval.strp);
                icl_hash_insert(table, keydup, keydup);
              }
              $$ = dstring_sprintf("%s.value", crit_lval.strp);
            }
            else
              $$ = dstring_sprintf("%s", crit_lval.strp);

            free(crit_lval.strp);
          }
        | '+' CONSTANT
          {
            $$ = dstring_sprintf("+%s", crit_lval.strp);
            free(crit_lval.strp);
          }
        | '-' CONSTANT
          {
            $$ = dstring_sprintf("-%s", crit_lval.strp);
            free(crit_lval.strp);
          }
        |  CONSTANT
          {
            $$ = dstring_sprintf("%s", crit_lval.strp);
            free(crit_lval.strp);
          }
        |  STRING_LITERAL
          {
            $$ = dstring_sprintf("%s", crit_lval.strp);
            free(crit_lval.strp);
          }
        |  '(' or_expr ')'
          {
            $$ = dstring_sprintf("(%s)", $2);
            free($2);
          }
;

%%

char *
gs_parse_client_criteria(char *client_criteria, int mode)
{
  struct crit__buffer_state * crit__scan_string(char *);
  extern int crit_parse();
  extern char *crit_query;
  extern int parse_mode;
  char *yybuf;

  parse_mode = mode;

  yybuf = strdup(client_criteria);
  crit__scan_string(yybuf);
  crit_parse();
  free(yybuf);
  if(!crit_query) {
    ERRPRINTF("Failed to parse client criteria: '%s'\n", client_criteria);
    return NULL;
  }

  return crit_query;
}

int
is_criteria_keyword(char *s)
{
  int i, is_keyword;

  is_keyword = 0;
  for(i=0;crit_keywords[i];i++)
    if(!strcasecmp(s, crit_keywords[i])) {
      is_keyword = 1;
      break;
    }

  return is_keyword;
}

void 
crit_parse_init(void)
{
  table = icl_hash_create(11, NULL);
  if(!table)
    crit_error("could not create hash table");
}

void
crit_error(const char *msg)
{
  fprintf(stderr,"%s\n", msg);
  if(crit_query) free(crit_query);
  crit_query = NULL;
}
