/**
 * @file
 *
 * This file contains the Fortran client calls to netsolve
 */
/* $Id: fortranclient.c,v 1.8 2007/07/24 19:24:32 yarkhan Exp $ */
/* $UTK_Copyright: $ */

#include <stdarg.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include "netsolveclient.h"
#include "grpc.h"

/**
 * This function is used to make NetSolve calls from a Fortran client 
 * in a blocking fashion.
 *
 * @param nickname -- Specifies the name of the problem that needs to be solved
 *    This should have the form "problemname()".  Remember the parens.
 * @param ... -- Variable length arguments to accommodate variable 
 *    arguments required by the problem
 */

void
fnetsl(char *nickname, ...)
{
  va_list argptr;
  char *buf;
  int *return_value, len;

  grpc_set_default_major("COL");

  /* Getting the problem's nickname */
  va_start(argptr, nickname);

#ifdef F2CSTRCRAYSTYLE
  va_arg(argptr, int);
#endif

  /* Get the address of the return value */
  return_value = (int *) va_arg(argptr, int *);

  /* The trailing "()" in the problem name signifies the end of the
   * string.  Strip this off since it isn't used by the GridRPC client.
   */
  buf = nickname;
  while((*buf != '(') && (*buf != '\0'))
    buf++;

  len = (int) (buf - nickname + 2);

  buf = (char *) malloc(len + 1);

  if(!buf) {
    *return_value = NetSolveSystemError;
    return;
  }

  memcpy(buf, nickname, len);
  buf[len] = '\0';

  *return_value = netslX(buf, argptr, GS_CALL_FROM_FORTRAN,
                         NS_BLOCK, NS_NOASSIGNMENT);

  free(buf);

  return;
}

/**
 * This function is used to print the error message associated with 
 * a NetSolve return value.  prints to stderr.
 *
 * @param info  -- Specifies the return value of a previous call to netsolve
 */

void
fnetslerr(int *info)
{
  fprintf(stderr, "%s\n", netsolveErrorMessage(*info));
}

/**
 * This routine sets the major (row- or column-wise) used
 * by the client.
 *
 * @param maj -- Takes as input "[Rr]*" or "[Cc]*"
 */

void
fnetslmajor(char *maj)
{
  netslmajor(maj);
}

/**
 * This function is used to make NetSolve calls from a Fortran client
 * in a non-blocking fashion.  This function will return as soon as the
 * arguments have been sent to the server.
 *
 * @param nickname -- Specifies the name of the problem that needs to be solved
 *    This should have the form "problemname()".  Remember the parens.
 * @param ... -- Variable length arguments to accomodate variable
 *    arguments required by the problem
 */

void
fnetslnb(char *nickname, ...)
{
  va_list argptr;
  char *buf;
  int *return_value, len;

  grpc_set_default_major("COL");

  /* Getting the problem's nickname */
  va_start(argptr, nickname);

#if defined(F2CSTRCRAYSTYLE)
  va_arg(argptr, int);
#endif

  /* Get the address of the return value */
  return_value = (int *) va_arg(argptr, int *);

  buf = nickname;
  while((*buf != '(') && (*buf != '\0'))
    buf++;

  len = (int) (buf - nickname + 2);

  buf = (char *) malloc(len + 1);

  if(!buf) {
    *return_value = NetSolveSystemError;
    return;
  }

  memcpy(buf, nickname, len);
  buf[len] = '\0';

  *return_value = netslX(buf, argptr, GS_CALL_FROM_FORTRAN,
                         NS_NOBLOCK, NS_NOASSIGNMENT);

  free(buf);

  return;
}

/**
 * This function is used to in conjuction with the non-blocking call
 * to check whether a previously submitted asynchronous request has completed.
 * It does not block, but returns a value signifying whether the call has 
 * completed or not.
 *
 * @param request_id -- Specifies the id of the previous non-blocking request
 * @param *info -- Stores the return value, which will be NetSolveOK on
 *    success.  On failure it will contain the status specifying the error 
 *    that occured.
 */

void
fnetslpr(int *request_id, int *info)
{
  if(!info || !request_id) return;

  *info = netslpr(*request_id);
  return;
}


/**
 * This function is used to in conjuction with the non-blocking call
 * to wait for a previously submitted asynchronous request to complete.
 * It blocks until the request has finished.
 *
 * @param request_id -- Specifies the id of the previous non-blocking request
 * @param *info -- Stores the return value, which will be NetSolveOK on
 *    success.  On failure it will contain the status specifying the error 
 *    that occured.
 */

void
fnetslwt(int *request_id, int *info)
{
  if(!info || !request_id) return;

  *info = netslwt(*request_id);
  return;
}
