/**
 * @file
 *
 * The functions in this file implement GridSolve basic communication 
 * infrastructure.  Users wanting to use their own communication and 
 * authentication mechanism will have to rewrite these functions using 
 * their mechanisms. Users behind NATs are advised to use this 
 * communication library.
 *
 */
/* $Id: comm_basics.c,v 1.48 2008/12/10 19:00:41 seymour Exp $ */
/* $UTK_Copyright: $ */
#ifdef HAVE_CONFIG_H
#include "config.h"
#include "gridsolve-config.h"
#endif

#include <signal.h>
#include <errno.h>
#include <string.h>
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#include <sys/types.h>

#if defined (solaris)
#include <sys/filio.h>
#endif

#include "portability.h"
#include "proxylib.h"
#include "comm_basics.h"
#include "utility.h"

/**
 * Write bytes to a descriptor (socket).
 *
 * @param fildes -- the descriptor (socket) to write to
 * @param buf -- the buffer containing the data to be written
 * @param nbyte -- the number of bytes to write
 *
 * @returns the number of bytes actually written.  returns -1 on error.
 */

ssize_t
gs_twrite(SOCKET fildes, const void *buf, size_t nbyte)
{
  static int sig = 1;
  if (sig) {
    (void) gs_signal(GS_SIGPIPE, SIG_IGN);
    sig = 0;
  }
  return gs_writen(fildes, buf, nbyte);
}


/**
 * Write until all bytes have been written (borrowed from Stevens 
 * "Unix Network Programming").
 *
 * @param fd -- the descriptor (socket) to write to
 * @param vptr -- the buffer containing the data to be written
 * @param n -- the number of bytes to write
 *
 * @returns the number of bytes written (which should equal the
 *   parameter 'n' since this routine keeps writing until 'n'
 *   bytes have been written).  Returns -1 on error.
 */

ssize_t
gs_writen(SOCKET fd, const void *vptr, size_t n)
{
  size_t          nleft;
  ssize_t         nwritten;
  const char      *ptr;

  ptr = vptr;
  DBGPRINTF("ns_writen: write %d bytes to socket %d num %d char %c\n", 
     n, fd, *((int *)ptr), (char)*ptr);
  nleft = n;
  while (nleft > 0) {
    if ( (nwritten = write_socket(fd, ptr, nleft)) == SOCKET_ERROR) {
      if (errno_socket() == EINTR)
        nwritten = 0;           /* and call write() agai n */
      else
        return(-1);             /* error */
    }

    nleft -= nwritten;
    ptr   += nwritten;
  }
  return(n);
}

/**
 * Read bytes from a descriptor (socket).
 *
 * @param d -- the descriptor (socket) to read from
 * @param b -- the buffer containing the data to be read
 * @param n -- the number of bytes to read
 *
 * @returns the number of bytes actually read.  Returns -1 on error.
 */

int 
gs_tread(SOCKET d, char *b, int n)
{
  static int sig = 1;
  int retval = -1;

  if (sig) {
    (void) gs_signal(GS_SIGPIPE, SIG_IGN);
    sig = 0;
  }
  
  retval = proxy_tread(d, b, n, PROXY_TIMEOUT_DEFAULT);

  return retval;
}

/**
 * Reads from descriptor until newline.
 *
 * @param fd -- descriptor to read from
 * @param vptr -- pointer to buffer to store string
 * @param maxlen -- length of vptr
 *
 * @returns number of bytes read, 0 on EOF, or -1 on error.
 */

ssize_t
gs_readline(int fd, void *vptr, size_t maxlen)
{
  int n, rc;
  char c, *ptr;

  ptr = vptr;
  for (n = 1; n < (int)maxlen; n++) {
    if ((rc = proxy_tread(fd, &c, 1, PROXY_TIMEOUT_DEFAULT)) == 1) {
      *ptr++ = c;
      if (c == '\n')
        break;                  /* newline is stored, like fgets() */
    } else if (rc == 0) {
      if (n == 1)
        return (0);             /* EOF, no data read */
      else
        break;                  /* EOF, some data was read */
    } else
      return (-1);              /* error, errno set by read() */
  }

  *ptr = 0;                     /* null terminate like fgets() */
  return (n);
}

/**
 * Get a socket and bind to it. Use the proxy library (gs_socket,
 * gs_bind) to get a new socket descriptor and bind to it.
 *
 * @param port -- Port number to be bound (on return this 
 *   parameter is set to the actual value bound)
 * @param reuse -- If TRUE, the DO_REUSEADDR option is set
 *   to allow reuse of the requested port.
 *
 * @returns The socket descriptor on success, -2 if the requested
 *   port is already in use, or -1 on failure.
 */

SOCKET 
gs_establish_socket(in_port_t * port, int reuse)
{
  SOCKET desc;
  struct sockaddr_in addr;
  socklen_t addr_length;
  int option_value = 1;

  if ((desc = proxy_socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
    return INVALID_SOCKET;
  }

  /* TODO */
  if (reuse) {
    if (setsockopt(desc, SOL_SOCKET, SO_REUSEADDR,
		   (char *) (&option_value), sizeof(int)) == -1) {
      return INVALID_SOCKET;
    }
  }

  if(setsockopt(desc, IPPROTO_TCP, TCP_NODELAY, (char *) &option_value, 
       sizeof(int)) < 0)
    DBGPRINTF("Warning: failed to set TCP_NODELAY for fd %d\n", desc);

  memset((char *) &addr, 0x0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(*port);

  if (proxy_bind(desc, (struct sockaddr *) &addr, sizeof(addr))) {
    if (errno_socket() == EADDRINUSE) {
      proxy_close(desc);
      return -2;
    }
    proxy_close(desc);
    return INVALID_SOCKET;
  }

  addr_length = sizeof(addr);
  if((proxy_getsockname(desc, (struct sockaddr *) &addr, &addr_length)) == -1) {
    proxy_close(desc);
    return INVALID_SOCKET;
  }
  *port = ntohs(addr.sin_port);

  return (desc);
}

/** 
 * Bind to the first available port greater than or equal to the
 * parameter 'initport'.  After GS_NUM_PORTS_TO_TRY attempts it fails.
 *
 * @param initport -- the requested port number
 * @param found_port -- on return, set to the actual port number bound
 *
 * @returns descriptor for the bound socket.  Returns -1 on error.
 */

SOCKET 
gs_bind_to_first_available_port(in_port_t initport, in_port_t * found_port)
{
  in_port_t port;
  in_port_t tryport;
  SOCKET sock;

  port = initport;
  tryport = port;

  /* 
   * It's important that we don't reuse the addresses here, as we will 
   * just attempt the next port if the current one fails - no problem...  
   * This call is typically used where the exact port doesn't matter much, 
   * but where we have multiple processes on the local host concurrently 
   * trying this call with the same starting port number. The proxy does 
   * this, and setting reuse=1 in the gs_establish_socket() call will spawn 
   * an unholy number of proxy processes. 
   */

  while ((sock = gs_establish_socket(&tryport, 0)) == -2) {
    port++;
    tryport = port;
    if (port > initport + GS_NUM_PORTS_TO_TRY) {
      ERRPRINTF("Impossible to bind to a port\n");
      return INVALID_SOCKET;
    }
  }

  if (proxy_is_socket_error(sock))
    return INVALID_SOCKET;

  *found_port = tryport;

  if (*found_port == 0) {
    *found_port = initport;
  }

  return sock;
}

/**
 * Set socket as listening.
 *
 * @param sock -- the socket descriptor
 *
 * @returns 0 on success, -1 on failure.
 */

int 
gs_listen_on_socket(SOCKET sock)
{
  return proxy_listen(sock, PROXY_MAX_CONNECTIONS);
}

/** 
 * Wait for a connection on this socket.
 * 
 * @param lisetening_socket -- the socket to wait for connections on
 *
 * @returns the descriptor for the accepted socket.
 *   on failure, returns -1.
 */

int 
gs_accept_connection(SOCKET listening_socket)
{
  struct sockaddr_in addr;	/* INET socket address */
  socklen_t addrlen;		/* address length */
  int sock;

  addrlen = sizeof(addr);
  sock = proxy_accept(listening_socket, (struct sockaddr *) &addr, &addrlen);
  while (sock == INVALID_SOCKET && errno_socket() == EINTR) {
    sock = proxy_accept(listening_socket, (struct sockaddr *) &addr, &addrlen);
  }
  return sock;
}

/**
 * Close this socket.
 *
 * @param sock -- the socket to be closed
 *
 * @returns 0 on success, -1 on failure
 */

int 
gs_close_socket(SOCKET sock)
{
  return proxy_close(sock);
}

/**
 * Establish a connection to the specified host.
 *
 * @param ID -- the component ID of the host to connect to.
 * @param ipaddr -- the IP address of the host to connect to.
 *    If connecting through proxy, set this to 0.
 * @param port -- the port to connect to.
 * @param proxyIP -- the IP address of the host's proxy server.
 *    If not proxied, set this to 0.
 * @param proxyPort -- the port number of the proxy server.
 * 
 * @returns the connected socket descriptor on success, -1 on failure.
 */

SOCKET 
gs_connect_to_host(char *ID, ipaddr_t ipaddr, int port, 
      ipaddr_t proxyIP, int proxyPort)
{
  PROXY_COMPONENTADDR servaddr;
  char dottedip[20];
  char dottedproxy[20];
  SOCKET sockfd, option_value = 1;

  memcpy(servaddr.ID, ID, CID_LEN);

  servaddr.IP = ipaddr;
  servaddr.port = htons(port);
  servaddr.proxyIP = proxyIP;
  servaddr.proxyPort = htons(proxyPort);

  sockfd = proxy_socket(AF_INET, SOCK_STREAM, 0);

  if(setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY, (char *) &option_value, 
       sizeof(int)) < 0)
    DBGPRINTF("Warning: failed to set TCP_NODELAY for fd %d\n", sockfd);

  proxy_ip_to_str(servaddr.IP, dottedip);
  proxy_ip_to_str(servaddr.proxyIP, dottedproxy);
  DBGPRINTF("Connecting to %s:%d (proxy %s:%d)\n", dottedip, 
    ntohs(servaddr.port), dottedproxy, ntohs(servaddr.proxyPort));

  if (proxy_connect(sockfd, &servaddr) < 0) {
    ERRPRINTF("Could not connect to %s:%d (proxy %s:%d)\n", dottedip, 
      ntohs(servaddr.port), dottedproxy, ntohs(servaddr.proxyPort));
    proxy_close(sockfd);
    return INVALID_SOCKET;
  }

  return sockfd;
}

/**
 * Connect directly (no proxy) to the specified host/port.
 *
 * @param host -- hostname to connect to
 * @param port -- port to connect to
 *
 * @returns the connected socket descriptor on success, -1 on failure.
 */

SOCKET 
gs_connect_direct(char *host, int port)
{
  char agent_cid[CID_LEN];
  struct hostent *hp;
  ipaddr_t ipaddr;
  
  DBGPRINTF("Connecting to host %s port %d\n", host, port);

  if ((hp = gethostbyname(host)) == NULL) {
    errno = errno_socket();
    ERRPRINTF("Could not gethostbyname for %s (errno %d) \n", host, errno);
    perror("gethostbyname()");
    return INVALID_SOCKET;
  }

  memcpy((void *) &ipaddr, hp->h_addr_list[0], sizeof(ipaddr));
  
  /* all ones will match any component ID */
  memset(agent_cid, 0xFF, CID_LEN);

  return gs_connect_to_host(agent_cid, ipaddr, port, 0, 0);
}
