/** 
 * @file
 *
 * This is a Yacc grammar representing the GridSolve Interface
 * Definition Language (IDL).
 *
 * To use, call yyparse() and upon return you'll find a list of
 * problem structures pointed to by the global variable 'problemp'.
 *
 */
/* $Id: problem_idl_parser.y,v 1.50 2007/03/14 22:00:26 seymour Exp $ */
/* $UTK_Copyright: $ */

%{
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#ifdef WIN32 
#  include "portability.h"
#else
#  include <unistd.h>
#endif

#include "idl_parsing_defs.h"
#include "problem.h"
#include "utility.h"

typedef struct expr_s {
  double val;
  char *str;
  struct expr_s *next;
} expr_t;

#define YYDEBUG 1
#define YYERROR_VERBOSE 1
  
struct idl__buffer_state *idl__scan_string( const char *idl__str );
void idl__delete_buffer( struct idl__buffer_state * b );

/* Local function prototypes */

static expr_t * new_expr_node(double, char *);
static void free_expr_node(expr_t *);
static void free_expr_list(expr_t *);
static gs_argument_t *gs_new_arg(enum objecttype, char *, char *, enum inout,
  enum datatype, char *, char *, char *, char *, char *, gs_arg_enum_t *);
static char *gs_idl_complexity_expression(gs_problem_t *problem) ;
static int gs_idl_get_major(gs_problem_t *problem);
static double handle_intrinsic(char *, expr_t *);
static char * get_argument_string(expr_t *);
static int get_two_args(expr_t *, double *, double *);
static int get_one_arg(expr_t *, double *);

int idl_lex(void);
void yyerror(const char *);
int gs_copy_global_attributes(gs_problem_t *, gs_info_t *);

/* A global variable */
gs_problem_t *problemp = NULL;

extern int lineno;

/* AYK TODO; Check if there is some way to set first tok correctly */
#ifdef __INTERIX
int first_tok = 257; 
#else 
int first_tok = GSIDL_PARSE; 
#endif

icl_hash_t *symtab = NULL;
icl_list_t *var_list = NULL;

double expr_result;

%}

%union {
  char *strp;
  gs_argument_t *argp;
  gs_arg_enum_t *enump;
  gs_problem_t *problemp;
  gs_info_t *infop;
  enum inout inout;
  enum datatype datatype;
  int op;
  struct expr_s *expr;
}

%token GSIDL_PARSE EXPR_EVAL_TOK
%token RIGHT_OP LEFT_OP AND_OP OR_OP LE_OP GE_OP EQ_OP NE_OP ENUM_TOK
%token IN_TOK OUT_TOK INOUT_TOK WORKSPACE VAROUT FUNCTION SUBROUTINE INT_TOK FLOAT_TOK 
%token DOUBLE_TOK CHAR_TOK SHORT_TOK LONG_TOK VOID_TOK SCOMPLEX DCOMPLEX FILE_TOK

%token <strp> LANGUAGE MAJOR IDENTIFIER CONSTANT STRING_LITERAL
%token <strp> INCLUDES COMPLEXITY PARALLEL CODE LIBS

%type <problemp> Start Problemlist Problem
%type <strp> Description Infotype Dim SpDim
%type <strp> Identifier Constant StringLiteral
%type <infop> OptionalInfolist Infolist Info
%type <datatype> Datatype
%type <argp> Arglist Arg ProbSpec
%type <enump> EnumDecl EnumValList EnumVal
%type <inout> Inout

%type <expr> DimEvaluated primary_expression postfix_expression expression
%type <expr> argument_expression_list unary_expression
%type <expr> cast_expression multiplicative_expression additive_expression
%type <expr> shift_expression relational_expression equality_expression
%type <expr> and_expression exclusive_or_expression inclusive_or_expression
%type <expr> logical_and_expression logical_or_expression

%type <strp> type_specifier unary_operator

%start Start

%%

Start:   GSIDL_PARSE OptionalInfolist Problemlist
         {
           /* set the global before finishing */
           problemp = $$ = $3;

           gs_copy_global_attributes(problemp, $2);
         }
     |   EXPR_EVAL_TOK DimEvaluated
         {
           expr_result = $2->val;
           free_expr_node($2);
         }
;

Identifier: IDENTIFIER
            {
              $$ = strdup($1);
            }
;

Constant: CONSTANT
          {
            $$ = strdup($1);
          }
;

StringLiteral: STRING_LITERAL
               {
                 $$ = strdup($1);
               }
;

Problemlist:  Problemlist Problem
              {
                gs_problem_t *prob;

                /* skip to the end of the list and link the new entry */
                for(prob = $1; prob && prob->next; prob=prob->next)
                  /* spin */;
                prob->next = $2;
                $$ = $1;
              }
           |  Problem
              {
                $$ = $1;
              }
;

Problem:  ProbSpec Identifier '(' Arglist ')' Description OptionalInfolist
          {
            gs_problem_t *prob;

            prob = (gs_problem_t *)calloc(1, sizeof(gs_problem_t));
            if(!prob) {
              ERRPRINTF("Could not allocate memory\n");
              exit(-1);
            }
            prob->name = $2;
            prob->arglist = $4;
            prob->description = $6;
            prob->infolist = $7;
            prob->work = NULL;
            prob->worksize = 0;
            prob->next = NULL;

            if((prob->major = gs_idl_get_major(prob)) < 0) {
              yyerror("Error: bad MAJOR specified.  Must be ROW or COLUMN.\n");
              exit(-1);
            }

            /* if there is a return value specified, then add a new argument
             * at the end of the calling sequence.
             */
            if($1) {
              gs_argument_t *a;

              /* skip to the end of the argument list */
              for(a = $4; a && a->next; a=a->next)
                /* spin */;

              a->next = $1;
              $1->next = NULL;

              prob->type = GS_FUNCTION;
            }
            else
              prob->type = GS_SUBROUTINE;

            $$ = prob;
          }
;

ProbSpec:  FUNCTION Datatype '[' Dim ']' '[' Dim ']'
           {
             $$ = gs_new_arg(GS_MATRIX, "__retval", "Return value", 
                     GS_OUT, $2, $4, $7, "0", "0", "0", NULL);
             free($4);
             free($7);
           }
        |  FUNCTION Datatype '[' Dim ']'
           {
             $$ = gs_new_arg(GS_VECTOR, "__retval", "Return value", 
                     GS_OUT, $2, $4, "1", "0", "0", "0", NULL);
             free($4);
           }
        |  FUNCTION Datatype
           {
             $$ = gs_new_arg(GS_SCALAR, "__retval", "Return value", 
                     GS_OUT, $2, "1", "1", "0", "0", "0", NULL);
           }
        |  SUBROUTINE
           {
             $$ = NULL;
           }
;

OptionalInfolist: Infolist
                  {
                    $$ = $1;
                  }
                | /* empty */
                  {
                    $$ = NULL;
                  }
;

Infolist:  Infolist Info
           {
             gs_info_t *infop;

             /* skip to the end of the list and link the new entry */
             for(infop = $1; infop && infop->next; infop=infop->next)
               /* spin */;
             infop->next = $2;
             $$ = $1;
           }
        |  Info
           {
             $$ = $1;
           }
;

Info:  Infotype '=' StringLiteral
       {
         gs_info_t *infop;

         infop = (gs_info_t *)malloc(sizeof(gs_info_t));
         if(!infop) {
           ERRPRINTF("Could not allocate memory\n");
           exit(-1);
         }
         infop->type = $1;
         infop->value = $3;
         infop->next = NULL;
         $$ = infop; 
       }
;

Arglist:  Arglist ',' Arg
          {
            gs_argument_t *argp;

            /* skip to the end of the list and link the new entry */
            for(argp = $1; argp && argp->next; argp=argp->next)
              /* spin */;
            argp->next = $3;
            $$ = $1;
          }
       |  Arg
          {
            $$ = $1;
          }
       |  /*empty*/
          {
            $$=NULL;
          }
;

Arg:  Inout Datatype Identifier '[' Dim ']' '[' Dim ']' '{' SpDim ',' SpDim ',' SpDim '}' Description
      {
        $$ = gs_new_arg(GS_SPARSEMATRIX, $3, $17, $1, $2, $5, $8, $11, $13, $15, NULL);
        free($3);
        free($5);
        free($8);
        free($11);
        free($13);
        free($15);
        free($17);
      }
    | Inout Datatype Identifier '[' Dim ']' '[' Dim ']' Description
      {
        $$ = gs_new_arg(GS_MATRIX, $3, $10, $1, $2, $5, $8, "0", "0", "0", NULL);
        free($3);
        free($5);
        free($8);
        free($10);
      }
    | Inout Datatype Identifier '[' Dim ']' Description
      {
        $$ = gs_new_arg(GS_VECTOR, $3, $7, $1, $2, $5, "1", "0", "0", "0", NULL);
        free($3);
        free($5);
        free($7);
      }
    | Inout Datatype Identifier Description EnumDecl
      {
        $$ = gs_new_arg(GS_SCALAR, $3, $4, $1, $2, "1", "1", "0", "0", "0", $5);
        free($3);
        free($4);
      }
    | Inout FILE_TOK Identifier Description
      {
        $$ = gs_new_arg(GS_FILE, $3, $4, $1, GS_CHAR, "1", "1", "0", "0", "0", NULL);
        free($3);
        free($4);
      }
    | Inout FILE_TOK Identifier '[' Dim ']' Description
      {
        $$ = gs_new_arg(GS_PACKEDFILE, $3, $7, $1, GS_CHAR, $5, "1", "0", "0", "0", NULL);
        free($3);
        free($5);
        free($7);
      }
;

SpDim: Identifier
       {
         $$ = $1;
       }
;

Inout:  IN_TOK
        {
          $$ = GS_IN;
        }
     |  INOUT_TOK
        {
          $$ = GS_INOUT;
        }
     |  OUT_TOK
        {
          $$ = GS_OUT;
        }
     |  VAROUT
        {
          $$ = GS_VAROUT;
        }
     |  WORKSPACE
        {
          $$ = GS_WORKSPACE;
        }
;
     
Description:  StringLiteral
              {
                $$ = $1;
              }
           |  /*empty*/
              {
                $$ = strdup("none");
              }
;

EnumDecl: ENUM_TOK '=' '<' EnumValList '>'
          {
            $$ = $4;
          }
        | /*empty*/
          {
            $$ = NULL;
          }
;

EnumValList:  EnumValList ',' EnumVal
          {
            gs_arg_enum_t *argp;

            /* skip to the end of the list and link the new entry */
            for(argp = $1; argp && argp->next; argp=argp->next)
              /* spin */;
            argp->next = $3;
            $$ = $1;
          }
       |  EnumVal
          {
            $$ = $1;
          }
;

EnumVal: StringLiteral
         {
           $$ = (gs_arg_enum_t *)malloc(sizeof(gs_arg_enum_t));
           $$->val = $1;
           $$->next = NULL;
         }
;

Datatype : INT_TOK
           {
             $$ = GS_INT;
           }
         | CHAR_TOK
           {
             $$ = GS_CHAR;
           }
         | FLOAT_TOK
           {
             $$ = GS_FLOAT;
           }
         | SCOMPLEX
           {
             $$ = GS_SCOMPLEX;
           }
         | DCOMPLEX
           {
             $$ = GS_DCOMPLEX;
           }
         | DOUBLE_TOK
           {
             $$ = GS_DOUBLE;
           };

/* allows for arbitrary code word */
Infotype:  LANGUAGE 
           {
             $$ = strdup($1);
           }
        |  MAJOR 
           {
             $$ = strdup($1);
           }
        |  LIBS 
           {
             $$ = strdup($1);
           }
        |  INCLUDES 
           {
             $$ = strdup($1);
           }
        |  COMPLEXITY 
           {
             $$ = strdup($1);
           }
        |  PARALLEL 
           {
             $$ = strdup($1);
           }
        |  CODE 
           {
             $$ = strdup($1);
           }
        |  Identifier 
           {
             $$ = $1;
           }
;

DimEvaluated: expression
              {
                $$ = $1;
              }

Dim: expression
     {
       $$ = strdup($1->str);
       free_expr_node($1);
     }

primary_expression
	: Identifier
          {
            double val = 0.0;

            if(symtab) {
              void *ht;

              ht = icl_hash_find(symtab, $1);

              if(!ht)
                ERRPRINTF("Unknown identifier: %s\n", $1);
              else
                val = *((double *)ht);
            }

            if(var_list)
              if(!icl_list_search(var_list, $1, 
                   (int (*)(void*, void*)) strcmp))
                icl_list_append(var_list, strdup($1));

            $$=new_expr_node(val, $1);
          }
	| Constant
          {
            if($1[0] == '\'') {
              $$=new_expr_node((double)$1[1], $1);
              if(strlen($1) > 3)
                ERRPRINTF("truncating multi-character literal (%s to %c)\n",
                  $1, $1[1]);
            }
            else
              $$=new_expr_node(atof($1), $1);
          }
	| '(' expression ')'
          {
            char *s;

            s = dstring_sprintf("(%s)", $2->str);
            $$=new_expr_node($2->val, s);
            free_expr_node($2);
          }
	;

postfix_expression
	: primary_expression
          {
              $$=$1;
          }
	| Identifier '(' ')'
          {
            char *s;

            ERRPRINTF("Unsupported intrinsic: %s\n", $1);

            s = dstring_sprintf("%s()", $1);
            $$=new_expr_node(0.0, s);
            free($1);
          }
	| Identifier '(' argument_expression_list ')'
          {
            double val;
            char *s, *astr;

            val = handle_intrinsic($1, $3);
            astr = get_argument_string($3);
            s = dstring_sprintf("%s(%s)", $1, astr);
            free(astr);
            $$=new_expr_node(val, s);
            free_expr_list($3);
            free($1);
          }
	;

argument_expression_list
	: expression
          {
            $$=$1;
          }
	| argument_expression_list ',' expression
          {
            expr_t *argp;

            /* skip to the end of the list and link the new entry */
            for(argp = $1; argp && argp->next; argp=argp->next)
              /* spin */;
            argp->next = $3;
            $$ = $1;
          }
	;

unary_expression
	: postfix_expression
          {
            $$=$1;
          }
	| unary_operator cast_expression
          {
            double val;
            char *s;

            if(!strcmp($1,"+"))
              val = (double) +$2->val;
            else if(!strcmp($1,"-"))
              val = (double) -$2->val;
            else if(!strcmp($1,"~"))
              val = (double) ~(int)$2->val;
            else if(!strcmp($1,"!"))
              val = (double) !$2->val;
            else
              val = (double) $2->val;

            s = dstring_sprintf("%s%s", $1, $2->str);
            $$=new_expr_node(val, s);
            free($1);
            free_expr_node($2);
          }
	;

unary_operator
	: '+'
          {
            $$=strdup("+");
          }
	| '-'
          {
            $$=strdup("-");
          }
	| '~'
          {
            $$=strdup("~");
          }
	| '!'
          {
            $$=strdup("!");
          }
	;

cast_expression
	: unary_expression
          {
            $$=$1;
          }
	| '(' type_specifier ')' cast_expression
          {
            /* these are delcared to suppress a bunch of purify
             * errors..  don't ask, i have no idea....
             */
            char gs_char[5] = {"char"}, gs_short[6] = {"short"},
              gs_int[4] = {"int"}, gs_long[5] = {"long"},
              gs_float[6] = {"float"};
            double val;
            char *s;

            if(!strcmp($2, gs_char))
              val = (double) ((char)$4->val);
            else if(!strcmp($2, gs_short))
              val = (double) ((short)$4->val);
            else if(!strcmp($2, gs_int))
              val = (double) ((int)$4->val);
            else if(!strcmp($2, gs_long))
              val = (double) ((long)$4->val);
            else if(!strcmp($2, gs_float))
              val = (double) ((float)$4->val);
            else
              val = $4->val;

            s = dstring_sprintf("(%s)%s", $2, $4->str);
            $$=new_expr_node(val, s);
            free($2);
            free_expr_node($4);
          }
	;

multiplicative_expression
	: cast_expression
          {
            $$=$1;
          }
	| multiplicative_expression '*' cast_expression
          {
            double val;
            char *s;

            val = (double) ($1->val * $3->val);

            s = dstring_sprintf("%s*%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	| multiplicative_expression '/' cast_expression
          {
            double val;
            char *s;

            val = (double) ($1->val / $3->val);

            s = dstring_sprintf("%s/%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	| multiplicative_expression '%' cast_expression
          {
            double val;
            char *s;

            val = (double) ((int)$1->val % (int)$3->val);

            s = dstring_sprintf("%s%%%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	;

additive_expression
	: multiplicative_expression
          {
            $$=$1;
          }
	| additive_expression '+' multiplicative_expression
          {
            double val;
            char *s;

            val = (double) ($1->val + $3->val);

            s = dstring_sprintf("%s+%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	| additive_expression '-' multiplicative_expression
          {
            double val;
            char *s;

            val = (double) ($1->val - $3->val);

            s = dstring_sprintf("%s-%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	;

shift_expression
	: additive_expression
          {
            $$=$1;
          }
	| shift_expression LEFT_OP additive_expression
          {
            double val;
            char *s;

            val = (double) ((int)$1->val << (int)$3->val);

            s = dstring_sprintf("%s<<%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	| shift_expression RIGHT_OP additive_expression
          {
            double val;
            char *s;

            val = (double) ((int)$1->val >> (int)$3->val);

            s = dstring_sprintf("%s>>%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	;

relational_expression
	: shift_expression
          {
            $$=$1;
          }
	| relational_expression '<' shift_expression
          {
            double val;
            char *s;

            val = (double) ($1->val < $3->val);

            s = dstring_sprintf("%s<%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	| relational_expression '>' shift_expression
          {
            double val;
            char *s;

            val = (double) ($1->val > $3->val);

            s = dstring_sprintf("%s>%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	| relational_expression LE_OP shift_expression
          {
            double val;
            char *s;

            val = (double) ($1->val <= $3->val);

            s = dstring_sprintf("%s<=%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	| relational_expression GE_OP shift_expression
          {
            double val;
            char *s;

            val = (double) ($1->val >= $3->val);

            s = dstring_sprintf("%s>=%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	;

equality_expression
	: relational_expression
          {
            $$=$1;
          }
	| equality_expression EQ_OP relational_expression
          {
            double val;
            char *s;

            val = (double) ($1->val == $3->val);

            s = dstring_sprintf("%s==%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	| equality_expression NE_OP relational_expression
          {
            double val;
            char *s;

            val = (double) ($1->val != $3->val);

            s = dstring_sprintf("%s!=%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	;

and_expression
	: equality_expression
          {
            $$=$1;
          }
	| and_expression '&' equality_expression
          {
            double val;
            char *s;

            val = (double) ((int)$1->val & (int)$3->val);

            s = dstring_sprintf("%s&%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	;

exclusive_or_expression
	: and_expression
          {
            $$=$1;
          }
	| exclusive_or_expression '^' and_expression
          {
            double val;
            char *s;

            val = (double) ((int)$1->val ^ (int)$3->val);

            s = dstring_sprintf("%s^%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	;

inclusive_or_expression
	: exclusive_or_expression
          {
            $$=$1;
          }
	| inclusive_or_expression '|' exclusive_or_expression
          {
            double val;
            char *s;

            val = (double) ((int)$1->val | (int)$3->val);

            s = dstring_sprintf("%s|%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	;

logical_and_expression
	: inclusive_or_expression
          {
            $$=$1;
          }
	| logical_and_expression AND_OP inclusive_or_expression
          {
            double val;
            char *s;

            val = ((int)$1->val && (int)$3->val) ? 1.0 : 0.0;

            s = dstring_sprintf("%s&&%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	;

logical_or_expression
	: logical_and_expression
          {
            $$=$1;
          }
	| logical_or_expression OR_OP logical_and_expression
          {
            double val;
            char *s;

            val = ((int)$1->val || (int)$3->val) ? 1.0 : 0.0;

            s = dstring_sprintf("%s||%s", $1->str, $3->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
          }
	;

expression
	: logical_or_expression
          {
            $$=$1;
          }
	| logical_or_expression '?' expression ':' expression
          {
            double val;
            char *s;

            val = (int)$1->val ? $3->val : $5->val;

            s = dstring_sprintf("%s?%s:%s", $1->str, $3->str, $5->str);
            $$=new_expr_node(val, s);
            free_expr_node($1);
            free_expr_node($3);
            free_expr_node($5);
          }
	;

type_specifier
	: CHAR_TOK
          {
            $$=strdup("char");
          }
	| SHORT_TOK
          {
            $$=strdup("short");
          }
	| INT_TOK
          {
            $$=strdup("int");
          }
	| LONG_TOK
          {
            $$=strdup("long");
          }
	| FLOAT_TOK
          {
            $$=strdup("float");
          }
	| DOUBLE_TOK
          {
            $$=strdup("double");
          }
	;
%%

/**
 * Standard YACC error reporting function.
 *
 * @param msg -- The message to be printed regarding this error.
 */

void 
yyerror(const char *msg) { 
  fprintf(stderr, "%d: %s\n", lineno, msg);
}

/**
 * Return a string representation of the problem complexity in terms
 * of the input parameters.
 *
 * If there is no specified complexity expression, the use the sum of
 * all the scalar input parameters. For example, matrixoperation(int
 * N, int M, double*A, double*B) would return a default complexity of
 * "N+M" if nothing else was specified in the problem description.
 *
 * @param problem -- The problem to examine for its complexity expression.
 *
 * @returns The complexity string.  Returns NULL on failure.
 */

static char *
gs_idl_complexity_expression(gs_problem_t *problem) 
{
  char *complexity = NULL;
  int maxlen=0;
  gs_argument_t *arg = NULL;

  DBGPRINTF("generating complexity expression \n");

  complexity = gs_problem_getinfo(problem, "COMPLEXITY", NULL);

  if(!complexity) {
    /* all scalar input or input/output arguments will be involved
     * in computing the complexity, so compute the length of the 
     * string accordingly.
     */
    for (maxlen=0,arg=problem->arglist; arg!=NULL; arg=arg->next) 
      if(((arg->inout == GS_IN) || (arg->inout == GS_INOUT)) && 
          (arg->objecttype == GS_SCALAR))
        maxlen += strlen(arg->name)+3;

    complexity = (char*)malloc(maxlen * sizeof(char));
    if(!complexity) {
      ERRPRINTF("Could not allocate space for complexity string.\n");
      return NULL;
    }

    complexity[0] = '\0';

    /* now simply sum up the scalar input or input/output arguments */
    for (arg=problem->arglist; arg!=NULL; arg=arg->next) 
      if(((arg->inout == GS_IN) || (arg->inout == GS_INOUT)) &&
          (arg->objecttype == GS_SCALAR))
      {
        strcat(complexity, arg->name);
        strcat(complexity, "+");
      }

    /* clobber the trailing '+' */
    if(complexity[strlen(complexity)-1] == '+')
      complexity[strlen(complexity)-1] = '\0';

    return(complexity);
  }
  else
    return(strdup(complexity));
}

/**
 * Determines whether this routine is implemented in row-major
 * or column-major.  If not specified by the user in the IDL, then
 * assume that routines implemented in C are row-major and 
 * routines implemented in Fortran are column-major.
 *
 * @param problem -- The problem struct representing the routine to be called.
 *
 * @returns Returns 114 ('r') for row-major or 99 ('c') for column-major.
 *    Returns -1 on failure.
 */

static int
gs_idl_get_major(gs_problem_t *problem)
{
  char *major, *language;

  if(!problem) return -1;

  language = gs_problem_getinfo(problem, "LANGUAGE", "C");
  major = gs_problem_getinfo(problem, "MAJOR", NULL);

  /* if major wasn't specified, use default of COLUMN for Fortran
   * and ROW for C.
   */
  if(!major)
    return strcmp(language,"C") ? 'c' : 'r';

  if(!strcmp(major, "ROW"))
    return 'r';
  else if(!strcmp(major, "COLUMN"))
    return 'c';

  return -1;
}

/**
 * Creates a new argument struct to hold information about the
 * return type of a function.
 *
 * @param otype -- The object type of the return value.
 * @param name -- The argument name.
 * @param desc -- The argument description.
 * @param io -- The in/out mode.
 * @param dtype -- The data type of the return value.
 * @param rowexp -- The expression for the number of rows.
 * @param colexp -- The expression for the number of columns.
 * @param nnzexp -- The expression for the number of nonzeroes.  applies to
 *    sparse matrices only.
 * @param indexp -- The name of the index vector.  applies to
 *    sparse matrices only.
 * @param ptrexp -- The name of the pointer array.  applies to
 *    sparse matrices only.
 * @param arg_enum -- the optional enum spec.  set to NULL if none.
 *
 * @returns Returns the new argument structure.
 */

static gs_argument_t *
gs_new_arg(enum objecttype otype, char *name, char *desc, enum inout io,
  enum datatype dtype, char *rowexp, char *colexp, char *nnzexp, char *indexp,
  char *ptrexp, gs_arg_enum_t *arg_enum)
{
  gs_argument_t *tmp;

  tmp = (gs_argument_t *)malloc(sizeof(gs_argument_t));
  if(!tmp) {
    ERRPRINTF("Could not allocate memory\n");
    exit(-1);
  }

  tmp->name = strdup(name);
  tmp->description = strdup(desc);
  tmp->arg_enum = arg_enum;
  tmp->rowexp = strdup(rowexp);
  tmp->colexp = strdup(colexp);
  tmp->sparse_attr.nnzexp = strdup(nnzexp);
  tmp->sparse_attr.indices = strdup(indexp);
  tmp->sparse_attr.pointer = strdup(ptrexp);

  if(!tmp->name || !tmp->description || !tmp->rowexp || !tmp->colexp
    || !tmp->sparse_attr.nnzexp || !tmp->sparse_attr.indices 
    || !tmp->sparse_attr.pointer) 
  {
    ERRPRINTF("uh, bad news.  no memory\n");
    exit(-1);
  }

  tmp->inout = io;
  tmp->datatype = dtype;
  tmp->objecttype = otype;
  tmp->rows = 1;
  tmp->cols = 1;
  tmp->sparse_attr.nnz = 1;
  tmp->data = NULL;
  tmp->prob = NULL;
  tmp->next = NULL;

  return tmp;
}

/**
 * Creates a new expression node.
 *
 * @param v -- numeric value of the sub-expression
 * @param s -- string representation of the sub-expression
 *
 * @returns pointer to the new node or NULL on failure.
 **/

static expr_t *
new_expr_node(double v, char *s)
{
  expr_t *nn;

  nn = (expr_t *) malloc(sizeof(expr_t));

  if(!nn) return NULL;

  nn->val = v;
  nn->str = s;
  nn->next = NULL;

  if(!nn->str) {
    free(nn);
    return NULL;
  }

  return nn;
}

/**
 * Frees a list of expression nodes.
 *
 * @param n -- the nodelist to be freed
 **/

static void
free_expr_list(expr_t *n)
{
  expr_t *tmp;

  tmp = n;

  while(tmp) {
    expr_t *next;

    next = tmp->next;
    free_expr_node(tmp);
    tmp = next;
  }
}

/**
 * Frees an expression node.
 *
 * @param n -- the nodelist to be freed
 **/

static void
free_expr_node(expr_t *n)
{
  if(!n) return;

  if(n->str) free(n->str);
  free(n);
}

/**
 * Evaluates the specified expression.
 *
 * @param s -- string containing the expression to be evaluated
 * @param result -- upon return, set to the value (double-precision)
 * @param tab -- hash table containing values for the variables
 *
 * @returns 0 on success, -1 on failure.
 **/

int
gs_expr_d(const char *s, double *result, icl_hash_t *tab)
{
  struct idl__buffer_state* buffer_state;
  char *yybuf;
  int err;

  first_tok = EXPR_EVAL_TOK;
  symtab = tab;

  yybuf = strdup(s);
  buffer_state = idl__scan_string(yybuf);
  err = idl_parse();
  free((void*)yybuf);
  idl__delete_buffer(buffer_state);

  *result = expr_result;
  symtab = NULL;

  return err != 0 ? -1 : 0;
}

/**
 * Evaluates the specified expression.
 *
 * @param s -- string containing the expression to be evaluated
 * @param result -- upon return, set to the value (integer)
 * @param tab -- hash table containing values for the variables
 *
 * @returns 0 on success, -1 on failure.
 **/

int
gs_expr_i(const char *s, int *result, icl_hash_t *tab)
{
  double r;
  int status;

  status = gs_expr_d(s,&r,tab);
  *result = (int)r;
  return status;
}

/**
 * Evaluates intrinsic functions supported in our expression
 * grammar.  This currently includes pow, sqrt, abs, exp, log, 
 * and log10.
 *
 * @param func -- the name of the function
 * @param args -- list of expression nodes representing the
 *   arguments to the function
 *
 * @returns the value of the computed function
 **/

static double
handle_intrinsic(char *func, expr_t *args)
{
  double val;

  if(!args || !func) return 0.0;

  if(!strcmp(func, "pow")) {
    double arg1, arg2;

    if(get_two_args(args, &arg1, &arg2) < 0) {
      ERRPRINTF("Wrong number of args for intrinsic %s\n", func);
      return 0.0;
    }

    val = pow(arg1, arg2);
  }
  else if(!strcmp(func, "min")) {
    double arg1, arg2;

    if(get_two_args(args, &arg1, &arg2) < 0) {
      ERRPRINTF("Wrong number of args for intrinsic %s\n", func);
      return 0.0;
    }

    val = arg1 < arg2 ? arg1 : arg2;
  }
  else if(!strcmp(func, "max")) {
    double arg1, arg2;

    if(get_two_args(args, &arg1, &arg2) < 0) {
      ERRPRINTF("Wrong number of args for intrinsic %s\n", func);
      return 0.0;
    }

    val = arg1 > arg2 ? arg1 : arg2;
  }
  else {
    double arg1;

    if(get_one_arg(args, &arg1) < 0) {
      ERRPRINTF("Wrong number of args for intrinsic %s\n", func);
      return 0.0;
    }

    if(!strcmp(func, "abs")) {
      val = arg1 < 0 ? -arg1 : arg1;
    }
    else if(!strcmp(func, "sqrt")) {
      val = sqrt(arg1);
    }
    else if(!strcmp(func, "exp")) {
      val = exp(arg1);
    }
    else if(!strcmp(func, "log")) {
      val = log(arg1);
    }
    else if(!strcmp(func, "log10")) {
      val = log10(arg1);
    }
    else {
      ERRPRINTF("Unsupported intrinsic: %s\n", func);
      val = 0.0;
    }
  }

  return val;
}

/**
 * Verifies that there's only one arg in the list and returns it.
 *
 * @param args -- the argument list
 * @param arg1 -- upon return, the value of the argument
 *
 * @returns 0 on success, -1 on failure.
 **/

static int
get_one_arg(expr_t *args, double *arg1)
{
  if(!args) return -1;

  *arg1 = args->val;

  if(args->next)
    return -1;

  return 0;
}

/**
 * Verifies that there are only two args in the list and returns them.
 *
 * @param args -- the argument list
 * @param arg1 -- upon return, the value of the first argument
 * @param arg2 -- upon return, the value of the second argument
 *
 * @returns 0 on success, -1 on failure.
 **/

static int
get_two_args(expr_t *args, double *arg1, double *arg2)
{
  if(!args) return -1;

  *arg1 = args->val;

  if(!args->next)
    return -1;

  *arg2 = args->next->val;

  if(args->next->next)
    return -1;

  return 0;
}

/**
 * Create the string representation of a call to an intrinsic 
 * function.
 *
 * @param args -- the list of args to the function
 * 
 * @returns the string representing the call, NULL on failure.
 **/

static char *
get_argument_string(expr_t *args)
{
  expr_t *tmp;
  char *rv;
  int len = 0;

  for(tmp=args;tmp;tmp=tmp->next)
    len += strlen(tmp->str) + 1;

  rv = (char *)malloc(len + 1);
  rv[0] = '\0';

  for(tmp=args;tmp;tmp=tmp->next) {
    strcat(rv, tmp->str);
    if(tmp->next != NULL)
      strcat(rv, ",");
  }

  return rv;
}

/**
 * Builds a linked list of the names of all the variables
 * used in the specified expression.
 *
 * @param expr -- the expression
 * @param list -- upon return, the list contains all the variable names
 *
 * @returns 0 on success, -1 on failure.
 **/

int
gs_get_var_list_from_expr(char *expr, icl_list_t *list)
{
  struct idl__buffer_state* buffer_state;
  char *yybuf;
  int err;

  first_tok = EXPR_EVAL_TOK;
  var_list = list;

  yybuf = strdup(expr);
  buffer_state = idl__scan_string(yybuf);
  err = idl_parse();
  free((void*)yybuf);
  idl__delete_buffer(buffer_state);

  var_list = NULL;

  return err != 0 ? -1 : 0;
}

/**
 * Copies global problem attributes to the problem structure before generating
 * the code.
 *
 * @param prob_list -- the list of problem structures that were parsed in this file
 * @param attr_list -- the list of attributes specified at the top of the file
 *
 * @return 0 on success, -1 on failure.
 */

int
gs_copy_global_attributes(gs_problem_t *prob_list, gs_info_t *attr_list)
{
  gs_problem_t *prob;
  gs_info_t *attr, *new_attr;

  for(prob = prob_list; prob; prob = prob->next) {
    for(attr = attr_list; attr; attr = attr->next) {
      if(!gs_problem_getinfo(prob, attr->type, NULL)) {
        /* if the problem doesn't already have this attribute, then copy it */

        new_attr = (gs_info_t *)malloc(sizeof(gs_info_t));

        if(!new_attr) {
          ERRPRINTF("malloc failed\n");
          exit(-1);
        }

        new_attr->type = strdup(attr->type);
        new_attr->value = strdup(attr->value);

        if(!new_attr->type || !new_attr->value) {
          ERRPRINTF("malloc failed\n");
          exit(-1);
        }

        new_attr->next = prob->infolist;
        prob->infolist = new_attr;
      }
    }

    /* if complexity wasn't specified, generate a default */
    if(gs_problem_getinfo(prob, "COMPLEXITY", NULL) == NULL) {
      new_attr = (gs_info_t *)malloc(sizeof(gs_info_t));
      if(!new_attr) {
        ERRPRINTF("Could not allocate memory\n");
        exit(-1);
      }
      new_attr->type = strdup("COMPLEXITY");
      new_attr->value = gs_idl_complexity_expression(prob);
      if(!new_attr->value) new_attr->value = strdup("1.0");

      if(!new_attr->type || !new_attr->value) {
        ERRPRINTF("malloc failed\n");
        exit(-1);
      }

      new_attr->next = prob->infolist;
      prob->infolist = new_attr;
    }
  }


  return 0;
}
