/**
 * @file
 *
 * Implementation of a queue.
 */
/* $Id: queue.c,v 1.3 2004/09/29 20:47:31 seymour Exp $ */
/* $UTK_Copyright: $ */

#include "queue.h"

/** Frees the memory associated with the queue */

#define DESTROY_QUEUE()     { if(q->mutex) free(q->mutex); \
    if(q->notFull) free(q->notFull); \
    if(q->notEmpty) free(q->notEmpty); \
    free(q); }

/**
 * Creates a new queue.
 *
 * @returns pointer to the new queue
 */

QUEUE *
new_queue()
{
  QUEUE *q;

  q = (QUEUE *) malloc(sizeof(QUEUE));

  if (!q)
    return NULL;

  q->head = q->tail = NULL;

  q->mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
  q->notFull = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));
  q->notEmpty = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));

  if (!q->mutex || !q->notFull || !q->notEmpty) {
    DESTROY_QUEUE();
    return NULL;
  }

  if (pthread_mutex_init(q->mutex, NULL)) {
    DESTROY_QUEUE();
    return NULL;
  }

  if (pthread_cond_init(q->notFull, NULL)) {
    pthread_mutex_destroy(q->mutex);
    DESTROY_QUEUE();
    return NULL;
  }

  if (pthread_cond_init(q->notEmpty, NULL)) {
    pthread_mutex_destroy(q->mutex);
    pthread_cond_destroy(q->notFull);
    DESTROY_QUEUE();
    return NULL;
  }

  return q;
}

/**
 * Frees the memory associated with the specified queue.
 *
 * @param q -- the queue to be destroyed
 *
 * @returns 0 on success, -1 on failure
 */

int 
destroy_queue(QUEUE * q)
{
  pthread_mutex_destroy(q->mutex);
  pthread_cond_destroy(q->notFull);
  pthread_cond_destroy(q->notEmpty);

  while (dequeue(q))		/* spin */
    ;
  DESTROY_QUEUE();
  return 0;
}

/**
 * Adds an item to the front of the queue.
 *
 * @param q -- the queue to which the item should be added
 * @param val -- pointer to the item to be added to the queue
 *
 * @returns 0 on success, -1 on failure
 */

int 
enqueue(QUEUE * q, void *val)
{
  Q_ITEM *item;

  item = (Q_ITEM *) malloc(sizeof(Q_ITEM));

  if (!item)
    return -1;

  item->val = val;
  item->next = q->head;
  item->prev = NULL;

  if (q->head)
    q->head->prev = item;

  q->head = item;
  if (!q->tail)
    q->tail = item;

  return 0;
}

/**
 * Removes an item from the end of the queue.
 *
 * @param q -- the queue whose last item should be removed
 *
 * @returns pointer to the item removed from the queue
 */

void *
dequeue(QUEUE * q)
{
  void *retval;
  Q_ITEM *tail;

  if (!q->tail)
    return NULL;

  tail = q->tail;

  if (q->tail->prev)
    q->tail->prev->next = NULL;
  q->tail = q->tail->prev;

  if (!q->tail)
    q->head = NULL;

  retval = tail->val;
  free(tail);

  return retval;
}
