/********************************************************************/
/*                           gs_sequence.c                          */
/*        the source code for the GridSolve sequencing API          */
/*                       Yinan Li, 05/17/2007                       */
/********************************************************************/


#include "gs_sequence.h"

/* the global GridSolve sequence DAG */
GS_DAG_t *GS_SEQ_DAG = NULL;


/* the current sequencing mode, could be either
   OPTIMISTIC_MODE
   or
   RESTRICTIVE_MODE
   or
   CONSERVATIVE_MODE
   or
   OFF_MODE
   OFF_MODE can not be used as the initial mode
*/
int grpc_sequence_mode;

/* the sequencing language, could be either
   CALL_FROM_C
   or
   CALL_FROM_FORTRAN */
int grpc_sequence_language;

/* the sequencing major, could be either 'r' or 'c' */
int grpc_sequence_major;



/**
 * Mark the begin of a GridRPC sequence 
 *
 * @return
 */
void grpc_sequence_begin(int mode) {
    /* set the sequencing mode to what is specified */
    grpc_sequence_mode = mode;
    /* set the sequencing language to default */
    grpc_sequence_language = GS_CALL_FROM_C;
    /* set the sequencing major to what is specified */
    grpc_sequence_major = 'r';

    /* make a new empty DAG */
    GS_SEQ_DAG = make_new_GS_DAG();
    if (GS_SEQ_DAG == NULL) {
        fprintf(stderr, "can't make sequencing DAG\n");
        exit(1);
    }

	printf("\n**********************************************\n");
    printf("Beginning a sequence ...\n\n");
}


/**
 * Mark the end of a GridRPC sequence
 * Schedule and run the sequence
 * 
 * @param
 * @return
 */
double grpc_sequence_end(int n, ...) {
    int status;
	double bt, et;
    va_list arg_list;

    va_start(arg_list, n);

    /* ---------------------- analyze and schedule DAG --------------------- */

    printf("Analyzing and building the DAG ...\n");
    /* analyze the dependencies of the
       current sequencing DAG */
    if ((status = analyze_dep_GS_DAG(GS_SEQ_DAG)) < 0) {
        fprintf(stderr, "error analyzing DAG\n");
        exit(1);
    }

    /* output the analyzed DAG to a dot graph */
    output_GS_DAG_to_dot(GS_SEQ_DAG, "pre_analysis.dot"); 

    printf("Post-analyzing the DAG ...\n");
    /* post analysis for DAG optimization */
    if ((status = post_analysis_GS_DAG(GS_SEQ_DAG, arg_list, n)) < 0) {
        fprintf(stderr, "error doing post-DAG-analysis\n");
        exit(1);
    }

	/*output the analyzed DAG to a dot graph */
	output_GS_DAG_to_dot(GS_SEQ_DAG, "post_analysis.dot");
 
    printf("Mapping the DAG to servers ...\n");
	/* map the DAG to servers - scheduling */	
	if ((status = map_to_servers_GS_DAG(GS_SEQ_DAG, AGENT)) < 0) {
	    fprintf(stderr, "error mapping DAG to servers\n");
	    exit(1);
	}

	if (make_new_data_handle_list(GS_SEQ_DAG) < 0) {
		fprintf(stderr, "error creating data handle list\n");
		exit(1);
	}

	if (gs_seq_set_lfs_dsi_data_storage_path() < 0) {
		exit(1);
	}
    
	/* --------------------------- execute DAG ----------------------------- */
    printf("\nExecuting the DAG ...\n");
    
	bt = walltime();
	
	/* execute the DAG */
	if ((status = execute_GS_DAG(GS_SEQ_DAG)) < 0) {
        fprintf(stderr, "error executing DAG\n");
        exit(1);
    }
	
	et = walltime();

    /* ---------------------------- finalize ------------------------------- */
    /* turn off the sequencing mode */
    grpc_sequence_mode = OFF_MODE;

    /* free the DAG's memory */
    free_GS_DAG(GS_SEQ_DAG);
    GS_SEQ_DAG = NULL;

	/* free the data handle list */
	free_data_handle_list();

	/* remove all data files */
	if (chdir(DATA_SEND_FILE_PATH) < 0) {
		fprintf(stderr, "error changing directory\n");
		return -1;
	}

	system("rm *.dat");

    printf("\nSequence done.\n");
	printf("**********************************************\n\n");

	return et - bt;
}


/**
 * Submit a GridRPC sequence request 
 * 
 * @param name -- the name of the problem to be solved 
 * @param
 * 
 * @return 0 on success, -1 on failure
 */
int grpc_submit(char *name, ...) {
    gs_va_list *arg_list;

    if (name == NULL) return -1;

    arg_list = (gs_va_list *) malloc(sizeof(gs_va_list));
    if (arg_list == NULL) {
        perror("malloc");
        exit(1);
    }

    /* create the variable argument list */
    va_start(arg_list->args, name);

    /* insert a new node into the DAG */
    return insert_node_GS_DAG(GS_SEQ_DAG, name, arg_list, NULL);
}


/**
 * Submit a GridRPC sequence request
 * 
 * @param name -- the name of the problem to be solved 
 * @param
 * 
 * @return 0 on success, -1 on failure
 */
int grpc_submit_arg_stack(char *name, grpc_arg_stack *arg_stack) {
    if (name == NULL || arg_stack == NULL) return -1;
    
    /* insert a new node into the DAG */
    return insert_node_GS_DAG(GS_SEQ_DAG, name, NULL, arg_stack);
}


/**
 * Get the current sequencing mode
 * 
 * @param
 *
 * @return the current sequencing mode
 */
int grpc_get_sequence_mode() {
    /* return the current sequencing mode */
    return grpc_sequence_mode;
}


/**
 * Get the current global GridSolve sequencing DAG
 * 
 * @param
 *
 * @return the DAG
 */
GS_DAG_t *grpc_get_sequence() {
    return GS_SEQ_DAG;
}
