/*
 *  @file
 *
 *  This file contains functions for manipulating
 *  the task to server mapping matrix and 
 *  task to server mapping vector which are used to generate
 *  the mapping solutions.
 */




#include "config.h"

#ifdef GS_SMART_GRIDSOLVE
#include "gs_smart_task_graph.h"
#include "gs_smart_mapping_graph.h"


/*
 * This function is used to iterate through every possible task to server 
 * assignment.  The input mapping_matrix is a binary matrix which records 
 * the assignment of tasks to servers.  This function increments the 
 * mapping_matrix to the next task-to-server assignment.
 *
 * @param mapping_matrix -- binary matrix which represent task to server assignemnt
 * @param number -- the current solution number
 * @param base - number of servers
 * @param nb_tasks -- number of tasks
 *
 */



int gs_smart_next_permutation(int ** mapping_matrix, int number, int base, int nb_tasks){
  int mod;
  int multiple=number;
  int fin=0;
  int col_nb=0;

  if(!mapping_matrix){
    ERRPRINTF("SMART: Error could not allocate memory to mapping matrix\n");
    return -1;
  }

  while(col_nb<nb_tasks){
    if(multiple>=base&&col_nb<nb_tasks){
      mod = (multiple%base);
      mapping_matrix[mod][col_nb]=1;
      multiple=(int)multiple/base;
      col_nb++;
    }
    if(fin==1&&col_nb<nb_tasks){
     mapping_matrix[0][col_nb]=1;
     col_nb++;
    }
    else{
      if(multiple<base&&col_nb<nb_tasks){
        mapping_matrix[multiple][col_nb]=1;
        col_nb++;
        fin=1;
      }
    }
  }
  return 0;
}






/*
 * This function resets/zeros the input binary mapping_matrix
 *
 * @param mapping_matrix -- binary matrix which represent task to server assignemnt
 * @param nb_servers - number of servers
 * @param nb_tasks -- number of tasks
 *
 */


int gs_smart_zero_mapping_matrix(int **mapping_matrix, int nb_servers,  int nb_tasks)
{
   int i;
   if(!mapping_matrix)  return -1;

   if((nb_tasks<=0)||(nb_servers<=0)){
     ERRPRINTF("SMART: Number of tasks or servers is not greater zero\n");
     return -1;
   }
   int j;
  for( i = 0; i < nb_servers; i++){
    for( j = 0; j < nb_tasks; j++){
      mapping_matrix[i][j] = 0;
    }
  }
  return 0;
}


/*
 * This function converts the  binary mapping_matrix to a task to server vector.
 * This vector is used to generate a mapping solution.
 *
 * @param mapping_matrix -- binary matrix which represent task to server assignemnt
 * @param nb_servers - number of servers
 * @param nb_tasks -- number of tasks
 * @param mapping_vector -- task to server assignment vector
 *
 */


int gs_smart_matrix_to_vec(int ** mapping_matrix, int nb_servers, int nb_tasks, int * mapping_vector){
  if(!mapping_matrix) return -1;

  int i, j;
  for(i=0;i<nb_servers;i++){
    for(j=0;j<nb_tasks;j++){
      if(mapping_matrix[i][j]==1){
        mapping_vector[j]=i;
      }
    }
  }

  return 0;
}


#endif
