/**
 * @file
 *
 * Tests whether probe, wait, and cancel work when the child process 
 * that is sending the input has not exited yet.  We send a large amount
 * of data to ensure that the child hasn't finished by the time we call
 * grpc_probe(), etc.  We also test using a small data size to check 
 * whether it realizes the child process has terminated.
 *
 */
/* $Id: async_test.c,v 1.6 2008/10/23 21:55:18 seymour Exp $ */
/* $UTK_Copyright: $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_VALUES_H
#include<values.h>
#else
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_FLOAT_H
#include <float.h>
#define MAXDOUBLE DBL_MAX
#endif
#endif

#include "portability.h"
#include "grpc.h"
#include "totaltest.h"

int submit(grpc_function_handle_t *, int *, int, int);

#define PROBE 1
#define WAIT 2
#define CANCEL 3

#define LARGE_DATA_SIZE 15000000
#define SMALL_DATA_SIZE 5

int
main(int argc, char *argv[])
{
  int i, *x, *y;
  grpc_function_handle_t handle;

  if(grpc_initialize(NULL) != GRPC_NO_ERROR) {
    grpc_perror("grpc_initialize");
    die(EXIT_FAILURE);
  }

  if(grpc_function_handle_default(&handle, "vpass_int") != GRPC_NO_ERROR) {
    fprintf(stderr,"Error creating function handle\n");
    die(EXIT_FAILURE);
  }

  if(argc == 3) {
    int data_size;

    data_size = atoi(argv[2]);

    x = (int*)malloc(data_size*sizeof(x[0]));

    printf("%s test for data size %lu bytes\n", argv[1], (unsigned long)(data_size*sizeof(x[0])));

    if(!strcasecmp("probe", argv[1]))
      submit(&handle, x, data_size, PROBE);
    else if(!strcasecmp("wait", argv[1]))
      submit(&handle, x, data_size, WAIT);
    else if(!strcasecmp("cancel", argv[1]))
      submit(&handle, x, data_size, CANCEL);
    else
      fprintf(stderr, "Bad test type '%s', no test performed.\n", argv[1]);
  }
  else {
    x = (int*)malloc(LARGE_DATA_SIZE*sizeof(x[0]));
    y = (int*)malloc(SMALL_DATA_SIZE*sizeof(y[0]));

    for(i=0;i<LARGE_DATA_SIZE;i++)
      x[i] = i;
    for(i=0;i<SMALL_DATA_SIZE;i++)
      y[i] = i;

    printf("Probe test for large size:\n");
    submit(&handle, x, LARGE_DATA_SIZE, PROBE);

    printf("Probe test for small size:\n");
    submit(&handle, y, SMALL_DATA_SIZE, PROBE);

    printf("Wait test for large size:\n");
    submit(&handle, x, LARGE_DATA_SIZE, WAIT);

    printf("Wait test for small size:\n");
    submit(&handle, y, SMALL_DATA_SIZE, WAIT);

    printf("Cancel test for large size:\n");
    submit(&handle, x, LARGE_DATA_SIZE, CANCEL);

    printf("Cancel test for small size:\n");
    submit(&handle, y, SMALL_DATA_SIZE, CANCEL);
  }

  grpc_function_handle_destruct(&handle);

  printf("Finalizing...\n");

  grpc_finalize();

  printf("Done.\n");

  die(EXIT_SUCCESS);
}

int
submit(grpc_function_handle_t *handle, int *vec, int len, int method)
{
  grpc_sessionid_t sessionID;
  grpc_error_t status;
  struct timeval tv;

  status = grpc_call_async(handle, &sessionID, vec, len);

  if(status != GRPC_NO_ERROR) {
    grpc_perror("grpc_call");
    die(sessionID);
  }

  switch(method) {
    case PROBE:
      printf("Probing for result.\n");

      while(1) {
        status = grpc_probe(sessionID);

        if(status == GRPC_NO_ERROR) {
          printf("Job complete.\n");
          break;
        }
        else if(status != GRPC_NOT_COMPLETED) {
          fprintf(stderr,"grpc_probe failed\n");
          die(EXIT_FAILURE);
        }

        printf("Not finished, sleeping....\n");
        sleep(1);
      }

      /* no break; fall through to wait after probe */

    case WAIT:
      status = grpc_wait(sessionID);

      if(status != GRPC_NO_ERROR) {
        fprintf(stderr,"grpc_wait failed\n");
        die(EXIT_FAILURE);
      }

      break;

    case CANCEL:
      tv.tv_sec = 0;
      tv.tv_usec = 100000;
    
      /* waste a bit of time */
      select(0, NULL, NULL, NULL, &tv);

      printf("Going to cancel.\n");
      status = grpc_cancel(sessionID);

      if(status == GRPC_NO_ERROR)
        printf("Successfully cancelled.\n");
      else
        printf("Could not cancel.\n");

      break;
  }

  return 0;
}
