/**
 * Test of dgemm (matrix multiply).
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>

#include "grpc.h"
#include "totaltest.h"

int set_scale(double *);
int set_trans(int *);

grpc_sessionid_t call_dgemm(double *a, double *b, double *c, int m, int n, int k, 
  int transa, int transb, double alpha, double beta, int async,
  grpc_function_handle_t *handle);

int main(int argc, char *argv[])
{
  int i, min, max, num_tests, range, async;
  grpc_function_handle_t *handles;
  struct timeval tv;
  int *m, *n, *k, *transa, *transb;
  double *alpha, *beta, **a, **b, **c;
  grpc_sessionid_t *ids;
  grpc_error_t status;

  if((argc != 4) && (argc != 5)) {
    fprintf(stderr, "Usage: %s <min> <max> <num_tests> [async/sync]\n", argv[0]);
    die(EXIT_FAILURE);
  }

  if(grpc_initialize("dgemm_random.cfg") != GRPC_NO_ERROR) {
    grpc_perror("grpc_initialize");
    return -1;
  }

  gettimeofday(&tv, NULL);

  srand((int)tv.tv_usec);

  min = atoi(argv[1]);
  max = atoi(argv[2]);
  num_tests = atoi(argv[3]);
  async = (argc == 5) && !strcmp("async", argv[4]);

  range = max-min;

  if(range < 0) {
    fprintf(stderr, "Bad min/max values\n");
    die(EXIT_FAILURE);
  }
  else if(range == 0)
    range = 1;

  handles = (grpc_function_handle_t *)malloc(num_tests *sizeof(grpc_function_handle_t));
  m = (int *)malloc(num_tests * sizeof(int));
  n = (int *)malloc(num_tests * sizeof(int));
  k = (int *)malloc(num_tests * sizeof(int));
  transa = (int *)malloc(num_tests * sizeof(int));
  transb = (int *)malloc(num_tests * sizeof(int));
  alpha = (double *)malloc(num_tests * sizeof(double));
  beta = (double *)malloc(num_tests * sizeof(double));
  a = (double **)malloc(num_tests * sizeof(double *));
  b = (double **)malloc(num_tests * sizeof(double *));
  c = (double **)malloc(num_tests * sizeof(double *));
  ids = (grpc_sessionid_t *)malloc(num_tests * sizeof(grpc_sessionid_t));

  if(!handles || !m || !n || !k || !transa || !transb || !alpha || !beta
       || !a || !b || !c || !ids) {
    fprintf(stderr, "malloc error\n");
    die(EXIT_FAILURE);
  }

  for(i=0;i<num_tests;i++) {
    printf("binding handle %d\n", i);
    if(grpc_function_handle_init(&handles[i], "bind_servers_at_call_time", "dgemm") != GRPC_NO_ERROR) {
      fprintf(stderr,"Error creating function handle\n");
      die(EXIT_FAILURE);
    }
   
    m[i] = (rand() % range) + min;
    n[i] = (rand() % range) + min;
    k[i] = (rand() % range) + min;

    set_trans(&transa[i]);
    set_trans(&transb[i]);
    set_scale(&alpha[i]);
    set_scale(&beta[i]);

    a[i] = (double *)malloc(m[i]*k[i]*sizeof(*a[i]));
    b[i] = (double *)malloc(k[i]*n[i]*sizeof(*b[i]));
    c[i] = (double *)malloc(m[i]*n[i]*sizeof(*c[i]));

printf("matrices = %p %p %p\n", a[i], b[i], c[i]);
    if(!a[i] || !b[i] || !c[i]) {
      fprintf(stderr, "malloc error, not submitting any more requests\n");
      break;
    }

    printf("making non-blocking call %d\n", i);
    ids[i] = call_dgemm(a[i], b[i], c[i], m[i], n[i], k[i],
        transa[i], transb[i], alpha[i], beta[i], async, &handles[i]);
printf("ids[i] = %d\n", ids[i]);
  }

  printf("waiting for all results.\n");

  status = grpc_wait_all();

  if(status != GRPC_NO_ERROR) {
    printf("Warning: one or more calls failed.\n");
    die(EXIT_FAILURE);
  }

  die(EXIT_SUCCESS);
}

int
set_trans(int *a)
{
  *a = (rand() % 2) == 0 ? 'N' : 'T';

  return 0;
}

int
set_scale(double *s)
{
  switch(rand() % 3) {
    case 0:
      *s = 0.0;
      break;
    case 1:
      *s = 1.0;
      break;
    case 2:
      *s = 432.3;
      break; 
  }

  return 0;
}

grpc_sessionid_t
call_dgemm(double *a, double *b, double *c, int m, int n, int k, 
  int transa, int transb, double alpha, double beta, int async,
  grpc_function_handle_t *handle)
{
  grpc_sessionid_t sessionID; 
  grpc_error_t status;
  int lda, ldb, ldc;

  if(transa=='N')
    lda = m;
  else
    lda = k;

  if(transb=='N')
    ldb = k;
  else
    ldb = n;

  ldc = m;

  printf("m=%d, n=%d, k=%d, transa=%c, transb=%c, alpha=%g, beta=%g\n",
     m, n, k, transa, transb, alpha, beta);

  status = grpc_call_async(handle, &sessionID, transa,transb,m,n,k,alpha,a,lda,b,ldb,beta,c,ldc);

  if(status != GRPC_NO_ERROR) {
    printf("GRPC error status  = %d\n", status);
    grpc_perror("grpc_call_async");
    return -1;
  }

  return sessionID;
}
