/**
 * Test of dgemm (matrix multiply).
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>

#include "grpc.h"
#include "totaltest.h"

int set_scale(double *);
int set_trans(int *);
int run_dgemm_test(int, int, int, int, int,
  double, double, int, grpc_function_handle_t *);

int main(int argc, char *argv[])
{
  int i, min, max, num_tests, range, async;
  grpc_function_handle_t handle;
  struct timeval tv;

  if((argc != 4) && (argc != 5)) {
    fprintf(stderr, "Usage: %s <min> <max> <num_tests> [async/sync]\n", argv[0]);
    die(EXIT_FAILURE);
  }

  if(grpc_initialize(NULL) != GRPC_NO_ERROR) {
    grpc_perror("grpc_initialize");
    return -1;
  }

  if(grpc_function_handle_init(&handle, "bind_servers_at_call_time", "dgemm") != GRPC_NO_ERROR) {
    fprintf(stderr,"Error creating function handle\n");
    return -1;
  }

  gettimeofday(&tv, NULL);

  srand((int)tv.tv_usec);

  min = atoi(argv[1]);
  max = atoi(argv[2]);
  num_tests = atoi(argv[3]);
  async = (argc == 5) && !strcmp("async", argv[4]);

  range = max-min;

  if(range < 0) {
    fprintf(stderr, "Bad min/max values\n");
    die(EXIT_FAILURE);
  }

  for(i=0;i<num_tests;i++) {
    int m, n, k, transa, transb;
    double alpha = 0.0, beta = 0.0;
   
    m = (rand() % range) + min;
    n = (rand() % range) + min;
    k = (rand() % range) + min;

    set_trans(&transa);
    set_trans(&transb);
    set_scale(&alpha);
    set_scale(&beta);

    run_dgemm_test(m, n, k, transa, transb, alpha, beta, async, &handle);
  }

  die(EXIT_SUCCESS);
}

int
set_trans(int *a)
{
  *a = (rand() % 2) == 0 ? 'N' : 'T';

  return 0;
}

int
set_scale(double *s)
{
  switch(rand() % 3) {
    case 0:
      *s = 0.0;
      break;
    case 1:
      *s = 1.0;
      break;
    case 2:
      *s = 432.3;
      break; 
  }

  return 0;
}

int
run_dgemm_test(int m, int n, int k, int transa, int transb,
  double alpha, double beta, int async, grpc_function_handle_t *handle)
{
  grpc_error_t status;
  int lda, ldb, ldc;
  double *a, *b, *c;

  a = (double *)malloc(m*k*sizeof(*a));
  b = (double *)malloc(k*n*sizeof(*b));
  c = (double *)malloc(m*n*sizeof(*c));

  if(!a || !b || !c) {
    fprintf(stderr, "Malloc failed\n");
    goto test_failed;
  }

  if(transa=='N')
    lda = m;
  else
    lda = k;

  if(transb=='N')
    ldb = k;
  else
    ldb = n;

  ldc = m;

  printf("m=%d, n=%d, k=%d, transa=%c, transb=%c, alpha=%g, beta=%g\n",
     m, n, k, transa, transb, alpha, beta);

  if(!async) {
    status = grpc_call(handle,transa,transb,m,n,k,alpha,a,lda,b,ldb,beta,c,ldc);

    if(status != GRPC_NO_ERROR) {
      printf("GRPC error status  = %d\n", status);
      grpc_perror("grpc_call");
      goto test_failed;
    }
  }
  else {
    grpc_sessionid_t sessionID; 

    status = grpc_call_async(handle, &sessionID, transa,transb,m,n,k,alpha,a,lda,b,ldb,beta,c,ldc);

    if(status != GRPC_NO_ERROR) {
      printf("GRPC error status  = %d\n", status);
      grpc_perror("grpc_call_async");
      goto test_failed;
    }

    printf("Waiting for result.\n");
  
    status = grpc_wait(sessionID);
  
    if(status != GRPC_NO_ERROR) {
      grpc_perror("grpc_wait");
      goto test_failed;
    }
  }

  if(a) free(a);
  if(b) free(b);
  if(c) free(c);

  return 0;

test_failed:
  if(a) free(a);
  if(b) free(b);
  if(c) free(c);

  return -1;
}
