/**
 * Test of dgemm (matrix multiply).
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>

#include "grpc.h"
#include "totaltest.h"

typedef struct {
  double *a, *b, *c, alpha, beta;
  int m, n, k, transa, transb;
  grpc_sessionid_t id;
  grpc_function_handle_t handle;
} dgemm_task_t;

grpc_sessionid_t call_dgemm(dgemm_task_t *);
void set_random_fields(dgemm_task_t *, int, int);
int set_scale(double *);
int set_trans(int *);

int main(int argc, char *argv[])
{
  int i, j, min, max, num_tests, range, window_size, outstanding, remaining;
  struct timeval tv;
  grpc_sessionid_t completed_id;
  grpc_error_t status;
  dgemm_task_t *tasks;

  if(argc != 5) {
    fprintf(stderr, "Usage: %s <min> <max> <num_tests> <window_size>\n", argv[0]);
    die(EXIT_FAILURE);
  }

  if(grpc_initialize(NULL) != GRPC_NO_ERROR) {
    grpc_perror("grpc_initialize");
    return -1;
  }

  gettimeofday(&tv, NULL);

  srand((int)tv.tv_usec);

  min = atoi(argv[1]);
  max = atoi(argv[2]);
  num_tests = atoi(argv[3]);
  window_size = atoi(argv[4]);

  range = max-min;

  if(range < 0) {
    fprintf(stderr, "Bad min/max values\n");
    die(EXIT_FAILURE);
  }
  else if(range == 0)
    range = 1;

  tasks = (dgemm_task_t *)malloc(window_size * sizeof(dgemm_task_t));

  if(!tasks) {
    fprintf(stderr, "malloc error\n");
    die(EXIT_FAILURE);
  }

  for(i=0;i<window_size;i++) {
    tasks[i].a = (double *)malloc(max*max*sizeof(*(tasks[i].a)));
    tasks[i].b = (double *)malloc(max*max*sizeof(*(tasks[i].b)));
    tasks[i].c = (double *)malloc(max*max*sizeof(*(tasks[i].c)));

    if(!tasks[i].a || !tasks[i].b || !tasks[i].c) {
      fprintf(stderr, "Out of memory.  Try reducing the matrix sizes or use a smaller window size.\n");
      die(EXIT_FAILURE);
    }
  }

  outstanding = 0;
  remaining = num_tests;

  for(i=0;i<window_size;i++) {
    printf("binding handle %d\n", i);
    if(grpc_function_handle_init(&(tasks[i].handle), "bind_servers_at_call_time", "dgemm") != GRPC_NO_ERROR) {
      fprintf(stderr,"Error creating function handle\n");
      die(EXIT_FAILURE);
    }
   
    set_random_fields(&tasks[i], min, range);

    printf("making non-blocking call idx=%d\n", i);
    call_dgemm(&tasks[i]);
    outstanding++;
    remaining--;
  }

  printf("initial window submitted, outstanding=%d, remaining=%d\n", outstanding, remaining);

  while((remaining > 0) || (outstanding > 0)) {
    status = grpc_wait_any(&completed_id);

    if(status != GRPC_NO_ERROR) {
      printf("Warning: one or more calls failed.\n");
      usleep(500000);
    }

    if(status == GRPC_NO_ERROR) {
      outstanding--;
      remaining--;

      printf("task id %d completed\n", completed_id);

      if(remaining >= 0) {
        i = -1;

        for(j=0;j<window_size;j++)
          if(tasks[j].id == completed_id) 
            i = j;
  
        if(i<0) {
          printf("Warning: lost track of id.\n");
          break;
        }

        /* free previous handle */
        grpc_function_handle_destruct(&(tasks[i].handle));

        printf("binding handle %d\n", i);
        if(grpc_function_handle_init(&(tasks[i].handle), "bind_servers_at_call_time", "dgemm") != GRPC_NO_ERROR) {
          fprintf(stderr,"Error creating function handle\n");
          die(EXIT_FAILURE);
        }

        set_random_fields(&tasks[i], min, range);

        printf("making non-blocking call idx=%d\n", i);
        call_dgemm(&tasks[i]);
        outstanding++;
      }
    }

    printf("outstanding=%d, remaining=%d\n", outstanding, remaining);
  }

  die(EXIT_SUCCESS);
}

int
set_trans(int *a)
{
  *a = (rand() % 2) == 0 ? 'N' : 'T';

  return 0;
}

int
set_scale(double *s)
{
  switch(rand() % 3) {
    case 0:
      *s = 0.0;
      break;
    case 1:
      *s = 1.0;
      break;
    case 2:
      *s = 432.3;
      break; 
  }

  return 0;
}

void
set_random_fields(dgemm_task_t *task, int min, int range)
{
  task->m = (rand() % range) + min;
  task->n = (rand() % range) + min;
  task->k = (rand() % range) + min;

  set_trans(&task->transa);
  set_trans(&task->transb);
  set_scale(&task->alpha);
  set_scale(&task->beta);
}

grpc_sessionid_t
call_dgemm(dgemm_task_t *task)
{
  grpc_sessionid_t sessionID; 
  grpc_error_t status;
  int lda, ldb, ldc;

  if(task->transa=='N')
    lda = task->m;
  else
    lda = task->k;

  if(task->transb=='N')
    ldb = task->k;
  else
    ldb = task->n;

  ldc = task->m;

  printf("m=%d, n=%d, k=%d, transa=%c, transb=%c, alpha=%g, beta=%g\n",
     task->m, task->n, task->k, task->transa, task->transb, task->alpha, task->beta);

  status = grpc_call_async(&(task->handle), &sessionID, task->transa, task->transb, task->m,
     task->n, task->k, task->alpha, task->a, lda, task->b, ldb, task->beta, task->c, ldc);

  if(status != GRPC_NO_ERROR) {
    printf("GRPC error status  = %d\n", status);
    grpc_perror("grpc_call_async");
    return -1;
  }

  task->id = sessionID;

  return sessionID;
}
