/**
 * @file
 *
 * This program tests the FILE object type.
 *
 */
/* $Id: filetest.c,v 1.5 2006/02/08 21:53:27 seymour Exp $ */
/* $UTK_Copyright: $ */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "grpc.h"
#include "totaltest.h"

int create_file(char *, char *);
int check_files(char *, char *, char *, char *, char *);
char * get_file_text(char *);

#define NUM_FILES 5

#define INTEXT     "this is the\ninput file.\n"
#define INOUTTEXT  "this is the\ninput/output file.\n"

#define PF_NAME_IN      "packedfile_%d.in"
#define PF_NAME_INOUT   "packedfile_%d.inout"
#define PF_NAME_OUT     "packedfile_%d.out"

#define PF_INTEXT "this is the packed\ninput file number %d.\n"
#define PF_INOUTTEXT "this is the packed\ninput/output file number %d.\n"

int
main(int argc, char *argv[])
{
  grpc_function_handle_t handle;
  grpc_sessionid_t sessionID;
  grpc_error_t status;
  char *infile, *inoutfile, *outfile;
  char **packed_infile, **packed_inoutfile, **packed_outfile;
  int i, async;

  async = strstr(argv[0], "async") != NULL;

  packed_infile = (char **)malloc(NUM_FILES * sizeof(char *));
  packed_inoutfile = (char **)malloc(NUM_FILES * sizeof(char *));
  packed_outfile = (char **)malloc(NUM_FILES * sizeof(char *));

  if(!packed_infile || !packed_inoutfile || !packed_outfile) {
    fprintf(stderr, "Error allocating memory.");
    die(EXIT_FAILURE);
  }

  if(grpc_initialize(NULL) != GRPC_NO_ERROR) {
    grpc_perror("grpc_initialize");
    die(EXIT_FAILURE);
  }

  if(grpc_function_handle_default(&handle, "filetest") != GRPC_NO_ERROR) {
    fprintf(stderr,"Error creating function handle\n");
    die(EXIT_FAILURE);
  }

  infile = "filetest.in";
  inoutfile = "filetest.inout";
  outfile = "filetest.out";

  for(i = 0; i < NUM_FILES; i++) {
    packed_infile[i] = (char *)malloc(strlen(PF_NAME_IN) + 5);
    packed_inoutfile[i] = (char *)malloc(strlen(PF_NAME_INOUT) + 5);
    packed_outfile[i] = (char *)malloc(strlen(PF_NAME_OUT) + 5);

    if(!packed_infile[i] || !packed_inoutfile[i] || !packed_outfile[i]) {
      fprintf(stderr, "Error allocating memory.");
      die(EXIT_FAILURE);
    }

    sprintf(packed_infile[i], PF_NAME_IN, i);
    sprintf(packed_inoutfile[i], PF_NAME_INOUT, i);
    sprintf(packed_outfile[i], PF_NAME_OUT, i);
  }

  if(create_file(infile, INTEXT) < 0 || create_file(inoutfile, INOUTTEXT) < 0) {
    fprintf(stderr,"Error creating files.\n");
    die(EXIT_FAILURE);
  }

  for(i = 0; i < NUM_FILES; i++) {
    char *pfin, *pfinout;

    pfin = malloc(strlen(PF_INTEXT) + 5);
    pfinout = malloc(strlen(PF_INOUTTEXT) + 5);

    if(!pfin || !pfinout) {
      fprintf(stderr, "Error allocating memory.");
      die(EXIT_FAILURE);
    }

    sprintf(pfin, PF_INTEXT, i);
    sprintf(pfinout, PF_INOUTTEXT, i);

    if(create_file(packed_infile[i], pfin) < 0 || 
       create_file(packed_inoutfile[i], pfinout) < 0) 
    {
      fprintf(stderr,"Error creating files.\n");
      die(EXIT_FAILURE);
    }

    free(pfin);
    free(pfinout);
  }

  if(async)
    status = grpc_call_async(&handle, &sessionID, infile, inoutfile, 
       outfile, packed_infile, packed_inoutfile, packed_outfile, NUM_FILES);
  else
    status = grpc_call(&handle, infile, inoutfile, outfile, packed_infile,
       packed_inoutfile, packed_outfile, NUM_FILES);

  if(status != GRPC_NO_ERROR) {
    printf("GRPC error status  = %d\n", status);
    grpc_perror("grpc_call");
    die(status);
  }

  if(async) {
    printf("Waiting for result...\n");
    status = grpc_wait(sessionID);

    if(status != GRPC_NO_ERROR) {
      grpc_perror("grpc_wait");
      die(status);
    }
  }

  if(check_files(infile, inoutfile, outfile, INTEXT, INOUTTEXT) < 0) {
    fprintf(stderr,"File check failed.\n");
    die(EXIT_FAILURE);
  }

  for(i = 0; i < NUM_FILES; i++) {
    char *pfin, *pfinout;

    pfin = malloc(strlen(PF_INTEXT) + 5);
    pfinout = malloc(strlen(PF_INOUTTEXT) + 5);

    if(!pfin || !pfinout) {
      fprintf(stderr, "Error allocating memory.");
      die(EXIT_FAILURE);
    }

    sprintf(pfin, PF_INTEXT, i);
    sprintf(pfinout, PF_INOUTTEXT, i);

    if(check_files(packed_infile[i], packed_inoutfile[i], packed_outfile[i], 
       pfin, pfinout) < 0) 
    {
      fprintf(stderr,"File check failed.\n");
      die(EXIT_FAILURE);
    }

    free(pfin);
    free(pfinout);
  }

  grpc_function_handle_destruct(&handle);

  printf("Finalizing...\n");

  grpc_finalize();

  printf("Done.\n");

  die(EXIT_SUCCESS);
}

int
create_file(char *name, char *text)
{
  int fd, len;

  fd = open(name, O_WRONLY | O_CREAT | O_TRUNC, 0600);
  if(fd < 0) {
    fprintf(stderr,"Could not open file '%s'\n", name);
    return -1;
  }

  len = strlen(text)+1;
  if(write(fd, text, len) != len) {
    fprintf(stderr,"Could not write %d bytes to file '%s'\n", len, name);
    return -1;
  }

  close(fd);

  return 0;
}

int 
check_files(char *infile, char *inoutfile, char *outfile, 
  char *intext_orig, char *inouttext_orig)
{
  char *intext, *inouttext, *outtext, expected;
  int i,failure;

  failure = 0;

  intext = get_file_text(infile);  
  inouttext = get_file_text(inoutfile);  
  outtext = get_file_text(outfile);  

  if(!intext || !inouttext || !outtext) {
    fprintf(stderr, "Failed to get file text.\n");
    return -1;
  }

  expected = 'a';

  if(isalpha(inouttext_orig[0])) {
    if(isupper(inouttext_orig[0]))
      expected = tolower(inouttext_orig[0]);
    else
      expected = toupper(inouttext_orig[0]);
  }

  if(inouttext[0] != expected) {
    fprintf(stderr,"Mismatch in input/output file (%c != %c)\n",
       inouttext[0], expected);
    failure = 1;
  }

  for(i=0;i<(int)strlen(intext_orig);i++)
    if(isalpha(intext_orig[i]))
      if(intext_orig[i] != outtext[i] - 1) {
        fprintf(stderr,"Mismatch in output file (%c != %c)\n",
          intext_orig[i], outtext[i] - 1);
        failure = 1;
        break;
      }

  free(intext);
  free(inouttext);
  free(outtext);

  return failure ? -1 : 0;
}

char *
get_file_text(char *name)
{
  int infd;
  char *file_text;
  struct stat st;

  if(stat(name, &st) < 0) {
    fprintf(stderr,"Could not stat file '%s'\n", name);
    return NULL;
  }

  file_text = (char *)malloc((int) st.st_size);
  if(!file_text) {
    fprintf(stderr,"Could not allocate space to store contents of file '%s'\n", name);
    return NULL;
  }

  infd = open(name, O_RDONLY, 0600);
  if(infd < 0) {
    fprintf(stderr,"Could not open file '%s'\n", name);
    return NULL;
  }

  if(read(infd, file_text, (int)st.st_size) != (int)st.st_size) {
    fprintf(stderr,"Error reading from file '%s'\n", name);
    return NULL;
  }

  return file_text;
}
