/**
 * @file
 *
 * This is an interactive test of the GridRPC API.  Using this
 * program you can test all the features of the standard API.
 * For information on the commands, enter '?' at the prompt.
 */
/* $Id: grpc_test.c,v 1.9 2006/01/30 18:00:13 yarkhan Exp $ */
/* $UTK_Copyright: $ */

#include <stdio.h>
#include <string.h>

#include "portability.h"
#include "grpc.h"

grpc_function_handle_t * fh;

void initialize(), finalize(), handle_init(char *), 
  handle_init_default(char *), handle_destruct(), blocking_call(), 
  nonblocking_call(), probe(int), wait_call(int), sleep_call(int), cancel(int),
  list_requests(), wait_any(), wait_all(), wait_all_list(char *),
  wait_any_list(char*);

int check_args(char *, int);

double gs_result = 0.0;
int sleep_result = 0;

int grpc_requests[100];

const char *delim = " ,\t\n";

#define MAXBUF 80

int
main(int argc, char *argv[])
{
  char buf[MAXBUF];
  int c,i;

  for(i = 0; i < 100; i++)
    grpc_requests[i] = -1;

  for(;;) {
    printf("GRPC test (? for help):");

    if(!fgets(buf, MAXBUF-1, stdin))
      exit(0);

    c = buf[0];

    switch(c) {
      case 'a':
        list_requests();
        break;
      case 'i':
        initialize();
        break;
      case 'f':
        finalize();
        break;
      case 'h':
        if(check_args(buf,1) == 0)
          handle_init(buf+1);
        break;
      case 'd':
        handle_destruct();
        break;
      case 's':
        if(check_args(buf,1) == 0)
          handle_init_default(buf+1);
        break;
      case 'c':
        blocking_call();
        break;
      case 'n':
        nonblocking_call();
        break;
      case 'l':
        if(check_args(buf,1) == 0)
          sleep_call(atoi(buf+1));
        break;
      case 'p':
        if(check_args(buf,1) == 0)
          probe(atoi(buf+1));
        break;
      case 'r':
        if(check_args(buf,1) == 0)
          cancel(atoi(buf+1));
        break;
      case 'w':
        if(check_args(buf,1) == 0)
          wait_call(atoi(buf+1));
        break;
      case 'x':
        wait_all();
        break;
      case 'y':
        wait_any();
        break;
      case 'z':
        if(check_args(buf,-1) == 0)
          wait_all_list(buf+1);
        break;
      case 't':
        if(check_args(buf,-1) == 0)
          wait_any_list(buf+1);
        break;
      case 'q':
        exit(0);
        break;
      case '?':
        printf("Commands:\n");
        printf("\n");
        printf("  i = Initialize GRPC\n");
        printf("  f = Finalize GRPC\n");
        printf("\n");
        printf("  h func_name = Initialize function handle\n");
        printf("  s func_name = Initialize function handle (default host)\n");
        printf("  d = Destruct function handle\n");
        printf("\n");
        printf("  c = Test blocking call.\n");
        printf("  n = Test non-blocking call.\n");
        printf("  l num = Test sleeping non-blocking call, <num> seconds.\n");
        printf("  p num = Probe non-blocking call, sessionid = <num>.\n");
        printf("  r num = Cancel non-blocking call, sessionid = <num>.\n");
        printf("  w num = Wait for results from non-blocking call, sessionid = <num>.\n");
        printf("\n");
        printf("  x = Wait for all non-blocking calls to complete.\n");
        printf("  y = Wait for any non-blocking call to complete.\n");
        printf("  z num num ... num = Wait for all listed non-blocking calls to complete.\n");
        printf("  t num num ... num = Wait for any listed non-blocking call to complete.\n");
        printf("\n");
        printf("  a = List outstanding non-blocking requests.\n");
        printf("\n");
        printf("  q = Quit the test program\n");
        printf("  ? = Help\n");
        break;
      default:
        printf("Unknown command.\n");
        break;
    }
  }
}

/**
 * Checks whether the number of arguments supplied on the command line
 * matches the expected number of arguments.
 *
 * @param cmd -- the command line
 * @param expected -- the number of arguments expected
 *
 * @returns 0 if the number of args matches the expected number,
 *    returns -1 otherwise
 **/

int
check_args(char *cmd, int expected)
{
  char *strtmp, *tok;
  int c = 0;

  strtmp = strdup(cmd+1);

  if(!strtmp)
    return -1;

  tok = strtok(strtmp, delim);

  while(tok) {
    c++;
    tok = strtok(NULL, delim);
  }

  free(strtmp);

  if(expected < 0) {
    if(c < -expected) {
      printf("This command expects at least %d arg(s).\n", -expected);
      return -1;
    }
    return 0;
  }

  if(c != expected) {
    printf("This command expects %d arg(s).\n", expected);
    return -1;
  }

  return 0;
}

/**
 * Prints a list of all pending asynchronous requests to the console.
 **/

void
list_requests() {
  int i;

  for(i = 0; i < 100; i++) {
    if(grpc_requests[i] != -1) {
      grpc_function_handle_t *f;
      grpc_error_t status;

      status = grpc_get_handle(&f, i);

      if(status == GRPC_NO_ERROR)
        printf("  %d: %s\n", i, f->func_name);
      else
        printf("  %d: function name unavailable\n",i);
    }
  }
}

/**
 * Converts a string containing a list of numbers into an array
 * of integers representing those numbers.
 *
 * @param l -- the string containing the list of numbers
 * @param list -- array whose elements will be set upon return
 *
 * @returns the number of elements assigned to the array (i.e. the
 *    number of values parsed from the string)
 **/

int
list2array(char *l, int *list) {
  char *str = strdup(l);
  char *tok;
  int c = 0;

  tok = strtok(str, delim);

  while(tok) {
    list[c++] = atoi(tok);

    tok = strtok(NULL, delim);
  }

  return c;
}

/**
 * Wait for any listed non-blocking call to complete.
 *
 * @param l -- string containing the session IDs to wait for
 **/

void
wait_any_list(char *l) {
  int r, jobno, len, list[100];

  len = list2array(l, list);

  if(len == 0) {
    printf("You must specify session IDs (e.g. 'z 3 6 9').\n");
    return;
  }

  printf("Going to wait for any of the following non-blocking jobs:\n");
  for(r = 0;r < len; r++)
    printf("%d ", list[r]); 
  printf("\n");

  r = grpc_wait_or(list, len, &jobno);

  if(r == GRPC_NO_ERROR) {
    grpc_function_handle_t *h;
    grpc_error_t status;

    status = grpc_get_handle(&h, jobno);

    if((status == GRPC_NO_ERROR) && !strcmp(h->func_name,"ddot"))
      printf("Job %d has completed, result = %g.\n", jobno, gs_result);
    else
      printf("Job %d has completed, result = %d.\n", jobno, sleep_result);

    grpc_requests[jobno] = -1;

  }
  else
    printf("There was an error (wait_all).\n");
}

/**
 * Wait for all listed non-blocking calls to complete.
 * 
 * @param l -- string containing the session IDs to wait for
 **/

void
wait_all_list(char *l) {
  int r, len, list[100];

  len = list2array(l, list);

  if(len == 0) {
    printf("You must specify session IDs (e.g. 'z 3 6 9').\n");
    return;
  }

  printf("Going to wait for all of the following non-blocking jobs:\n");
  for(r = 0;r < len; r++)
    printf("%d ", list[r]); 
  printf("\n");

  r = grpc_wait_and(list, len);

  if(r == GRPC_NO_ERROR) {
    int i;

    printf("All listed jobs have completed.\n");

    for(i = 0; i < len; i++)
      grpc_requests[list[i]] = -1;
  }
  else
    printf("There was an error (wait_all).\n");
}

/**
 * Wait for all non-blocking calls to complete.
 **/

void
wait_all() {
  int r;

  printf("Going to wait for all non-blocking jobs.\n");

  r = grpc_wait_all();

  if(r == GRPC_NO_ERROR) {
    int i;

    printf("All jobs have completed.\n");

    for(i = 0; i < 100; i++)
      grpc_requests[i] = -1;
  }
  else
    printf("There was an error (wait_all).\n");

}

/**
 * Wait for any non-blocking call to complete.
 **/

void
wait_any() {
  int r, jobno;

  printf("Going to wait for any non-blocking job.\n");

  r = grpc_wait_any(&jobno);

  if(r == GRPC_NO_ERROR) {
    grpc_function_handle_t *h;
    grpc_error_t status;

    status = grpc_get_handle(&h, jobno);

    if((status == GRPC_NO_ERROR) && !strcmp(h->func_name,"ddot"))
      printf("Job %d has completed, result = %g.\n", jobno, gs_result);
    else
      printf("Job %d has completed, result = %d.\n", jobno, sleep_result);

    grpc_requests[jobno] = -1;
  }
  else
    printf("There was an error (wait_any).\n");

}

/**
 * Wait for the specified non-blocking call to complete.
 * 
 * @param s -- the session ID to wait for
 **/

void
wait_call(int s) {
  int r;

  if(s < 0) {
    printf("Bad sessionid: %d.\n", s);
    return;
  }

  printf("Going to wait for session %d\n", s);

  r = grpc_wait(s);
 
  if(r == GRPC_NO_ERROR) {
    grpc_function_handle_t *h;
    grpc_error_t status;

    status = grpc_get_handle(&h, s);

    if((status == GRPC_NO_ERROR) && !strcmp(h->func_name,"ddot"))
      printf("The job has completed, result = %g.\n", gs_result);
    else
      printf("The job has completed, result = %d.\n", sleep_result);

    grpc_requests[s] = -1;
  }
  else
    printf("The job has not completed.\n");
}

/**
 * Cancel the specified non-blocking call.
 * 
 * @param s -- the session IDs to cancel
 **/

void
cancel(int s) {
  int r;

  if(s < 0) {
    printf("Bad sessionid: %d.\n", s);
    return;
  }

  printf("Going to cancel session %d\n", s);

  r = grpc_cancel(s);

  if(r == GRPC_NO_ERROR) {
    printf("The job has been cancelled.\n");
    grpc_requests[s] = -1;
  }
  else
    printf("The job could not be cancelled.\n");
}

/**
 * Check whether the specified non-blocking call has completed.
 * 
 * @param s -- the session IDs to probe
 **/

void
probe(int s) {
  int r;

  if(s < 0) {
    printf("Bad sessionid: %d.\n", s);
    return;
  }

  printf("Going to probe session %d\n", s);

  r = grpc_probe(s);
  
  if(r == GRPC_NO_ERROR)
    printf("The job has completed.\n");
  else
    printf("The job has not completed.\n");
}

/**
 * Test sleeping non-blocking call for <len> seconds.
 * global sleep_result will get clobbered on subsequent calls but who cares...
 *
 * @param len -- the number of seconds to sleep
 **/

void
sleep_call(int len) {
  grpc_sessionid_t sessionID;
  grpc_error_t r;

  sleep_result = (len <= 0) ? 20 : len;

  printf("Going to make a non-blocking call (sleeps for %d sec).\n",
     sleep_result);

  fh = (grpc_function_handle_t *) malloc(sizeof(grpc_function_handle_t));

  r = grpc_function_handle_default(fh, "sleeptest");

  if(r == GRPC_NO_ERROR)
    printf("Function handle successfully initialized.\n");
  else
    printf("Error initializing function handle.\n");

  r = grpc_call_async(fh, &sessionID, sleep_result);

  if(r != GRPC_NO_ERROR) {
    printf("Error making non-blocking call.\n");
    return;
  }

  grpc_requests[r] = r;

  printf("Nonblocking call successful, session = %d\n", r);
}

/**
 * Submit a non-blocking call.
 **/

void
nonblocking_call() {
  grpc_sessionid_t sessionID;
  int i, n, incx, incy;
  double *dx, *dy;
  grpc_error_t r;

  gs_result = 0.0;

  printf("Going to make a non-blocking call.\n");

  fh = (grpc_function_handle_t *) malloc(sizeof(grpc_function_handle_t));

  r = grpc_function_handle_default(fh, "ddot");

  if(r == GRPC_NO_ERROR)
    printf("Function handle successfully initialized.\n");
  else
    printf("Error initializing function handle.\n");

  n  = 10;

  dx = (double *)malloc(n * sizeof(double));
  dy = (double *)malloc(n * sizeof(double));

  for(i=0 ; i< n; i++) {
    dx[i] = i + 100.0;
    dy[i] = i + 1000.0;
  }

  incx = incy = 1;

  r = grpc_call_async(fh, &sessionID, n, dx, incx, dy, incy, &gs_result);

  if(r != GRPC_NO_ERROR) {
    printf("Error making non-blocking call.\n");
    return;
  }

  printf("Nonblocking call successful, session = %d\n", r);
}

/**
 * Submit a blocking call.
 **/

void
blocking_call() {
  int i, n, incx, incy;
  double *dx, *dy;
  int r;

  gs_result = 0.0;

  printf("Going to make a blocking call.\n");

  fh = (grpc_function_handle_t *) malloc(sizeof(grpc_function_handle_t));

  r = grpc_function_handle_default(fh, "ddot");

  if(r == GRPC_NO_ERROR)
    printf("Function handle successfully initialized.\n");
  else
    printf("Error initializing function handle.\n");

  n  = 10;

  dx = (double *)malloc(n * sizeof(double));
  dy = (double *)malloc(n * sizeof(double));

  for(i=0 ; i< n; i++) {
    dx[i] = i + 100.0;
    dy[i] = i + 1000.0;
  }
  
  incx = incy = 1;

  r = grpc_call(fh, n, dx, incx, dy, incy, &gs_result);

  if(r != GRPC_NO_ERROR) {
    printf("Error making blocking call.\n");
    return;
  }

  printf("result = %g (%s)\n", gs_result,
    gs_result == 1049785.0 ? "CORRECT" : "INCORRECT");
}

/**
 * Create a function handle for the specified function
 * allowing the agent to assign the server to be used.
 *
 * @param func -- the name of the function to be called
 **/

void
handle_init_default(char *func) {
  char *tok;
  int r;

  if(!func) {
    printf("No function name specified.\n");
    return;
  }

  tok = strtok(func, delim);

  printf("Going to Initialize a function handle (func = %s).\n", tok);

  fh = (grpc_function_handle_t *) malloc(sizeof(grpc_function_handle_t));

  r = grpc_function_handle_default(fh, tok);

  if(r == GRPC_NO_ERROR)
    printf("Function handle successfully initialized.\n");
  else
    printf("Error initializing function handle.\n");
}

/**
 * Create a function handle for the specified function
 * using hard-coded server name: "netsolve.cs.utk.edu".
 *
 * @param func -- the name of the function to be called
 **/

void
handle_init(char *func) {
  char *tok;
  int r;

  if(!func) {
    printf("No function name specified.\n");
    return;
  }

  tok = strtok(func, delim);

  printf("Going to Initialize a function handle (func = %s).\n", tok);

  fh = (grpc_function_handle_t *) malloc(sizeof(grpc_function_handle_t));
  r = grpc_function_handle_init(fh, "netsolve.cs.utk.edu", tok);
  
  if(r == GRPC_NO_ERROR)
    printf("Function handle successfully initialized.\n");
  else
    printf("Error initializing function handle.\n");
}

/**
 * Destroy previously created function handle (global var fh).
 **/

void
handle_destruct() {
  int r;

  printf("Going to destruct a function handle.\n");

  if(!fh) {
    printf("Function handle not initialized.\n");
    return;
  }

  r = grpc_function_handle_destruct(fh);

  if(r == GRPC_NO_ERROR)
    printf("Function handle successfully destructed.\n");
  else
    printf("Error destructing function handle.\n");

  fh = NULL;
}

/**
 * Initialize the GridRPC system.  This should be done before
 * any other GridRPC calls are made.
 **/

void
initialize() {
  int r,i;

  printf("Going to Initialize GRPC.\n");

  r = grpc_initialize("ignored_filename");

  for(i = 0; i < 100; i++)
    grpc_requests[i] = -1;

  if(r == GRPC_NO_ERROR)
    printf("GRPC successfully initialized.\n");
  else
    printf("Error initializing GRPC.\n");
}

/**
 * Finalize GridRPC, free resources, etc.
 **/

void
finalize() {
  int r,i;

  printf("Going to finalize GRPC.\n");

  r = grpc_finalize();

  for(i = 0; i < 100; i++)
    grpc_requests[i] = -1;

  if(r == GRPC_NO_ERROR)
    printf("GRPC successfully finalized.\n");
  else
    printf("Error finalizing GRPC.\n");
}
