/**
 * @file
 *
 * Test of the "return_double_scalar" problem.  This submits
 * a number of non-blocking requests specified on the command line:
 *    # return_test 5
 */
/* $Id: return_test.c,v 1.10 2006/02/08 21:53:27 seymour Exp $ */
/* $UTK_Copyright: $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include<stdio.h>
#include<stdlib.h>
#ifdef HAVE_VALUES_H
#include<values.h>
#else
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_FLOAT_H
#include <float.h>
#define MAXDOUBLE DBL_MAX
#endif
#endif

#include "portability.h"
#include "grpc.h"
#include "totaltest.h"

void test_all_return_types(void);
int test_scalar_return_types(void);
int test_vector_return_types(void);
int test_matrix_return_types(void);

int 
main(int argc, char** argv)
{
  grpc_function_handle_t handle;
  grpc_sessionid_t *sessionId;
  int i, num_req, failures;
  grpc_error_t status;
  double d, *foo;

  if(argc != 2) {
    fprintf(stderr,"Usage: %s [num]\n", argv[0]);
    die(EXIT_FAILURE);
  }

  if(grpc_initialize(NULL) != GRPC_NO_ERROR) {
    grpc_perror("grpc_initialize");
    die(EXIT_FAILURE);
  }

  if(grpc_function_handle_default(&handle, "return_double_scalar") != GRPC_NO_ERROR) {
    fprintf(stderr,"Error creating function handle\n");
    die(EXIT_FAILURE);
  }

  num_req = atoi(argv[1]);

  d = 666.0;

  sessionId = (grpc_sessionid_t *)malloc(num_req * sizeof(grpc_sessionid_t));
  foo = (double *)malloc(num_req * sizeof(double));

  for(i=0;i<num_req;i++) {
    foo[i] = 0.0;
    status = grpc_call_async(&handle, &(sessionId[i]), d, &(foo[i]));
    printf("Submitted request %d\n", sessionId[i]);

    if(status != GRPC_NO_ERROR) {
      grpc_perror("grpc_call_async");
      die(EXIT_FAILURE);
    }
  }

  failures=0;

  for(i=0;i<num_req;i++) {
    grpc_error_t reqstatus;

    printf("Probing for request %d\n", sessionId[i]);

    do {
      reqstatus = grpc_probe(sessionId[i]);
      if(reqstatus == GRPC_NO_ERROR)
        break;
      printf(".");
      sleep(1);
    } while(reqstatus == GRPC_NOT_COMPLETED);

    printf("\n");

    if(reqstatus != GRPC_NO_ERROR) {
      printf("grpc_probe failed\n");
      die(EXIT_FAILURE);
    }

    reqstatus = grpc_wait(sessionId[i]);

    fflush(stdout);
    if(reqstatus != GRPC_NO_ERROR) {
      grpc_perror("grpc_wait");
      die(EXIT_FAILURE);
    }

    if(foo[i] != d+1.0)
      failures++;
  }

  if(failures > 0) {
    printf("%d of the requests returned the wrong value\n", failures);
    die(EXIT_FAILURE);
  }

  if(grpc_function_handle_destruct(&handle) != GRPC_NO_ERROR)
    fprintf(stderr,"Warning: failed to destroy function handle\n");

  test_all_return_types();

  die(EXIT_SUCCESS);
}

void
get_handle(grpc_function_handle_t *handle, char *s)
{
  printf("Testing %s\n",s);
  if(grpc_function_handle_default(handle, s) != GRPC_NO_ERROR) {
    fprintf(stderr,"Error creating function handle\n");
    die(EXIT_FAILURE);
  }
}

int
check_and_wait(grpc_function_handle_t *handle, grpc_error_t status, 
  grpc_sessionid_t sessionId)
{
  if(status != GRPC_NO_ERROR) {
    printf("return_int_scalar failed.\n");
    return 1;
  }
  else {
    status = grpc_wait(sessionId);

    if(status != GRPC_NO_ERROR) {
      printf("grpc_wait failed.\n");
      return 1;
    }
  }

  if(grpc_function_handle_destruct(handle) != GRPC_NO_ERROR)
    fprintf(stderr,"Warning: failed to destroy function handle\n");

  return 0;
}

void
test_all_return_types(void)
{
  int num_failures = 0;

  num_failures += test_scalar_return_types();
  num_failures += test_vector_return_types();
  num_failures += test_matrix_return_types();

  if(num_failures > 0) {
    fprintf(stderr, "There were %d failures in the return tests\n", 
       num_failures);
    die(EXIT_FAILURE);
  }
}

int
test_matrix_return_types(void)
{   
  grpc_function_handle_t handle;
  grpc_sessionid_t sessionId;
  grpc_error_t status;
  int i, num_failures;

#define MDIM 2
  
  int int_mat[MDIM*MDIM] = {9, 8, 7, 6}, int_mat_rv[MDIM*MDIM];
  float float_mat[MDIM*MDIM] = {6.0, 7.0, 8.0, 9.0}, float_mat_rv[MDIM*MDIM];
  double double_mat[MDIM*MDIM] = {3.0, 4.0, 5.0, 6.0}, double_mat_rv[MDIM*MDIM];
  char char_mat[MDIM*MDIM] = {'a', 's', 'd', 'f'}, char_mat_rv[MDIM*MDIM];
  
  num_failures = 0;
  
  get_handle(&handle, "return_int_matrix");
  status = grpc_call_async(&handle, &sessionId, int_mat, MDIM, MDIM, 
      &int_mat_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  for(i=0;i<MDIM*MDIM;i++) { 
    if(GS_NE(int_mat[i] + 1, int_mat_rv[i])) {
      printf("return_int_matrix value wrong (%d, %d)\n",
         int_mat[i] + 1, int_mat_rv[i]);
      num_failures++;
      break;
    }
  }

  get_handle(&handle, "return_float_matrix");
  status = grpc_call_async(&handle, &sessionId, float_mat, MDIM, MDIM,
      &float_mat_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  for(i=0;i<MDIM*MDIM;i++) {
    if(GS_NE(float_mat[i] + 1.0, float_mat_rv[i])) {
      printf("return_float_matrix value wrong (%f, %f)\n",
         float_mat[i] + 1.0, float_mat_rv[i]);
      num_failures++;
      break;
    }
  }

  get_handle(&handle, "return_double_matrix");
  status = grpc_call_async(&handle, &sessionId, double_mat, MDIM, MDIM,
      &double_mat_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  for(i=0;i<MDIM*MDIM;i++) {
    if(GS_NE(double_mat[i] + 1.0, double_mat_rv[i])) {
      printf("return_double_matrix value wrong (%g, %g)\n",
         double_mat[i] + 1.0, double_mat_rv[i]);
      num_failures++;
      break;
    }
  }

  get_handle(&handle, "return_char_matrix");
  status = grpc_call_async(&handle, &sessionId, char_mat, MDIM, MDIM,
      &char_mat_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  for(i=0;i<MDIM*MDIM;i++) {
    if(GS_NE(char_mat[i] + 1, char_mat_rv[i])) {
      printf("return_char_matrix value wrong (%c, %c)\n",
         char_mat[i] + 1, char_mat_rv[i]);
      num_failures++;
      break;
    }
  }

  return num_failures;
}

int
test_vector_return_types(void)
{   
  grpc_function_handle_t handle;
  grpc_sessionid_t sessionId;
  grpc_error_t status;
  int i, num_failures;

#define VLEN 4

  int int_vec[VLEN] = {9, 8, 7, 6}, int_vec_rv[VLEN];
  float float_vec[VLEN] = {6.0, 7.0, 8.0, 9.0}, float_vec_rv[VLEN];
  double double_vec[VLEN] = {3.0, 4.0, 5.0, 6.0}, double_vec_rv[VLEN];
  char char_vec[VLEN] = {'a', 's', 'd', 'f'}, char_vec_rv[VLEN];

  num_failures = 0;

  get_handle(&handle, "return_int_vector");
  status = grpc_call_async(&handle, &sessionId, int_vec, VLEN, &int_vec_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  for(i=0;i<VLEN;i++) {
    if(GS_NE(int_vec[i] + 1, int_vec_rv[i])) {
      printf("return_int_vector value wrong (%d, %d)\n",
         int_vec[i] + 1, int_vec_rv[i]);
      num_failures++;
      break;
    }
  }

  get_handle(&handle, "return_float_vector");
  status = grpc_call_async(&handle, &sessionId, float_vec, VLEN, &float_vec_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  for(i=0;i<VLEN;i++) {
    if(GS_NE(float_vec[i] + 1.0, float_vec_rv[i])) {
      printf("return_float_vector value wrong (%f, %f)\n",
         float_vec[i] + 1.0, float_vec_rv[i]);
      num_failures++;
      break;
    }
  }

  get_handle(&handle, "return_double_vector");
  status = grpc_call_async(&handle, &sessionId, double_vec, VLEN, &double_vec_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  for(i=0;i<VLEN;i++) {
    if(GS_NE(double_vec[i] + 1.0, double_vec_rv[i])) {
      printf("return_double_vector value wrong (%g, %g)\n",
         double_vec[i] + 1.0, double_vec_rv[i]);
      num_failures++;
      break;
    }
  }

  get_handle(&handle, "return_char_vector");
  status = grpc_call_async(&handle, &sessionId, char_vec, VLEN, &char_vec_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  for(i=0;i<VLEN;i++) {
    if(GS_NE(char_vec[i] + 1, char_vec_rv[i])) {
      printf("return_char_vector value wrong (%c, %c)\n",
         char_vec[i] + 1, char_vec_rv[i]);
      num_failures++;
      break;
    }
  }

  return num_failures;
}

int
test_scalar_return_types(void)
{
  grpc_function_handle_t handle;
  grpc_sessionid_t sessionId;
  grpc_error_t status;
  int num_failures;

  int int_scalar = 666, int_scalar_rv;
  float float_scalar = 123.0, float_scalar_rv;
  double double_scalar = 456.0, double_scalar_rv;
  char char_scalar = 'K', char_scalar_rv;

  num_failures = 0;

  get_handle(&handle, "return_int_scalar");
  status = grpc_call_async(&handle, &sessionId, int_scalar, &int_scalar_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  if(GS_NE(int_scalar + 1, int_scalar_rv)) {
    printf("return_int_scalar value wrong (%d, %d)\n", 
       int_scalar + 1, int_scalar_rv);
    num_failures++;
  }

  get_handle(&handle, "return_float_scalar");
  status = grpc_call_async(&handle, &sessionId, float_scalar, &float_scalar_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  if(GS_NE(float_scalar + 1.0, float_scalar_rv)) {
    printf("return_float_scalar value wrong (%f, %f)\n",
       float_scalar + 1.0, float_scalar_rv);
    num_failures++;
  }

  get_handle(&handle, "return_double_scalar");
  status = grpc_call_async(&handle, &sessionId, double_scalar, &double_scalar_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  if(GS_NE(double_scalar + 1.0, double_scalar_rv)) {
    printf("return_double_scalar value wrong (%g, %g)\n",
       double_scalar + 1.0, double_scalar_rv);
    num_failures++;
  }

  get_handle(&handle, "return_char_scalar");
  status = grpc_call_async(&handle, &sessionId, char_scalar, &char_scalar_rv);
  num_failures += check_and_wait(&handle, status, sessionId);
  if(GS_NE(char_scalar + 1, char_scalar_rv)) {
    printf("return_int_scalar value wrong (%c, %c)\n",
       char_scalar + 1, char_scalar_rv);
    num_failures++;
  }

  return num_failures;

}
