/**
 * @file
 *
 * This program performs ping-pong bandwidth tests using the GridRPC API.
 * The program is invoked as follows:
 * 
 *   rtt_test <lo> <hi> <inc> <reps>
 *
 * Transmits arrays ranging in size from <lo> to <hi>
 * in increments of <inc>.  The number of repetitions
 * each size is specified by <reps>
 *
 * The output consists of two columns of numbers.  The first number is
 * the size in bytes of the data sent.  The second number is the wall time
 * elapsed from the time the data was sent until the same size data was
 * received back from the server.
 *
 */
/* $Id: rtt_test.c,v 1.13 2006/02/21 22:02:23 seymour Exp $ */
/* $UTK_Copyright: $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_VALUES_H
#include<values.h>
#else  /* HAVE_VALUES_H */
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif /* HAVE_LIMITS_H */
#ifdef HAVE_FLOAT_H
#include <float.h>
#define MAXDOUBLE DBL_MAX
#endif /* HAVE_FLOAT_H */
#endif /* HAVE_VALUES_H */

#ifndef MAXDOUBLE
#define MAXDOUBLE 1.7976931348623157e+308
#endif

#include "portability.h"	/* for <sys/time.h> */
#include "grpc.h"
#include "totaltest.h"

double elapsed_time(const struct timeval*, const struct timeval*);

int
main(int argc, char *argv[])
{
  int i, j, lo, hi, incr, rep, *x, status, prof_enabled;
  grpc_function_handle_t handle;
  struct timeval t1, t2;
  double et, min_time;
  grpc_profile_t *gsprof, *min_prof;

  if(argc != 5) {
    fprintf(stderr,"\nUsage: %s <lo> <hi> <inc> <reps>\n", argv[0]);
    fprintf(stderr,"\n  Transmits arrays ranging in size from <lo>..<hi>\n");
    fprintf(stderr,"  in increments of <inc>.  The number of repetitions\n");
    fprintf(stderr,"  at each size is specified by <reps>.\n");
    die(EXIT_FAILURE);
  }

  lo = atoi(argv[1]);
  hi = atoi(argv[2]);
  incr = atoi(argv[3]);
  rep = atoi(argv[4]);
  prof_enabled = 0;

  x = (int*)malloc(hi*sizeof(x[0]));
  gsprof = (grpc_profile_t *)calloc(rep, sizeof(grpc_profile_t));

  if(!x || !gsprof) {
    grpc_perror("malloc");
    die(EXIT_FAILURE);
  }

  if(grpc_initialize(NULL) != GRPC_NO_ERROR) {
    grpc_perror("grpc_initialize");
    die(EXIT_FAILURE);
  }

  if(grpc_function_handle_default(&handle, "vpass_int") != GRPC_NO_ERROR) {
    fprintf(stderr,"Error creating function handle\n");
    die(EXIT_FAILURE);
  }

  for(i=0;i<hi;i++)
    x[i] = i;

  for(i=lo;i<=hi;i+=incr) {
    min_time = MAXDOUBLE;
    min_prof = gsprof;

    for(j=0;j<rep;j++) {
 
      prof_enabled = grpc_profile(&(gsprof[j])) == GRPC_NO_ERROR;

      gettimeofday(&t1, 0);
      status = grpc_call(&handle, x, i);
      gettimeofday(&t2, 0);

      if(status != GRPC_NO_ERROR) {
        printf("GRPC error status  = %d\n", status);
        grpc_perror("grpc_call");
        die(status);
      }

      et = elapsed_time(&t1,&t2);
      
      if(et < min_time) {
        min_time = et;
        min_prof = gsprof;
      }
    }

    if(prof_enabled)
      printf("%d: %g %g %g %g %g %g %g\n",(int)(i*sizeof(x[0])), min_time,
        min_prof->proxy_start, min_prof->object_init, min_prof->agent_comm,
        min_prof->send_input, min_prof->job_complete, min_prof->recv_output);
    else
      printf("%d: %g\n",(int)(i*sizeof(x[0])),min_time);
  }

  grpc_function_handle_destruct(&handle);

  printf("Finalizing...\n");

  grpc_finalize();

  printf("Done.\n");

  die(EXIT_SUCCESS);
}

/**
 * Computes the elapsed wall time between the two given 
 * times (timeval structs).
 *
 * @param t0 -- starting time
 * @param t1 -- ending time
 *
 * @returns the elapsed time between the two times
 **/

double
elapsed_time(const struct timeval* t0, const struct timeval* t1)
{
  double ret = t1->tv_sec - t0->tv_sec;
  double uret = t1->tv_usec - t0->tv_usec;
  return ret + uret / 1E6;
}
