/******************************************************************************/
/*                              strassen_dgemm.c                              */
/*  The implementation of the Strassen's algorithm for matrix multiplication  */
/*             This is used to test GridSolve request sequencing              */
/*                            Yinan Li, 05/17/2007                            */
/******************************************************************************/


#include <time.h>
#include <math.h>
#include <float.h>
#include "gs_sequence.h"
/* #include "c_timer.h" */
/* #include "cblas.h" */


/* get a square submatrix of size 'subn' 
   from the original matrix of size 'n' */
int get_submatrix(const double *m, double *subm, int n, int subn) {
    int i, j;
    
    for (i = 0; i < subn; i++) {
        for (j = 0; j < subn; j++) {
            subm[i * subn + j] = m[i * n + j];
        }
    } 
    
    return 0;  
}


/* copy submatrices back after computation */
void assemble_matrix(const double *c11, const double *c12, 
const double *c21, const double *c22, double *c, int n) {
	int i, j;
	int m;
	
	m = n / 2;

	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			if (i < m && j < m) {
				c[i * n + j] = c11[i * m + j];
			}
			else if (i < m && j >= m) {
				c[i * n + j] = c12[i * m + j - m];
			}
			else if (i >= m && j < m) {
				c[i * n + j] = c21[(i - m) * m + j];
			}
			else if (i >= m && j >= m) {
				c[i * n + j] = c22[(i - m) * m + j - m];
			}
		}
	}
}


/* initialize all elements in a matrix to 0.0 */
void init_matrix(double *mat, int n) {
	int i, j;
	
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			mat[i * n + j] = 0.000;
		}
	}
}


/* print the matrix */
void print_matrix(double *mat, int n) {
	int i, j;

	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			printf("%.3f ", mat[i * n + j]);
		}
		printf("\n");
	}
}


/* the Strassen's algorithm for matrix multiplication */
double strassen_dgemm(const double *a, const double *b, double *c, int n) {
	double *a11, *a12, *a21, *a22, *b11, *b12, *b21, *b22;
    double *t[10];
    double *q[7];
	double *s[4];
    double *c11, *c12, *c21, *c22;
    int m;
    double alpha, beta;
    int i;
    double et;
	
	
    m = n / 2;
     
	a11 = (double *) malloc(sizeof(double) * m * m);
	a12 = (double *) malloc(sizeof(double) * m * m);
	a21 = (double *) malloc(sizeof(double) * m * m);
	a22 = (double *) malloc(sizeof(double) * m * m);
	b11 = (double *) malloc(sizeof(double) * m * m);
	b12 = (double *) malloc(sizeof(double) * m * m);
	b21 = (double *) malloc(sizeof(double) * m * m);
	b22 = (double *) malloc(sizeof(double) * m * m);
	c11 = (double *) malloc(sizeof(double) * m * m);
	c12 = (double *) malloc(sizeof(double) * m * m);
	c21 = (double *) malloc(sizeof(double) * m * m);
	c22 = (double *) malloc(sizeof(double) * m * m);
	
    /* get four submatrices of a */
    get_submatrix(&a[0], a11, n, m);
    get_submatrix(&a[m], a12, n, m);
    get_submatrix(&a[n * m], a21, n, m);
    get_submatrix(&a[n * m + m], a22, n, m);
    
    /* get four submatrices of b */
    get_submatrix(&b[0], b11, n, m);
    get_submatrix(&b[m], b12, n, m);
    get_submatrix(&b[n * m], b21, n, m);
    get_submatrix(&b[n * m + m], b22, n, m);  
    
	/* get four submatrices of c */
    /*
	get_matrix(&c[0], c11, n, m);
	get_matrix(&c[m], c12, n, m);
	get_matrix(&c[n * m], c21, n, m);
	get_matrix(&c[n * m + m], c22, n, m);
	*/

    /* allocate space for those temporary matrices */
    for (i = 0; i < 10; i++)
        t[i] = (double *) malloc(sizeof(double) * m * m);
    
    for (i = 0; i < 7; i++) 
        q[i] = (double *) malloc(sizeof(double) * m * m);
        
	for (i = 0; i < 4; i++) 
		s[i] = (double *) malloc(sizeof(double) * m * m);
	
    alpha = 1.000;
    beta = 0.000;
	
    /* the GridSolve sequence */
    grpc_sequence_begin(OPTIMISTIC_MODE);
    
    /* the first level */
    grpc_submit("matadd", a11, a22, t[0], m);
    grpc_submit("matadd", a21, a22, t[1], m);
    grpc_submit("matadd", a11, a12, t[2], m);
    grpc_submit("matsub", a21, a11, t[3], m);
    grpc_submit("matsub", a12, a22, t[4], m);
    grpc_submit("matadd", b11, b22, t[5], m);
    grpc_submit("matsub", b12, b22, t[6], m);
    grpc_submit("matsub", b21, b11, t[7], m);
    grpc_submit("matadd", b11, b12, t[8], m);
    grpc_submit("matadd", b21, b22, t[9], m);
    
    /* the second level */  
	/*
	grpc_submit("matmul", t[0], t[5], q[0], m);
	grpc_submit("matmul", t[1], b11, q[1], m);
	grpc_submit("matmul", a11, t[6], q[2], m);
	grpc_submit("matmul", a22, t[7], q[3], m);
	grpc_submit("matmul", t[2], b22, q[4], m);
	grpc_submit("matmul", t[3], t[8], q[5], m);
	grpc_submit("matmul", t[4], t[9], q[6], m);
	*/
	
    grpc_submit("dgemm", 'N', 'N', m, m, m, alpha, t[0], m, t[5], m, beta, q[0], m);
    grpc_submit("dgemm", 'N', 'N', m, m, m, alpha, t[1], m, b11, m, beta, q[1], m);
    grpc_submit("dgemm", 'N', 'N', m, m, m, alpha, a11, m, t[6], m, beta, q[2], m);
    grpc_submit("dgemm", 'N', 'N', m, m, m, alpha, a22, m, t[7], m, beta, q[3], m);
    grpc_submit("dgemm", 'N', 'N', m, m, m, alpha, t[2], m, b22, m, beta, q[4], m);
    grpc_submit("dgemm", 'N', 'N', m, m, m, alpha, t[3], m, t[8], m, beta, q[5], m);
    grpc_submit("dgemm", 'N', 'N', m, m, m, alpha, t[4], m, t[9], m, beta, q[6], m);

    /* the third level */
    grpc_submit("matadd", q[0], q[3], s[0], m);
    grpc_submit("matsub", q[4], q[6], s[1], m);
    grpc_submit("matadd", q[2], q[0], s[2], m);
    grpc_submit("matsub", q[1], q[5], s[3], m);
    
    /* the forth level */
    grpc_submit("matsub", s[0], s[1], c11, m);
    grpc_submit("matadd", q[2], q[4], c12, m);
    grpc_submit("matadd", q[1], q[3], c21, m);
    grpc_submit("matsub", s[2], s[3], c22, m);

    /*  et = grpc_sequence_end(21, t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8],
            t[9], q[0], q[1], q[2], q[3], q[4], q[5], q[6], s[0], s[1], s[2], s[3]); 
     */
	
	et = grpc_sequence_end(0);
	/* end sequencing */

	assemble_matrix(c11, c12, c21, c22, c, n);

	/* release memory */
	free(a11);
	free(a12);
	free(a21);
	free(a22);
	free(b11);
	free(b12);
	free(b21);
	free(b22);
	free(c11);
	free(c12);
	free(c21);
	free(c22);

    for (i = 0; i < 10; i++) 
        free(t[i]);

    for (i = 0; i < 7; i++)
        free(q[i]);

	for (i = 0; i < 4; i++)
		free(s[i]);
	           
    return et;    
}


/* cblas matrix multiplication */
/*
double std_dgemm(double *a, double *b, double *c, int n) {
    double alpha, beta;
    double bt, et;


    alpha = 1.000;
	beta = 0.000;
    
	bt = walltime();

    cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, 
	n, n, n, alpha, a, n, b, n, beta, c, n);

	et = walltime();

	return et - bt;
}
*/


/* verify that the computation result is correct */
double verify_result(const double *c, const double *c_cblas, int n) {
    double diff, sum_diff, sum;
    int i, j;
    
    
	diff = 0.000;
    sum_diff = 0.000;
    sum = 0.000;
    for (i = 0; i < n; i++) {
        for (j = 0; j < n; j++) {
            sum += c[i * n + j] * c[i * n + j];
            diff = c[i * n + j] - c_cblas[i * n + j];
            sum_diff += diff * diff;
        }
    }

    return sqrt(sum_diff) / (sqrt(sum) * DBL_EPSILON);
}


int main(int argc, char *argv[]) {
    int i, n;
    double *a, *b, *c;
    grpc_error_t status;
    double min = -5.0, max = 5.0;
    double time_strassen;
    
	
	if (argc != 2) {
		fprintf(stderr, "Usage: strassen_dgemm <dimension size>\n");
		exit(1);
	}

    n = atoi(argv[1]);
    
    /* allocate spaces for the matrices */
    a = (double *) malloc(sizeof(double) * n * n);
    b = (double *) malloc(sizeof(double) * n * n);
    c = (double *) malloc(sizeof(double) * n * n);  
    /* c_cblas = (double *) malloc(sizeof(double) * n * n);   */
    if (!a || !b || !c) {
        perror("malloc");
        exit(1);
    }
    
    srand(time(0));
    /* randomly initialize the matrices */
    for (i = 0; i < n * n; i++) {
        a[i] = min + (double) rand() * (max - min) / (double) RAND_MAX;
        b[i] = min + (double) rand() * (max - min) / (double) RAND_MAX;
        c[i] = 0.000;
    }

    /* initialize the GridRPC system */
    status = grpc_initialize(NULL);
    if (status != GRPC_NO_ERROR && status != GRPC_ALREADY_INITIALIZED) {
        fprintf(stderr, "%s\n", grpc_error_string(status));
        exit(1);
    }
	
    /* call the Strassen's algorithm */
    time_strassen = strassen_dgemm(a, b, c, n);
	/* time_std = std_dgemm(a, b, c_cblas, n); */
	
	printf("%d\t%f\n", n, time_strassen);

    /* verify the result is correctly computed */
    /* val = verify_result(c, c_cblas, n); */
    /* printf("Result verification: %.16f\n", val); */
        
    /* finalize and done */
    grpc_finalize();    
    
    return 0;
}
