/**
 * @file
 *
 * This file contains common code used by the variations of
 * 'totaltest'.  Primarily this includes routines to print
 *  the results.
 *
 */
/* $Id: totaltest_common.c,v 1.7 2005/11/10 20:46:06 seymour Exp $ */
/* $UTK_Copyright: $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "grpc.h"
#include "totaltest.h"

/**
 * Parses the command line for all totaltest related test programs.
 * The command line should consist of:
 *   argv[0]: the executable name, as usual
 *   argv[1]: optionally contains the test to be run (ctotal or ftotal)
 *
 * @param argc -- number of args
 * @param argv -- array of strings containing the arguments
 *
 * @returns the name of the test to be run.  If not specified on the
 *   command line, returns "ctotal".
 **/

char *
parse_totaltest_cmdline(int argc, char *argv[])
{
  char *test_name;

  if(argc == 2) {
    if(!strcmp(argv[1], "ctotal"))
      test_name = argv[1];
    else if(!strcmp(argv[1], "ftotal"))
      test_name = argv[1];
    else {
      fprintf(stderr,"Test name must be ctotal or ftotal.\n");
      die(EXIT_FAILURE);
    }
  }
  else
    test_name = "ctotal";

  return test_name;
}

/**
 * Prints the given integer matrix to the console.
 *
 * @param str -- string to be printed before the matrix
 * @param mat -- the matrix to be printed
 * @param rows -- the number of rows in the matrix
 * @param cols -- the number of columns in the matrix
 *
 * @returns 0 on success, -1 on failure.
 **/

int
print_imat(char *str, int *mat, int rows, int cols)
{
  int i,j;

  printf("%s:\n",str);
  for(i=0;i<rows;i++) {
    for(j=0;j<cols;j++)
      printf("%d ", mat[i*cols+j]);
    printf("\n");
  }

  return 0;
}

/**
 * Prints the given single precision matrix to the console.
 *
 * @param str -- string to be printed before the matrix
 * @param mat -- the matrix to be printed
 * @param rows -- the number of rows in the matrix
 * @param cols -- the number of columns in the matrix
 *
 * @returns 0 on success, -1 on failure.
 **/

int
print_fmat(char *str, float *mat, int rows, int cols) {
  int i,j;

  printf("%s:\n",str);
  for(i=0;i<rows;i++) {
    for(j=0;j<cols;j++)
      printf("%g ", mat[i*cols+j]);
    printf("\n");
  }

  return 0;
} 

/**
 * Prints the given double precision matrix to the console.
 *
 * @param str -- string to be printed before the matrix
 * @param mat -- the matrix to be printed
 * @param rows -- the number of rows in the matrix
 * @param cols -- the number of columns in the matrix
 *
 * @returns 0 on success, -1 on failure.
 **/

int
print_dmat(char *str, double *mat, int rows, int cols) {
  int i,j;

  printf("%s:\n",str);
  for(i=0;i<rows;i++) {
    for(j=0;j<cols;j++)
      printf("%g ", mat[i*cols+j]);
    printf("\n");
  }

  return 0;
} 

/**
 * Prints the given single precision complex matrix to the console.
 *
 * @param str -- string to be printed before the matrix
 * @param mat -- the matrix to be printed 
 * @param rows -- the number of rows in the matrix
 * @param cols -- the number of columns in the matrix
 *
 * @returns 0 on success, -1 on failure.
 **/
    
int   
print_scmp_mat(char *str, gs_scomplex *mat, int rows, int cols) {
  int i,j;

  printf("%s:\n",str);
  for(i=0;i<rows;i++) {
    for(j=0;j<cols;j++)
      printf("%g+%gi ", mat[i*cols+j].r, mat[i*cols+j].i);
    printf("\n");
  }
 
  return 0;
}  

/**
 * Prints the given double precision complex matrix to the console.
 * 
 * @param str -- string to be printed before the matrix
 * @param mat -- the matrix to be printed 
 * @param rows -- the number of rows in the matrix
 * @param cols -- the number of columns in the matrix
 *
 * @returns 0 on success, -1 on failure.
 **/

int
print_dcmp_mat(char *str, gs_dcomplex *mat, int rows, int cols) {
  int i,j;

  printf("%s:\n",str);
  for(i=0;i<rows;i++) {
    for(j=0;j<cols;j++)
      printf("%g+%gi ", mat[i*cols+j].r, mat[i*cols+j].i);
    printf("\n");
  }

  return 0;
}

/**
 * Prints the given character matrix to the console.
 *
 * @param str -- string to be printed before the matrix
 * @param mat -- the matrix to be printed
 * @param rows -- the number of rows in the matrix
 * @param cols -- the number of columns in the matrix
 *
 * @returns 0 on success, -1 on failure.
 **/

int
print_cmat(char *str, char *mat, int rows, int cols) {
  int i,j;

  printf("%s:\n",str);
  for(i=0;i<rows;i++) {
    for(j=0;j<cols;j++)
      printf("%c ", mat[i*cols+j]);
    printf("\n");
  }

  return 0;
} 

/**
 * Prints the given single precision vector to the console.
 *
 * @param str -- string to be printed before the vector
 * @param vec -- the vector to be printed
 * @param len -- the number of elements in the vector
 *
 * @returns 0 on success, -1 on failure.
 **/

int
print_fvec(char *str, float *vec, int len)
{
  int i;

  printf("%s = ",str);
  for(i=0;i<len;i++)
    printf("%g ", vec[i]);
  printf("\n");

  return 0;
}

/**
 * Prints the given integer vector to the console.
 *
 * @param str -- string to be printed before the vector
 * @param vec -- the vector to be printed
 * @param len -- the number of elements in the vector
 *
 * @returns 0 on success, -1 on failure.
 **/

int
print_ivec(char *str, int *vec, int len)
{
  int i;

  printf("%s = ",str);
  for(i=0;i<len;i++)
    printf("%d ", vec[i]);
  printf("\n");

  return 0;
}

/**
 * Prints the given double precision vector to the console.
 *
 * @param str -- string to be printed before the vector
 * @param vec -- the vector to be printed
 * @param len -- the number of elements in the vector
 *
 * @returns 0 on success, -1 on failure.
 **/

int
print_dvec(char *str, double *vec, int len)
{
  int i;

  printf("%s = ",str);
  for(i=0;i<len;i++)
    printf("%g ", vec[i]);
  printf("\n");

  return 0;
}

/**
 * Prints the given single precision complex vector to the console.
 *
 * @param str -- string to be printed before the vector
 * @param vec -- the vector to be printed
 * @param len -- the number of elements in the vector
 *
 * @returns 0 on success, -1 on failure.
 **/

int
print_scmp_vec(char *str, gs_scomplex *vec, int len)
{
  int i;
  
  printf("%s = ",str);
  for(i=0;i<len;i++)
    printf("%g+%gi ", vec[i].r, vec[i].i);
  printf("\n");
  
  return 0;
}

/**
 * Prints the given double precision complex vector to the console.
 * 
 * @param str -- string to be printed before the vector
 * @param vec -- the vector to be printed
 * @param len -- the number of elements in the vector
 * 
 * @returns 0 on success, -1 on failure.
 **/
 
int
print_dcmp_vec(char *str, gs_dcomplex *vec, int len)
{
  int i;
  
  printf("%s = ",str);
  for(i=0;i<len;i++)
    printf("%g+%gi ", vec[i].r, vec[i].i);
  printf("\n");
  
  return 0;
}

/**
 * Prints the given character vector to the console.
 *
 * @param str -- string to be printed before the vector
 * @param vec -- the vector to be printed
 * @param len -- the number of elements in the vector
 *
 * @returns 0 on success, -1 on failure.
 **/

int
print_cvec(char *str, char *vec, int len)
{
  int i;

  printf("%s = ",str);
  for(i=0;i<len;i++)
    printf("%c ", vec[i]);
  printf("\n");

  return 0;
}
