% mandelbrot_gs(xa,dx,xb,ya,dy,yb,threshold,nb_block,max_requests)
% 
% threshhold: Limit for the Mandelbrot iterations
% nb_block: number of blocks
% max_requests: number of concurrent requests
%
% Sample call 
% mandelbrot_gs(-2,0.02,2,-2,0.02,2,60,3,3)
%
% Doing the GridSolve calls directly for testing purposes
% sid = gs_call_async('mandel',32.000000,0.000000,0.005000,0.500000,0.000000,0.005000,0.505000,1,10302)
% [r] = gs_wait(sid);
%

function mandelbrot_gs(xa,dx,xb,ya,dy,yb,threshold,nb_block,max_requests)

clf;
h = axes('XLim',[xa xb],'YLim',[ya yb]);
xlabel(['x-interval: [' num2str(xa) ', ' num2str(xb) ']']);
ylabel(['y-interval: [' num2str(ya) ', ' num2str(yb) ']']);
set(gca,'XTick',[]);
set(gca,'YTick',[]);
hold on;

requests = zeros(nb_block,nb_block);

startx = zeros(1,nb_block);
stopx = zeros(1,nb_block);
starty = zeros(1,nb_block);
stopy = zeros(1,nb_block);

xstep = ceil(ceil((xb-xa)/dx)/nb_block);
ystep = ceil(ceil((yb-ya)/dy)/nb_block);
for i=1:nb_block
  startx(i) = xa + (i-1)*xstep*dx;
  stopx(i) = min([startx(i)+xstep*dx xb]);
  starty(i) = ya + (i-1)*ystep*dy;
  stopy(i) = min([starty(i)+ystep*dy+dy yb]);
end

nblocks = nb_block^2;
requests = cell(2,max_requests);
block_sequence = randperm(nblocks);
req_issued = 0;

for i = 1:min(max_requests,nblocks)
  req_issued = req_issued + 1;
  requests{1,i} = block_sequence(req_issued);
  a = ceil(requests{1,i}/nb_block);
  b = mod(requests{1,i},nb_block);
  if b == 0 
    b = nb_block;
  end
  n1 = ceil((stopx(a)-startx(a))/dx)+1;
  n2 = ceil((stopy(b)-starty(b))/dy)+1;
  % fprintf('gs_call_async("mandel",%f,%f,%f,%f,%f,%f,%f,%d,%d)\n',threshold,startx(a),dx,stopx(a),starty(b),dy,stopy(b),1,n1*n2);
  requests{2,i} = gs_call_async('mandel',threshold,startx(a),dx,stopx(a),starty(b),dy,stopy(b),1,n1*n2);
  errnum = gs_get_last_error;
  if (errnum ~= 0)              % if error
    m = gs_error_string(errnum);
    fprintf('Error during mandelbrot call: %s\n', m);
    return;
  end
end

while (1==1)
  finished = 1;
  for i=1:max_requests
    if (requests{1,i} > 0)
      e = gs_probe(requests{2,i});
      if (e == 0)                       % GRPC_NO_ERROR

	
	a = ceil(requests{1,i}/nb_block);
	b = mod(requests{1,i},nb_block);
	if b == 0 
	  b = nb_block;
	end
	n1 = ceil((stopx(a)-startx(a))/dx)+1;
	n2 = ceil((stopy(b)-starty(b))/dy)+1;
	
	[r] = gs_wait(requests{2,i});
        
	r = reshape(r, n2, n1);
	subx = (startx(a):dx:stopx(a))';
	suby = (starty(b):dy:stopy(b))';
 
        % disable excessive warnings if a newer version of OpenGL is installed
        warning off;
	surface(subx,suby,r(1:size(suby,1),1:size(subx,1)));
	shading flat;
	drawnow;
	refresh;
        warning on;

	requests{1,i} = 0;

	if (req_issued < nblocks)
	  req_issued = req_issued + 1;
	  requests{1,i} = block_sequence(req_issued);
	  a = ceil(requests{1,i}/nb_block);
	  b = mod(requests{1,i},nb_block);
	  if b == 0
	    b = nb_block;
	  end
	  n1 = ceil((stopx(a)-startx(a))/dx)+1;
	  n2 = ceil((stopy(b)-starty(b))/dy)+1;
	  requests{2,i} = gs_call_async('mandel',threshold, startx(a),dx,stopx(a),starty(b),dy,stopy(b),1,n1*n2);
          errnum = gs_get_last_error;
	  if (errnum ~= 0)
	    m = gs_error_string(netsolve_err);
	    fprintf('Error during mandelbrot call async (2): %s\n', m);
	    return;
	  end
	end
      end
    end
  end
  
  finished = 1;
  for i=1:1:min(max_requests,nblocks)
    if (requests{1,i} ~= 0)
      finished = 0;
    end
  end
  
  if (finished == 1)
    break;
  end
  pause(1)
end

hold off;
 
return;

