function varargout = mandelbrot_gui(varargin)
% The Mandelbrot set consists of all of those (complex) c-values
% for which the iteration z_{n+1} = z^2 + c does not escape to
% infinity. 
% The set is illustrated through colorful plots. The user may zoom
% into the figure using the MATLAB's zoom function, and redraw the
% figure in a refined resolution.
%
% Use the figure zoom buttons to enlarge an area, and click the 
% Redraw button to compute a refined image of this area.
% Click the Reset button to restart with the given computation grid
% ranges.
% 
% Start the GUI with mandelbrot_gui from the MATLAB prompt.

% Last Modified by GUIDE v2.5 22-Oct-2004 11:11:22

% Modified from original code by Andreas Klimke, 
% Stuttgart University (v 1.1; May 3, 2003)


if nargin == 0  % LAUNCH GUI
  
  fig = openfig(mfilename,'reuse');
  
  % Use system color scheme for figure:
  set(fig,'Color',get(0,'defaultUicontrolBackgroundColor'));
  
  % Generate a structure of handles to pass to callbacks, and store it. 
  handles = guihandles(fig);
  guidata(fig, handles);
  set(handles.cmap,'Value',10);
  set(fig,'Units','pixels');

  %resetbutton_Callback([],[],handles);
  s = get(handles.cmap,'String');
  colormap(s{get(handles.cmap,'Value')});
  resolution = str2double(get(handles.resolution,'String'));
  iter = str2double(get(handles.iter,'String'));
  %escape = str2double(get(handles.escape,'String'));
  escape = 2.0;
  dx1 = str2double(get(handles.dx1,'String'));
  dx2 = str2double(get(handles.dx2,'String'));
  dy1 = str2double(get(handles.dy1,'String'));
  dy2 = str2double(get(handles.dy2,'String'));
  set(handles.dx1,'UserData',[dx1, dx2, dy1, dy2, resolution]);
  xlabel(['x-interval: [' num2str(dx1) ', ' num2str(dx2) ']']);
  ylabel(['y-interval: [' num2str(dy1) ', ' num2str(dy2) ']']);
  set(gca,'XTick',[]);
  set(gca,'YTick',[]);

  guiviewsize = get(fig,'Position');
  screensize = get(0,'ScreenSize');
  screensize = screensize(3:4);
  viewsize = [400 300];
  viewoffset = [50 200];
  set(fig,'Position', [screensize(1)-viewsize(1)- ...
	  2*viewoffset(1)-guiviewsize(3), screensize(2)- ...
	  viewoffset(2)-guiviewsize(4), ...
	  guiviewsize(3:4)]);
  set(gcf,'Position',[screensize(1)-viewoffset(1)-viewsize(1), ...
	  screensize(2)-viewoffset(2)-viewsize(2), viewsize])
  if nargout > 0
    varargout{1} = fig;
  end
  
elseif ischar(varargin{1}) % INVOKE NAMED SUBFUNCTION OR CALLBACK
  
  try
  [varargout{1:nargout}] = feval(varargin{:}); % FEVAL switchyard
  catch
  disp(lasterr);
end

end

% --------------------------------------------------------------------
function varargout = edit1_Callback(h, eventdata, handles, varargin)
return;

% --------------------------------------------------------------------
function varargout = popupmenu1_Callback(h, eventdata, handles, varargin)
% Stub for Callback of the uicontrol handles.popupmenu1.
s = get(handles.cmap,'String');
colormap(s{get(handles.cmap,'Value')});

% --------------------------------------------------------------------
function varargout = resetbutton_Callback(h, eventdata, handles, ...
    varargin)

s = get(handles.cmap,'String');
colormap(s{get(handles.cmap,'Value')});
resolution = str2double(get(handles.resolution,'String'));
iter = str2double(get(handles.iter,'String'));
escape = str2double(get(handles.escape,'String'));
dx1 = str2double(get(handles.dx1,'String'));
dx2 = str2double(get(handles.dx2,'String'));
dy1 = str2double(get(handles.dy1,'String'));
dy2 = str2double(get(handles.dy2,'String'));
set(handles.dx1,'UserData',[dx1, dx2, dy1, dy2, resolution]);
mandelbrot_wrap(handles, resolution, escape, iter, [dx1 dx2], [dy1 dy2]);
xlabel(['x-interval: [' num2str(dx1) ', ' num2str(dx2) ']']);
ylabel(['y-interval: [' num2str(dy1) ', ' num2str(dy2) ']']);
set(gca,'XTick',[]);
set(gca,'YTick',[]);

% --------------------------------------------------------------------
function varargout = redrawbutton1_Callback(h, eventdata, handles, ...
    varargin)

s = get(handles.cmap,'String');
colormap(s{get(handles.cmap,'Value')});
resolution = str2double(get(handles.resolution,'String'));
iter = str2double(get(handles.iter,'String'));
escape = str2double(get(handles.escape,'String'));
userd = get(handles.dx1,'UserData');
if isempty(userd)
  dx1 = str2double(get(handles.dx1,'String'));
  dx2 = str2double(get(handles.dx2,'String'));
  dy1 = str2double(get(handles.dy1,'String'));
  dy2 = str2double(get(handles.dy2,'String'));
  oldres = resolution;
else
  dx1 = userd(1);
  dx2 = userd(2);
  dy1 = userd(3);
  dy2 = userd(4);
  oldres = userd(5);
end
nx = get(gca,'XLim');
ny = get(gca,'YLim');
oldresy = oldres*(dy2-dy1)/(dx2-dx1);

% Note, the following 4 lines are commented out because mandelbrot_gs.m works
% on a different axis than the scalar mandelbrot.m example
%nx = nx -0.5; 
%nx = (nx/oldres)*(dx2-dx1)+dx1;
%ny = ny - 0.5;
%ny = (ny/oldresy)*(dy2-dy1)+dy1;

userd = [nx ny resolution];
set(handles.dx1,'UserData',userd);
mandelbrot_wrap(handles, resolution, escape, iter, nx, ny);
xlabel(['x-interval: [' num2str(nx(1),10) ', ' num2str(nx(2),10) ']']);
ylabel(['y-interval: [' num2str(ny(1),10) ', ' num2str(ny(2),10) ']']);
set(gca,'XTick',[]);
set(gca,'YTick',[]);

% --------------------------------------------------------------------
function varargout = helpbutton_Callback(h, eventdata, handles, varargin)
helpwin('mandelbrot_gui.m');

% --------------------------------------------------------------------

function mandelbrot_wrap(handles, resolution, escape, iter, dx, dy)

xa = dx(1);
xb = dx(2);
dx = (xb-xa)/resolution;
ya = dy(1);
yb = dy(2);
dy = (yb-ya)/round(resolution*(yb-ya)/(xb-xa));
nb_block = str2double(get(handles.nblocks,'String'));
max_requests = str2double(get(handles.ncalls,'String'));
threshold = iter;

mandelbrot_gs(xa,dx,xb,ya,dy,yb,threshold,nb_block,max_requests)

% --------------------------------------------------------------------
