/**
 * @file
 *
 * This program performs ping-pong bandwidth tests using the NetSolve API.
 * The program is invoked as follows:
 * 
 *   rtt_test <lo> <hi> <inc> <reps>
 *
 * Transmits arrays ranging in size from <lo> to <hi>
 * in increments of <inc>.  The number of repetitions
 * each size is specified by <reps>
 *
 * The output consists of two columns of numbers.  The first number is
 * the size in bytes of the data sent.  The second number is the wall time
 * for the entire transaction (including contacting the agent for the list
 * of servers).
 */
/* $Id: rtt_test.c,v 1.10 2005/05/25 18:06:42 seymour Exp $ */
/* $UTK_Copyright: $ */

#include "config.h"

#include<stdio.h>
#include<stdlib.h>
#ifdef HAVE_VALUES_H
#include<values.h>
#else
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_FLOAT_H
#include <float.h>
#define MAXDOUBLE DBL_MAX
#endif
#endif
#include<sys/time.h>
#include "netsolve.h"
#include "netsolveclient.h"
#include "totaltest.h"

double elapsed_time(const struct timeval*, const struct timeval*);

int 
main(int argc, char** argv)
{
  int *A, i, j, lo, hi, incr, rep, status, prof_enabled;
  struct timeval t1, t2;
  double et, min_time;
  NS_profile *nsprof, *min_prof;

  if(argc != 5) {
    fprintf(stderr,"\nUsage: %s <lo> <hi> <inc> <reps>\n", argv[0]);
    fprintf(stderr,"\n  Transmits arrays ranging in size from <lo>..<hi>\n");
    fprintf(stderr,"  in increments of <inc>.  The number of repetitions\n");
    fprintf(stderr,"  at each size is specified by <reps>.\n");
    die(EXIT_FAILURE);
  }

  lo = atoi(argv[1]);
  hi = atoi(argv[2]);
  incr = atoi(argv[3]);
  rep = atoi(argv[4]);
  prof_enabled = 0;

  netslmajor("col");

  A = (int*)malloc(hi*sizeof(*A));
  nsprof = (NS_profile *)calloc(rep, sizeof(NS_profile));

  for (i=0; i<hi; i++) A[i]=i;

  for(i=lo;i<=hi;i+=incr) {
    min_time = MAXDOUBLE;
    min_prof = nsprof;

    for(j=0;j<rep;j++) {
 
      prof_enabled = netsl_profile(&(nsprof[j])) == 0;

      gettimeofday(&t1, 0);
      status = netsl("vpass_int()", A, i);
      gettimeofday(&t2, 0);

      if(status < 0) {
        netslerr(status);
        die(EXIT_FAILURE);
      }

      et = elapsed_time(&t1,&t2);
      
      if(et < min_time) {
        min_time = et;
        min_prof = nsprof + j;
      }
    }

    if(prof_enabled)
      printf("%d: %g %g %g %g %g %g %g\n",(int)(i*sizeof(*A)),min_time,
        min_prof->proxy_start, min_prof->object_init, min_prof->agent_comm,
        min_prof->send_input, min_prof->job_complete, min_prof->recv_output);
    else
      printf("%d: %g\n",(int)(i*sizeof(*A)),min_time);
  }

  free(A);
  die(EXIT_SUCCESS);
}

/**
 * Computes the elapsed wall time between the two given 
 * times (timeval structs).
 *
 * @param t0 -- starting time
 * @param t1 -- ending time
 *
 * @returns the elapsed time between the two times
 **/

double 
elapsed_time(const struct timeval* t0, const struct timeval* t1)
{
  double ret = t1->tv_sec - t0->tv_sec;
  double uret = t1->tv_usec - t0->tv_usec;
  return ret + uret / 1E6;
}
