/**
 * @file
 *
 * -- SmartGridRPC implementation of mm1_gs.c --
 *
 * Application executes the matrix chain product.
 * Given x matrices A1, A2, ...., Ax of size n*n, the matrix
 * chain product is to compute A1 * A2 *, ..., * Ax.
 * This application obtains the matrix product using the 
 * standard binary tree method.  
 *
 * For example if x=4 then the first set of matrix operations
 * would consist of 
 * 
 * A1 * A2 = A5 and A3 * A4 = A6
 *
 * And the second set of operations would consist of
 * 
 * A5 * A6 = A7
 *  
 * where A7 would be the matrix product of A1 * A2 * A3 * A4.
 *
 * -----------------------------------------------------------
 * 
 * When this program is run the outputs of one set of matrix
 * multiplications will be the input to the next set. So using
 * the SmartGridRPC implementation the outputs will be sent
 * directly to the destination servers. Or when the source
 * server and destination server are the same then the output 
 * will be cached.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "gs_smart_clib.h"
#include "grpc.h"
#include "totaltest.h"


double
elapsed_time(const struct timeval*, const struct timeval*);


int
main(int argc, char *argv[])
{
  grpc_function_handle_t * handles;
  int * status;

  struct timeval t1, t2;
  int i, j, n,  lda, ldb, ldc,  x, start_mat, end_mat, mul, nb_mat_mul, nb_sets;
  int task_nb, start_task, end_task, rem, total_mat, total_tasks, nb_mat_this_set;
  double **a, et;
  char * mapper;
  char transa = 'T';
  char transb = 'N';
  double alpha = 1.0;
  double beta = 0.0;
  if(argc!=4){
    fprintf(stderr," Error: Incorrect argument parameters. \n Usage: ./mm1_smart mapper x n\n mapper is the name of the mapping heuristic (rwalk_map, greedy_map, basic_map)\nx is the number of matrix multiplications in the matrix chain. x must be a power of 2 (i.e. 2, 4, 8, 16, 32 ..) \n n is the dimensions of the matrices\n");
    exit(1); 
  }
  mapper=argv[1];
  x=atoi(argv[2]);
  n=atoi(argv[3]);

  i=0;
  if(x<2){
    fprintf(stderr," Error: Incorrect argument parameters. \n Usage: ./mm1_smart mapper x n\n mapper is the name of the mapping heuristic (rwalk_map, greedy_map, basic_map)\nx is the number of matrix multiplications in the matrix chain. x must be a power of 2 (i.e. 2, 4, 8, 16, 32 ..) \n n is the dimensions of the matrices\n");
    exit(1); 
  }

  printf("x=%d\n", x);

  /* 
   * check to make sure x is a power of 2
   * and calculate the total number of matrices required
   * and the total number of tasks that will be executed.
   */
  total_mat=0;
  total_tasks=0;
  rem=x;
  while(rem>=0){
    total_mat=total_mat+rem;
    total_tasks=total_tasks+rem/2;
    if(rem%2!=0) break;
    rem=rem/2;
  }
  printf("total_mat=%d\n", total_mat);
  printf("total_tasks=%d\n", total_tasks);
  /* if x is not a power of two */
  if(rem!=1){
    fprintf(stderr," Error: Incorrect argument parameters. \n Usage: ./mm1_smart mapper x n\n mapper is the name of the mapping heuristic (rwalk_map, greedy_map, basic_map)\nx is the number of matrix multiplications in the matrix chain. x must be a power of 2 (i.e. 2, 4, 8, 16, 32 ..) \n n is the dimensions of the matrices\n");
    exit(1); 


  }


  /*
   * Allocate memory to sessionId, handles, status array and matrix array (a).
   */ 
  grpc_sessionid_t sessionIDs[total_tasks];
  handles=(grpc_function_handle_t *)(calloc(total_tasks, sizeof(grpc_function_handle_t )));
  status=(int *)(calloc(total_tasks, sizeof(int)));
  a=(double **)(calloc(total_mat, sizeof(double *)));
  
  


  if(grpc_initialize(NULL) != GRPC_NO_ERROR) { 
    grpc_perror("grpc_initialize");
    die(EXIT_FAILURE);
  }

  /*
   * Set up handles
   */ 

   for(i=0; i<total_tasks;i++){
     if(grpc_function_handle_init(&handles[i], "bind_servers_at_call_time", "dgemm") != GRPC_NO_ERROR){
        fprintf(stderr,"Error creating function handle\n");
        die(EXIT_FAILURE);
     }

   }

  /*
   * allocate memory for each matrix in matrix array
   */ 
  for(i=0;i<total_mat;i++){
   a[i] = (double*)malloc(n*n*sizeof(double));

  }

  /*
   * initialize the first x matrices
   */ 
  
  for(i=0;i<x;i++){
    for(j=0;j<n*n;j++){
      a[i][j]=i+j;
    }
  }

  gettimeofday(&t1, 0);
  grpc_map(mapper){
    /* number of sets of matrix muliplications */
    nb_sets=0;
    /* number of matrices in next set of matrix multiplications */
    nb_mat_this_set=0;
    /* number of matrix multiplications performed in current set */
    nb_mat_mul=0;
    /* this is the task nb currently executing */
    task_nb=0;
    /* the first and last task in the current matrix set */
    start_task=0, end_task=0;
    lda = n;
    ldb = n;
    ldc=n;
    /* value used to determine the number of matrices in next set */
    mul=pow(2, nb_sets);
    
    nb_sets++;
    start_mat=0;
    nb_mat_this_set=x/mul;
    /* 
     * while there are more than one matrix in the 
     * next set of matrices then perform multiplications
     * if there is only one matrix left then this
     * matrix is the result of the chain product
     */
    while(nb_mat_this_set>1){
      /* number of matrices in the next matrix set */
      nb_mat_this_set=x/mul;
      /* id of the last matrix in the matrix set */
      end_mat=start_mat+nb_mat_this_set;
     
      nb_mat_mul=0;
      start_task=task_nb;
      for(i=start_mat;i<end_mat;i=i+2){
        printf("a[%d]*a[%d]=a[%d]\n", i, i+1, end_mat+nb_mat_mul);
        status[task_nb] = grpc_call_async(&handles[task_nb],&(sessionIDs[task_nb]),transa,transb,n,n,n,alpha,a[i],lda,a[i+1],ldb,beta,a[end_mat+nb_mat_mul],ldc);

        if(status[task_nb] != GRPC_NO_ERROR) {
          grpc_perror("grpc_call");
          exit(-1);
        }

        task_nb++;
        nb_mat_mul++;
      }
      end_task=task_nb;
      grpc_local(){
        printf("Waiting for tasks %d-%d to finish.\n", start_task, end_task);
        for(i=start_task;i<end_task;i++){
          if(status[i] != GRPC_NO_ERROR) {
            grpc_perror("grpc_call");
            return -1;
          }
          status[i] = grpc_wait(sessionIDs[i]);

          printf("%d completed\n", i);

          if(status[i] != GRPC_NO_ERROR) {
            printf("Warning: one or more tasks failed.\n");
            exit(-1);
          } 
        }
      }

      mul=pow(2, nb_sets);
      start_mat=end_mat;
    nb_mat_this_set=x/mul;
        

      nb_sets++;
    }
  } 
  gettimeofday(&t2, 0);
  et = elapsed_time(&t1,&t2);
  
  printf("Elapsed time to get matrix chain product of %d (x) matrices : %g\n", x, et);
  printf("Matrix dimensions = %d*%d\n", n, n);
  printf("Finalizing...\n");
  grpc_finalize();
  free(handles);
  free(status);
  for(i=0;i<total_mat;i++){
    free(a[i]);
  }
  free(a);
  die(EXIT_SUCCESS);
}


double
elapsed_time(const struct timeval* t0, const struct timeval* t1)
{
  double ret = t1->tv_sec - t0->tv_sec;
  double uret = t1->tv_usec - t0->tv_usec;
  return ret + uret / 1E6;
}
