/**
 * @file
 *
 * -- GridRPC implementation --
 *
 * This program makes a number (n) of calls to the "dgesv" 
 * routine.
 * The first set of calls are a set of blocking calls to 
 * this routine where the tasks ascend in size.
 * The second set of calls are a set of non-blocking calls 
 * and the tasks descend in size.
 * The third set of calls are non-blocking calls
 * and the tasks ascend in size. 
 * The fourth set are blocking calls and the task
 * descend in size
 *
 * The input x and y both determine the sizes of matrices. 
 * The value of x specifies the starting size and y specifies
 * how quickly the size of the matrices either increase or
 * decrease.
 * There are no dependencies between any tasks so this 
 * program is ideal for testing different mapping heuristics 
 * and timing their task-to-server mappings. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "gs_smart_clib.h"
#include "grpc.h"
#include "totaltest.h"


double
elapsed_time(const struct timeval*, const struct timeval*);


int
main(int argc, char *argv[])
{
  grpc_function_handle_t * handles;
 int * status;
  int x, y;

  struct timeval t1, t2;
  int i, n, nrhs, lda, **ipiv, ldb, info, num_iters;
  double **a, **b, et;
  int iter;

  if(argc!=4){
    fprintf(stderr," Error: Incorrect argument parameters. \n Usage: ./mm6_gs n x y\n n is the number of matrix multiplications to perform\n x and y are the parameters which determine the matrix sizes\n matrix dimensions = x*(i+1)^y (where 0<i<n) \n");
    exit(1); 
  }
  num_iters=atoi(argv[1]);
  x=atoi(argv[2]);
  y=atoi(argv[3]);

  if(num_iters<1){
    fprintf(stderr," Error: Incorrect argument parameters. \n Usage: ./mm6_gs n x y\n n is the number of matrix multiplications to perform\n x and y are the parameters which determine the matrix sizes\n matrix dimensions = x*(i+1)^y (where 0<i<n) \n");
    exit(1); 
  }
 
  grpc_sessionid_t sessionIDs[num_iters]; 
  handles=(grpc_function_handle_t *)(calloc(num_iters, sizeof(grpc_function_handle_t )));
  status=(int *)(calloc(num_iters, sizeof(int)));
  a=(double **)(calloc(num_iters, sizeof(double *)));
  b=(double **)(calloc(num_iters, sizeof(double *)));
  ipiv=(int **)(calloc(num_iters, sizeof(int *)));

  if(grpc_initialize(NULL) != GRPC_NO_ERROR) {
    grpc_perror("grpc_initialize");
    die(EXIT_FAILURE);
  }
  i=0;
  for(iter=0; iter<num_iters;iter++){
     if(grpc_function_handle_init(&handles[iter], "bind_servers_at_call_time", "dgesv") != GRPC_NO_ERROR)
    {
        fprintf(stderr,"Error creating function handle\n");
        die(EXIT_FAILURE);
    }
  }

  for(iter=0; iter<num_iters/4;iter++){
    n=x * pow((iter+1), y);
    a[iter] = (double*)malloc(n*n*sizeof(double));
    b[iter]= (double*)malloc(n*n*sizeof(double));
    ipiv[iter] = (int*)malloc(n*sizeof(int));
    for(i=0;i<n*n;i++) {
        a[iter][i] = (double)i;
        b[iter][i] = (double)(n+i);
    }
    for(i=0;i<n;i++)
      ipiv[iter][i] = n-i;
  }
  
  for(iter=num_iters/4;iter<num_iters/2;iter++ ){       
    n=x * pow(((num_iters/2)-(iter)), y);
    a[iter] = (double*)malloc(n*n*sizeof(double));
    b[iter]= (double*)malloc(n*n*sizeof(double));
    ipiv[iter] = (int*)malloc(n*sizeof(int));
    for(i=0;i<n*n;i++) {
        a[iter][i] = (double)i;
        b[iter][i] = (double)(n+i);
    }
    for(i=0;i<n;i++)
      ipiv[iter][i] = n-i;

  }	  


  for(iter=num_iters/2;iter<(num_iters/4+num_iters/2);iter++){
    n=x * pow((iter+1), y);
    a[iter] = (double*)malloc(n*n*sizeof(double));
    b[iter]= (double*)malloc(n*n*sizeof(double));
    ipiv[iter] = (int*)malloc(n*sizeof(int));
    for(i=0;i<n*n;i++) {
        a[iter][i] = (double)i;
        b[iter][i] = (double)(n+i);
    }
    for(i=0;i<n;i++)
      ipiv[iter][i] = n-i;


  }
  int j=0;
  for(iter= (num_iters/4+num_iters/2);iter<num_iters;iter++ ){       
    n=x * pow((num_iters-j), y);
    a[iter] = (double*)malloc(n*n*sizeof(double));
    b[iter]= (double*)malloc(n*n*sizeof(double));
    ipiv[iter] = (int*)malloc(n*sizeof(int));
    for(i=0;i<n*n;i++) {
        a[iter][i] = (double)i;
        b[iter][i] = (double)(n+i);
    }
    for(i=0;i<n;i++)
      ipiv[iter][i] = n-i;


    j++;
  }




  gettimeofday(&t1, 0);
  for(iter=0;iter<num_iters/4;iter++){
    n=x * pow((iter+1), y);
    nrhs = n;
    lda = n;
    ldb = n;
    status[iter] = grpc_call(&handles[iter], n, nrhs, a[iter], lda, ipiv[iter], b[iter], ldb, &info);
  }
  for(iter=num_iters/4;iter<num_iters/2;iter++ ){       
    n=x * pow(((num_iters/2)-(iter)), y);
    nrhs = n;
    lda = n;
    ldb = n;
    status[iter] = grpc_call_async(&handles[iter], &(sessionIDs[iter]), n, nrhs, a[iter], lda, ipiv[iter], b[iter], ldb, &info);
    if(status[iter] != GRPC_NO_ERROR) {
      grpc_perror("grpc_call");
      exit(-1);
    }
  } 
  printf("Waiting for tasks to finish.\n");
  for(iter=num_iters/4;iter<num_iters/2;iter++){       
    if(status[iter] != GRPC_NO_ERROR) {
      grpc_perror("grpc_call");
      return -1;
    }
    status[iter] = grpc_wait(sessionIDs[iter]);

    printf("%d completed\n", iter);

    if(status[iter] != GRPC_NO_ERROR) {
      printf("Warning: one or more tasks failed.\n");
      exit(-1);
    }
  }
  

  j=0;
  for(iter= (num_iters/4+num_iters/2);iter<num_iters;iter++ ){       
    n=x * pow((num_iters-j), y);
    nrhs = n;
    lda = n;
    ldb = n;
    status[iter] = grpc_call_async(&handles[iter], &(sessionIDs[iter]), n, nrhs, a[iter], lda, ipiv[iter], b[iter], ldb, &info);

    if(status[iter] != GRPC_NO_ERROR) {
      grpc_perror("grpc_call");
      exit(-1);
    }
    j++;
  }
  for(iter=num_iters/2;iter<(num_iters/4+num_iters/2);iter++){
    n=x * pow((iter+1), y);
    nrhs = n;
    lda = n;
    ldb = n;
    status[iter] = grpc_call(&handles[iter], n, nrhs, a[iter], lda, ipiv[iter], b[iter], ldb, &info);
  }
  printf("Waiting for tasks to finish.\n");
  for(iter= (num_iters/4+num_iters/2);iter<num_iters;iter++ ){       
    if(status[iter] != GRPC_NO_ERROR) {
      grpc_perror("grpc_call");
      return -1;
    }

    status[iter] = grpc_wait(sessionIDs[iter]);

    printf("%d completed\n", iter);

    if(status[iter] != GRPC_NO_ERROR) {
      printf("Warning: one or more tasks failed.\n");
      exit(-1);
    }
  }

  for(iter=0;iter<num_iters;iter++){
     grpc_function_handle_destruct(&handles[iter]);
  }

  gettimeofday(&t2, 0);
  et = elapsed_time(&t1,&t2);
  
  printf("Elapsed time for %d (n) matrix multiplications : %g\n", num_iters, et);
  printf("Matrix dimensions = %d*(i+1)^%d (where 0<i<%d)", x, y, num_iters);
  printf("Finalizing...\n");

  grpc_finalize();

  printf("Done.\n");

  free(handles);
  free(status);
  for(i=0;i<num_iters;i++){
    free(ipiv[i]);
    free(a[i]);
    free(b[i]);
  }
  free(ipiv);
  free(a);
  free(b);
  die(EXIT_SUCCESS);
}


double
elapsed_time(const struct timeval* t0, const struct timeval* t1)
{
  double ret = t1->tv_sec - t0->tv_sec;
  double uret = t1->tv_usec - t0->tv_usec;
  return ret + uret / 1E6;
}
