/**
 * @file
 *
 * This file contains routines for performing conversions between
 * different machine representations.
 *
 * The arrays of function pointers are indexed by the byte order
 * part of the data signature such that the value at (m,n) is the
 * function that converts between byte order 'm' and byte order 'n'.
 *
 */
/* $Id: data_conversion.c,v 1.6 2005/11/10 20:46:08 seymour Exp $ */
/* $UTK_Copyright: $ */

#include "utility.h"

/** Array of integer byte swapping functions */

gs_int_swap_f gs_int_swap_func[4][4] = {
   {gs_swap_int_nop,   gs_swap_int_undef, gs_swap_int_undef, gs_swap_int},
   {gs_swap_int_undef, gs_swap_int_nop,   gs_swap_int_undef, gs_swap_int_undef},
   {gs_swap_int_undef, gs_swap_int_undef, gs_swap_int_nop,   gs_swap_int_undef},
   {gs_swap_int,       gs_swap_int_undef, gs_swap_int_undef, gs_swap_int_nop}};

/** Array of float byte swapping functions */

gs_float_swap_f gs_float_swap_func[4][4] = {
   {gs_swap_float_nop,   gs_swap_float_undef, gs_swap_float_undef, gs_swap_float},
   {gs_swap_float_undef, gs_swap_float_nop,   gs_swap_float_undef, gs_swap_float_undef},
   {gs_swap_float_undef, gs_swap_float_undef, gs_swap_float_nop,   gs_swap_float_undef},
   {gs_swap_float,       gs_swap_float_undef, gs_swap_float_undef, gs_swap_float_nop}};

/** Array of double byte swapping functions */

gs_double_swap_f gs_double_swap_func[4][4] = {
   {gs_swap_double_nop,   gs_swap_double_undef, gs_swap_double_undef, gs_swap_double},
   {gs_swap_double_undef, gs_swap_double_nop,   gs_swap_double_undef, gs_swap_double_undef},
   {gs_swap_double_undef, gs_swap_double_undef, gs_swap_double_nop,   gs_swap_double_undef},
   {gs_swap_double,       gs_swap_double_undef, gs_swap_double_undef, gs_swap_double_nop}};

/**
 * Swaps the bytes in an integer.
 *
 * @param x -- pointer to the integer to be swapped
 *
 * @returns 0 on success, -1 on failure
 */

int
gs_swap_int(int *x)
{
  switch(sizeof(int)) {
    case 4:
      *x = ((*x & 0xFF) << 24) |
           ((*x & 0xFF00) << 8) |
           ((*x & 0xFF0000) >> 8) |
           ((*x & 0xFF000000) >> 24);
      return 0;
    case 2:
      *x = ((*x & 0xFF) << 8) | 
           ((*x & 0xFF00) >> 8);
      return 0;
    default:
      ERRPRINTF("Error: unexpected int size\n");
      return -1;
  }
}

/**
 * Does nothing.  This is used along the diagonal of the
 * conversion matrix since that represents a conversion
 * between the same representations.
 *
 * @param x -- pointer to the integer to be swapped
 *
 * @returns 0 on success, -1 on failure.
 *   currently always returns 0.
 */

int
gs_swap_int_nop(int *x)
{
  return 0;
}

/**
 * Does nothing.  This is used in the conversion matrix
 * in places where we have not implemented a conversion
 * between the two representations.
 *
 * @param x -- pointer to the integer to be swapped
 *
 * @returns 0 on success, -1 on failure.
 *   currently always returns -1.
 */

int
gs_swap_int_undef(int *x)
{
  ERRPRINTF("Error: this conversion is undefined\n");
  return -1;
}

/**
 * Swaps the bytes in a float.
 *
 * @param x -- pointer to the float to be swapped
 *
 * @returns 0 on success, -1 on failure
 */

int
gs_swap_float(float *f)
{
  int i, size = sizeof(float)-1;
  char *p, tmp;

  p = (char*)f;

  for(i=size;i>size/2;i--) {
    tmp = p[i];
    p[i] = p[size-i];
    p[size-i] = tmp;
  }

  return 0;
}

/**
 * Does nothing.  This is used along the diagonal of the
 * conversion matrix since that represents a conversion
 * between the same representations.
 *
 * @param x -- pointer to the float to be swapped
 *
 * @returns 0 on success, -1 on failure.
 *   currently always returns 0.
 */

int
gs_swap_float_nop(float *x)
{
  return 0;
}

/**
 * Does nothing.  This is used in the conversion matrix
 * in places where we have not implemented a conversion
 * between the two representations.
 *
 * @param x -- pointer to the float to be swapped
 *
 * @returns 0 on success, -1 on failure.
 *   currently always returns -1.
 */

int
gs_swap_float_undef(float *x)
{
  ERRPRINTF("Error: this conversion is undefined\n");
  return -1;
}

/**
 * Swaps the bytes in a double.
 *
 * @param d -- pointer to the double to be swapped
 *
 * @returns 0 on success, -1 on failure
 */

int
gs_swap_double(double *d)
{
  int i, size = sizeof(double)-1;
  char *p, tmp;

  p = (char*)d;

  for(i=size;i>size/2;i--) {
    tmp = p[i];
    p[i] = p[size-i];
    p[size-i] = tmp;
  }

  return 0;
}

/**
 * Does nothing.  This is used along the diagonal of the
 * conversion matrix since that represents a conversion
 * between the same representations.
 *
 * @param d -- pointer to the double to be swapped
 *
 * @returns 0 on success, -1 on failure.
 *   currently always returns 0.
 */

int
gs_swap_double_nop(double *d)
{
  return 0;
}

/**
 * Does nothing.  This is used in the conversion matrix
 * in places where we have not implemented a conversion
 * between the two representations.
 *
 * @param d -- pointer to the double to be swapped
 *
 * @returns 0 on success, -1 on failure.
 *   currently always returns -1.
 */

int
gs_swap_double_undef(double *d)
{
  ERRPRINTF("Error: this conversion is undefined\n");
  return -1;
}
