/**
 * @file
 *
 * This file contains dynamic string handling functions.
 *
 */
/* $Id: dstring_sprintf.c,v 1.8 2004/08/16 21:28:34 seymour Exp $ */
/* $UTK_Copyright: $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "utility.h"
#include "xnprintf.h"

/** 
 * Taken from the Linux man page of snprintf.  To allocate a
 * sufficiently large string and print into it (code correct for both
 * glibc 2.0 and glibc 2.1).  This string needs to be deallocated
 * manually.
 *
 * @param fmt -- the format string (same as printf)
 * @param ... -- the arguments to be printed
 *
 * @returns a dynamically allocated string containing the output of
 *   the sprintf.
 */

char *dstring_sprintf(const char *fmt, ...) {
  char *buf;
  int status;
  va_list ap;

  va_start (ap, fmt);
  status = vasprintf (&buf, fmt, ap);
  va_end (ap);
  
  return(buf);
}

#define DO_FREE 1
#define DO_NOT_FREE 0

/**
 * Common routine for the dstring_append functions.
 * Appends s2 to s1.
 *
 * @param s1 -- the string to which we will append
 * @param s2 -- the string to append to s1
 * @param do_free -- if TRUE, free s2 after appending.
 *   otherwise,  do not free s2.
 *
 * @returns pointer to the new appended string
 */

char *dstring_append_common(char *s1, char* s2, int do_free) 
{
  if ((s1 = realloc(s1, strlen(s1)+strlen(s2)+1)) == NULL) 
    return NULL;
  s1 = strcat(s1,s2);
  if (do_free && s2!=NULL) free(s2);
  return(s1);
}

/**
 * Appends s2 to s1 without freeing s2.
 *
 * @param s1 -- the string to which we will append
 * @param s2 -- the string to append to s1
 * 
 * @returns pointer to the new appended string
 */

char *dstring_append(char *s1, char *s2)
{
  return(dstring_append_common(s1,s2,DO_NOT_FREE));
}

/**
 * Appends s2 to s1 and frees s2.
 *
 * @param s1 -- the string to which we will append
 * @param s2 -- the string to append to s1
 * 
 * @returns pointer to the new appended string
 */

char *dstring_append_free(char *s1, char* s2) 
{
  return(dstring_append_common(s1,s2,DO_FREE));
}
