/**
 * @file
 *
 * This file contains various routines that help in portability to
 * multiple platforms.  Especially, there is a WIN32 section that
 * implements various functions that are missing on WIN32.
 *
 */
/* $Id: portability.c,v 1.4 2006/03/20 20:20:37 seymour Exp $ */
/* $UTK_Copyright: $ */

#include <stdio.h>

#include "portability.h"
#include "utility.h"


/* -------------------------------------------------- */
#ifdef WIN32
/* -------------------------------------------------- */

int gettimeofday(struct timeval* tp, void* tzp) 
{
  DWORD t;
  t = timeGetTime();
  tp->tv_sec = t / 1000;
  tp->tv_usec = t % 1000;
  /* 0 indicates that the call succeeded. */
  return 0;
}

int kill(int pid, int sig) 
{
  int status;
  HANDLE pidh = OpenProcess(PROCESS_ALL_ACCESS, TRUE, pid);
  status = TerminateProcess(pidh, (UINT)0);
  CloseHandle(pidh);
  if (status == 0) 
    return -1;
  else return 0;
}

int waitpid(int pid, int* status, int options)
{
  int retval;
  HANDLE pidh = OpenProcess(PROCESS_ALL_ACCESS, TRUE, pid);
  retval = WaitForSingleObject(pidh, options);
  CloseHandle(pidh);
  return retval;
}

int read_socket(SOCKET sock, const char*buffer, int len)
{
  /* DBGPRINTF("read_socket: sock %d len %d buffer[1] %d \n", sock, len, (int)*buffer); */
  return recv(sock, buffer, len, 0);
}

int write_socket(SOCKET sock, const char*buffer, int len)
{
  /* DBGPRINTF("write_socket: sock %d len %d buffer[1] %d \n", sock, len, (int)*buffer); */
  return send(sock, buffer, len, 0);
}

int close_socket(SOCKET sock)
{
  /* DBGPRINTF("close_socket\n"); */
  return closesocket(sock);
}

int getdomainname(char *name, size_t len)
{
  return -1;
}

int gs_lock_fd_nowait(int fd, int lock_type)
{
  ERRPRINTF("File locking is not implemented for Win32\n");
  return -1;
}

int gs_lock_fd(int fd, int lock_type) 
{
  ERRPRINTF("File locking is not implemented for Win32\n");
  return -1;
}

int gs_unlock_fd(int fd) 
{
  ERRPRINTF("File locking is not implemented for Win32\n");
  return -1;
}

void initialize_sockets()
{
  WORD wVersionRequested = MAKEWORD( 2, 0 );
  WSADATA wsaData;
  WSAStartup(wVersionRequested, &wsaData );
}    


void cleanup_sockets()
{
  WSACleanup();
}    



/* -------------------------------------------------- */
#else  /* not WIN32 */
/* -------------------------------------------------- */

int gs_lock_fd_nowait(int fd, int lock_type)
{
  struct flock fl;
  
  fl.l_type   = lock_type;
  fl.l_whence = SEEK_SET;
  fl.l_start  = 0;
  fl.l_len    = sizeof(int);
  fl.l_pid    = getpid();
  
  return fcntl(fd, F_SETLK, &fl);
}

int gs_lock_fd(int fd, int lock_type) 
{				
  struct flock fl;
  
  fl.l_type   = lock_type;
  fl.l_whence = SEEK_SET;
  fl.l_start  = 0;
  fl.l_len    = sizeof(int);
  fl.l_pid    = getpid();
  
  return fcntl(fd, F_SETLKW, &fl);
}

int gs_unlock_fd(int fd) 
{
  struct flock fl;

  fl.l_type   = F_UNLCK;
  fl.l_whence = SEEK_SET;
  fl.l_start  = 0;
  fl.l_len    = sizeof(int);
  fl.l_pid    = getpid();

  if(fcntl(fd, F_SETLK, &fl) == -1) {
    return -1;
  }

  return 0;
}


void initialize_sockets() {}    

void cleanup_sockets() {}    


#endif
